#include <glib.h>
#include <unistd.h>
#include <fcntl.h>
#include <errno.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <gst/gst.h>

#include <term_source.h>
#include <enums.h>

#define LAGSET_SRC_NAME "src"
#define INPUT_STAGE_SOCK_NAME "issink"

typedef struct pipeline {
    GstElement *pipeline;
    gchar *input_stage;
    gchar *name;
    gint sock;
    gboolean finished;
} pipeline_t;

static void parse_options(gint argc, gchar *argv[], gchar **input,
                          gchar ***pipeline_desc);
static void run_pipeline(pipeline_t *pipeline);
static gboolean bus_cb(GstBus *bus, GstMessage *msg, gpointer data);

gint
main(gint argc, gchar *argv[])
{
    gchar *input = NULL;
    gchar **pipeline_desc = NULL;
    pipeline_t pipeline = {.pipeline = NULL, .sock = -1};

    if (!g_thread_supported()) {
        g_thread_init(NULL);
    }
    enums_register();

    parse_options(argc, argv, &input, &pipeline_desc);
    GPtrArray *full_pipeline = g_ptr_array_new();
    g_ptr_array_add(full_pipeline, g_strdup("wcbe_lagset_src"));
    g_ptr_array_add(full_pipeline, g_strdup("name="LAGSET_SRC_NAME));
    gchar *insock = g_strdup_printf("input=%s", input);
    g_ptr_array_add(full_pipeline, insock);
    g_ptr_array_add(full_pipeline, g_strdup("!"));
    gchar **component = pipeline_desc;
    while (component != NULL && *component != NULL) {
        g_ptr_array_add(full_pipeline, g_strdup(*component));
        ++component;
    }
    g_ptr_array_add(full_pipeline, NULL);
    g_free(pipeline_desc);

    GError *err = NULL;
    pipeline.pipeline =
        gst_parse_launchv((const gchar **)full_pipeline->pdata, &err);
    if (err != NULL) {
        g_error("Failed to create pipeline: %s", err->message);
        if (pipeline.pipeline != NULL) {
            gst_object_unref(GST_OBJECT(pipeline.pipeline));
            pipeline.pipeline = NULL;
        }
    }
    gchar **pdata = (gchar **)g_ptr_array_free(full_pipeline, FALSE);
    g_strfreev(pdata);

    if (pipeline.pipeline != NULL) {
        // TODO: not sure of the encoding the g_path... functions work
        // on; ignored for now, need to experiment
        gchar *basename = g_path_get_basename(input);
        gchar *confdir = g_path_get_dirname(input);
        gchar *confname = g_path_get_basename(confdir);
        gchar *bedir = g_path_get_dirname(confdir);
        pipeline.input_stage =
            g_build_filename(bedir, INPUT_STAGE_SOCK_NAME, NULL);
        pipeline.name = g_build_filename(confname, basename, NULL);
        gst_element_set_name(pipeline.pipeline, basename);
        g_free(bedir);
        g_free(confname);
        g_free(confdir);
        g_free(basename);
        pipeline.sock = socket(PF_UNIX, SOCK_DGRAM, 0);
        if (pipeline.sock != -1) {
            gint rc = fcntl(pipeline.sock, F_SETFL, O_NONBLOCK);
            if (rc != 0) {
                GST_ERROR("Failed to set socket to non-blocking mode: %s",
                          g_strerror(errno));
            }
            run_pipeline(&pipeline);
            close(pipeline.sock);
        }
        else {
            g_printerr("Failed to open message passing socket: %s",
                       g_strerror(errno));
        }
        gst_object_unref(GST_OBJECT(pipeline.pipeline));
        g_free(pipeline.input_stage);
        g_free(pipeline.name);
    }
    return 0;
}

static void
parse_options(gint argc, gchar *argv[], gchar **input, gchar ***pipeline_desc)
{
    const GOptionEntry entries[] = {
        {"input", 'i', 0, G_OPTION_ARG_STRING, input,
         "Lag set source socket name", "NAME"},
        {NULL}
    };

    GError *error = NULL;
    GOptionContext *context = g_option_context_new(
        "PIPELINE - WIDAR correlator backend lag set pipeline");
    g_option_context_set_summary(
        context,
        "Lag set processing pipeline stage for compute "
        "node of WIDAR correlator backend system.\n"
        "PIPELINE argument should be given in the syntax "
        "used by gst-launch.");
    g_option_context_set_description(
        context,
        "Please report feature requests, bugs, etc. to\n"
        "mpokorny@nrao.edu.");
    g_option_context_set_help_enabled(context, TRUE);
    g_option_context_set_ignore_unknown_options(context, TRUE);
    g_option_context_add_main_entries(context, entries, NULL);
    g_option_context_add_group(context, gst_init_get_option_group());
    gboolean ctxt_parse =
        g_option_context_parse(context, &argc, &argv, &error);

    if (ctxt_parse) {
        GPtrArray *pipeline = g_ptr_array_new();
        guint i;
        for (i = 1; i < argc; ++i) {
            g_ptr_array_add(pipeline, argv[i]);
        }
        g_ptr_array_add(pipeline, NULL);
        *pipeline_desc = (gchar **)g_ptr_array_free(pipeline, FALSE);
    }

    if (!ctxt_parse) {
        if (error != NULL) {
            g_printerr("%s\n", error->message);
            g_error_free(error);
        }
        exit(-1);
    }
    g_option_context_free(context);
}

static void
run_pipeline(pipeline_t *pipeline)
{
    GstBus *bus = NULL;

    bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->pipeline));
    if (bus == NULL) {
        g_printerr("Failed to get pipeline message bus");
        goto cleanup_and_return;
    }
    gst_bus_add_watch(bus, bus_cb, pipeline);

    GstStateChangeReturn ret =
        gst_element_set_state(pipeline->pipeline, GST_STATE_PLAYING);
    if (ret != GST_STATE_CHANGE_FAILURE) {
        GMainLoop *loop = g_main_loop_new(NULL, FALSE);

        // set up TermSource to catch termination signals
        pipeline->finished = FALSE;
        gboolean term_handler(gint *signum) {
            pipeline->finished = TRUE;
            return FALSE;
        }
        TermSource *term_source = term_source_new();
        term_source_set_callback(term_source, term_handler);
        term_source_attach(term_source, NULL);

        // On interrupt, try an orderly pipeline shutdown, then quit
        // main loop.
        gboolean done(gpointer data) {
            if (g_atomic_int_get(&(pipeline->finished))) {
                ret = gst_element_set_state(pipeline->pipeline,
                                            GST_STATE_NULL);
                if (ret == GST_STATE_CHANGE_ASYNC) {
                    gst_element_get_state(pipeline->pipeline, NULL, NULL,
                                          10 * GST_SECOND);
                }
                GstMessage *msg = gst_message_new_application(
                    NULL,
                    gst_structure_new(
                        "pipeline-state",
                        "value", PIPELINE_STATE_ENUM_TYPE,
                        PIPELINE_STATE_STOPPED,
                        NULL));
                if (!bus_cb(bus, msg, pipeline)) {
                    g_debug("Failed to post 'pipeline-state:value=stopped' "
                            "message to message bus");
                }
                gst_message_unref(msg);
                g_main_loop_quit(loop);
            }
            return TRUE;
        };
        g_timeout_add(100, done, NULL);

        GstMessage *msg = gst_message_new_application(
            NULL,
            gst_structure_new(
                "pipeline-state",
                "value", PIPELINE_STATE_ENUM_TYPE, PIPELINE_STATE_RUNNING,
                NULL));

        if (!bus_cb(bus, msg, pipeline)) {
            g_debug("Failed to post 'pipeline-state:value=running' message "
                    "to message bus");
        }
        gst_message_unref(msg);

        g_main_loop_run(loop);

        g_main_loop_unref(loop);
    }
    else {
        g_warning("Failed to start pipeline");
        GstMessage *msg = gst_bus_poll(bus, GST_MESSAGE_ERROR, 0);
        if (msg != NULL) {
            GError *error = NULL;
            gst_message_parse_error(msg, &error, NULL);
            g_printerr("ERROR: %s\n", error->message);
            g_error_free(error);
            gst_message_unref(msg);
        }
    }

 cleanup_and_return:
    if (bus != NULL) {
        gst_object_unref(GST_OBJECT(bus));
    }
}

static gboolean
bus_cb(GstBus *bus, GstMessage *msg, gpointer data)
{
    pipeline_t *pipeline = data;

    switch (GST_MESSAGE_TYPE(msg)) {
      case GST_MESSAGE_ERROR: {
        gchar *debug = NULL;
        GError *err = NULL;
        gst_message_parse_error(msg, &err, &debug);
        g_printerr("ERROR: %s\n", err->message);
        if (debug != NULL) {
            g_printerr("Debug details: %s\n", debug);
            g_free(debug);
        }
        g_error_free(err);
        break;
      }
      case GST_MESSAGE_EOS:
        g_atomic_int_set(&(pipeline->finished), TRUE);
        break;
      case GST_MESSAGE_APPLICATION: {
        GstStructure *structure =
            gst_structure_copy(gst_message_get_structure(msg));
        if (GST_MESSAGE_SRC(msg) != NULL) {
            gchar *src_name = gst_object_get_name(GST_MESSAGE_SRC(msg));
            gchar *src =
                g_strjoin(G_DIR_SEPARATOR_S, pipeline->name, src_name, NULL);
            gst_structure_set(structure,
                              "src", G_TYPE_STRING, src,
                              NULL);
            g_free(src_name);
            g_free(src);
        }
        else {
            gst_structure_set(structure,
                              "src", G_TYPE_STRING, pipeline->name,
                              NULL);
        }
        gchar *str = gst_structure_to_string(structure);
        guint str_len = strlen(str) + 1;
        struct sockaddr_un addr = {.sun_family = AF_UNIX};
        gint rc = g_strlcpy(addr.sun_path, pipeline->input_stage,
                            sizeof(addr.sun_path));
        g_assert(0 < rc && rc < sizeof(addr.sun_path));
        struct iovec iovec[] =
            {{.iov_base = &str_len, .iov_len = sizeof(str_len)},
             {.iov_base = str, .iov_len = str_len}};
        struct msghdr msghdr = {.msg_name = &addr, .msg_namelen = sizeof(addr),
                                .msg_control = NULL, .msg_controllen = 0,
                                .msg_iov = iovec, .msg_iovlen = 2};
        gssize bytes_written = sendmsg(pipeline->sock, &msghdr, 0);
        if (bytes_written != iovec[0].iov_len + iovec[1].iov_len) {
            g_warning("Failed to write application message (%s) to socket: %s",
                      str, g_strerror(errno));
        }
        g_free(str);
        gst_structure_free(structure);
        break;
      }
      default:
        break;
    }
    return TRUE;
}
