#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <signal.h>
#include <fcntl.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/wait.h>
#include <unistd.h>
#include <glib.h>
#include <glib/gstdio.h>
#include <gst/gst.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>

#include <term_source.h>
#include <enums.h>
#include <wcbe_app.h>
#include <wcbe.h>
#include <wcbe_src.h>
#include <wcbe_framedumper.h>
#include <wcbe_decoder.h>
#include <wcbe_sorter.h>
#include <wcbe_lagset_sink.h>


#ifdef G_HAVE_ISO_VARARGS
#define g_info(...)    g_log (G_LOG_DOMAIN,       \
                              G_LOG_LEVEL_INFO,   \
                              __VA_ARGS__)
#elif defined(G_HAVE_GNUC_VARARGS)
#define g_info(format...)      g_log (G_LOG_DOMAIN,         \
                                      G_LOG_LEVEL_INFO,     \
                                      format)
#else   /* no varargs macros */
static void
g_info (const gchar *format,
        ...)
{
  va_list args;
  va_start (args, format);
  g_logv (G_LOG_DOMAIN, G_LOG_LEVEL_INFO, format, args);
  va_end (args);
}
#endif  /* !__GNUC__ */
#define CONFIG_TIMEOUT_MS 5000
#define DELAYED_ACTIVATE_INTERVAL 250

#define DEFAULT_PIPELINE_PIPELINE NULL
#define DEFAULT_PIPELINE_NAME NULL
#define DEFAULT_PIPELINE_SRC NULL
#define DEFAULT_PIPELINE_CONFIGS NULL
#define DEFAULT_PIPELINE_OLD_CONFIGS NULL
#define DEFAULT_PIPELINE_PENDING_CONFIGS NULL
#define DEFAULT_PIPELINE_CONFIG NULL
#define DEFAULT_PIPELINE_LIVE_TS FALSE
#define DEFAULT_PIPELINE_FRAMEDUMP WCBE_DUMP_NONE
#define DEFAULT_PIPELINE_FLOW WCBE_FLOW_OFF
#define DEFAULT_PIPELINE_FPS 100
#define DEFAULT_PIPELINE_SOCKDIR NULL
#define DEFAULT_PIPELINE_LSP_CHANNEL NULL
#define DEFAULT_PIPELINE_LSP_SOURCE 0

#define CONTROL_SOCK_NAME "ctlsock"
#define INPUT_STAGE_SOCK_NAME "issink"

#define MAX_SPAWN_ATTEMPTS 3

typedef struct configuration_ configuration_t;

typedef struct pipeline {
    GMainLoop *loop;
    gchar *name;
    gchar *src;
    gboolean live_ts;
    wcbe_dump_format_t framedump;
    wcbe_flow_t flow;
    guint fps;
    gchar *sockdir;
    gint sockfd;
    GstElement *is_pipeline;
    GstElement *old_is_pipeline;
    GHashTable *configs;
    GSList *old_configs;
    GSList *pending_configs;
    configuration_t *config;
    GSList *old_ls_pipelines;
    GIOChannel *lsp_channel;
    guint lsp_source;
} pipeline_t;

typedef enum {CONFIG_STATE_INITIALIZING, CONFIG_STATE_READY,
              CONFIG_STATE_NOT_READY, CONFIG_STATE_FINALIZING} config_state_t;

struct configuration_ {
    gchar *name;
    gchar *desc;
    gchar *dir;
    GSList *ls_pipelines;
    gboolean destroy;
    pipeline_t *pipeline;
    config_state_t state;
};

typedef struct ls_pipeline {
    gchar *name;
    GPid pid;
    gchar **argv;
    guint spawn_count;
    pipeline_state_t state;
    pipeline_t *pipeline;
    configuration_t *config;
    gboolean destroy;
    guint exit_timeout;
} ls_pipeline_t;

typedef struct pending_config {
    enum {PENDING_CREATE, PENDING_ACTIVATE, PENDING_DESTROY} action;
    configuration_t *config;
} pending_config_t;

typedef gssize (*request_fn_t)(pipeline_t *, wcbe_msg_t **);

static void rm_path(const gchar *path, const gchar *dir);
static void parse_options(gint argc, gchar *argv[], pipeline_t *pipeline,
                          gchar **conffile);
static gchar *parse_src(const gchar *value);
static gchar *parse_name(const gchar *value);
static gboolean parse_framedump(const gchar *value,
                                wcbe_dump_format_t *format);
static gboolean parse_flow(const gchar *value, wcbe_flow_t *flow);
static GIOChannel *open_comm_channel(const gchar *sockdir, const gchar *name);
static gint open_sock(const gchar *dir, const gchar *name);
static void create_pipeline(pipeline_t *pipeline);
static void destroy_pipeline(pipeline_t *pipeline);
static void cleanup_old_pipeline(pipeline_t *pipeline);
static gboolean bus_cb(GstBus *bus, GstMessage *msg, gpointer data);
static void handle_pipeline_state(pipeline_t *pipeline,
                                  GstStructure *structure);
static gboolean handle_comm(GIOChannel *channel, GIOCondition condition,
                            gpointer data);
static gboolean handle_child_comm(GIOChannel *source, GIOCondition condition,
                                  gpointer data);
static void start_pipeline(pipeline_t *pipeline, gboolean in_main_loop);
static void stop_pipeline(pipeline_t *pipeline, gboolean in_main_loop);
static void change_pipeline_state(pipeline_t *pipeline, gboolean in_main_loop,
                                  GstState state);
ls_pipeline_t *find_ls_pipeline(GSList *list, const gchar *name);
static void destroy_configuration(gpointer data);
static gssize config_create_command(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize config_activate_command(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize config_destroy_command(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize config_list_query(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize src_command(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize src_query(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize live_ts_command(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize live_ts_query(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize framedump_command(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize framedump_query(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize flow_command(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize flow_query(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize rate_command(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize rate_query(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize name_query(pipeline_t *pipeline, wcbe_msg_t **msg);
static gssize quit_command(pipeline_t *pipeline, wcbe_msg_t **msg);
static gboolean spawn_ls_pipeline(const gchar *cfg_name,
                                  xmlNodePtr xml_pipeline,
                                  ls_pipeline_t *lsp);
static void create_ls_pipelines(pipeline_t *pipeline, configuration_t *config);
static void pipeline_exit_notify(GPid pid, gint status, gpointer data);
static gboolean pipeline_kill(gpointer data);
static void cleanup_config(configuration_t *config);
static configuration_t *find_config(pipeline_t *pipeline, const gchar *name);
static void insert_pending_conf(pipeline_t *pipeline, guint action,
                                configuration_t *config);
static gboolean create_config(pipeline_t *pipeline, configuration_t *config);
static gboolean delayed_activate_config(gpointer data);
static gboolean activate_config(pipeline_t *pipeline, configuration_t *config);
static void destroy_config(pipeline_t *pipeline, configuration_t *config);
static GSList *find_pending_config(GSList *pending_configs, const gchar *name);
static void free_ls_pipeline(ls_pipeline_t *lsp);
static void connect_all_signals(GstElement *pipeline, GstElement *element);
static void config_activated_handler(GstElement *element, const gchar *name,
                                     gboolean result, gpointer data);
static void config_created_handler(GstElement *element, const gchar *name,
                                   gboolean result, gpointer data);
static void config_destroyed_handler(GstElement *element, const gchar *name,
                                     gboolean result, gpointer data);
static void config_deactivated_handler(GstElement *element, const gchar *name,
                                       gpointer data);

static request_fn_t request_handlers[WCBE_NUM_MSG_TYPES][WCBE_NUM_MSG_INTENTS] = {
    [WCBE_CONFIG_CREATE_MSG]   = {[WCBE_COMMAND] = config_create_command,
                                  [WCBE_QUERY]   = NULL},
    [WCBE_CONFIG_ACTIVATE_MSG] = {[WCBE_COMMAND] = config_activate_command,
                                  [WCBE_QUERY]   = NULL},
    [WCBE_CONFIG_DESTROY_MSG]  = {[WCBE_COMMAND] = config_destroy_command,
                                  [WCBE_QUERY]   = NULL},
    [WCBE_CONFIG_LIST_MSG]     = {[WCBE_COMMAND] = NULL,
                                  [WCBE_QUERY]   = config_list_query},
    [WCBE_SRC_MSG]             = {[WCBE_COMMAND] = src_command,
                                  [WCBE_QUERY]   = src_query},
    [WCBE_LIVE_TS_MSG]         = {[WCBE_COMMAND] = live_ts_command,
                                  [WCBE_QUERY]   = live_ts_query},
    [WCBE_FRAMEDUMP_MSG]       = {[WCBE_COMMAND] = framedump_command,
                                  [WCBE_QUERY]   = framedump_query},
    [WCBE_FLOW_MSG]            = {[WCBE_COMMAND] = flow_command,
                                  [WCBE_QUERY]   = flow_query},
    [WCBE_RATE_MSG]            = {[WCBE_COMMAND] = rate_command,
                                  [WCBE_QUERY]   = rate_query},
    [WCBE_NAME_MSG]            = {[WCBE_COMMAND] = NULL,
                                  [WCBE_QUERY]   = name_query},
    [WCBE_QUIT_MSG]            = {[WCBE_COMMAND] = quit_command,
                                  [WCBE_QUERY]   = NULL}
};

gint
main(gint argc, char *argv[])
{
    pipeline_t pipeline =
        {.name = DEFAULT_PIPELINE_NAME,
         .src = DEFAULT_PIPELINE_SRC,
         .config = DEFAULT_PIPELINE_CONFIG,
         .is_pipeline = DEFAULT_PIPELINE_PIPELINE,
         .old_is_pipeline = DEFAULT_PIPELINE_PIPELINE,
         .live_ts = DEFAULT_PIPELINE_LIVE_TS,
         .framedump = DEFAULT_PIPELINE_FRAMEDUMP,
         .flow = DEFAULT_PIPELINE_FLOW,
         .fps = DEFAULT_PIPELINE_FPS,
         .sockdir = DEFAULT_PIPELINE_SOCKDIR,
         .old_ls_pipelines = NULL,
         .old_configs = DEFAULT_PIPELINE_OLD_CONFIGS,
         .pending_configs = DEFAULT_PIPELINE_PENDING_CONFIGS,
         .lsp_channel = DEFAULT_PIPELINE_LSP_CHANNEL,
         .lsp_source = DEFAULT_PIPELINE_LSP_SOURCE
        };
    gchar *conffile = NULL;

    if (!g_thread_supported()) {
        g_thread_init(NULL);
    }
    enums_register();
    pipeline.loop = g_main_loop_new(NULL, FALSE);
    pipeline.configs = g_hash_table_new_full(
        g_str_hash, g_str_equal, NULL, destroy_configuration);

    // Parse command line
    parse_options(argc, argv, &pipeline, &conffile);
    if (pipeline.sockdir != NULL) {
        pipeline.sockdir = g_strdup(pipeline.sockdir);
    }
    else {
        pipeline.sockdir = g_strdup(g_get_tmp_dir());
    }

    // Create socket directory
    if (pipeline.name == NULL) {
        guint i = 0;
        while (pipeline.name == NULL && i < 100) {
            gchar *trial = g_strdup_printf("cbe%u", i);
            gchar *dir = g_build_filename(pipeline.sockdir, trial, NULL);
            gint rc = mkdir(dir, S_IRWXU);
            if (rc == 0) {
                pipeline.name = trial;
                g_free(pipeline.sockdir);
                pipeline.sockdir = dir;
            }
            else {
                g_free(trial);
                g_free(dir);
            }
            ++i;
        }
    }
    else {
        GError *err = NULL;
        gchar *name =
            g_filename_from_utf8(pipeline.name, -1, NULL, NULL, &err);
        if (name != NULL) {
            gchar *dir = g_build_filename(pipeline.sockdir, name, NULL);
            gint rc = mkdir(dir, S_IRWXU);
            if (rc == 0) {
                g_free(pipeline.sockdir);
                pipeline.sockdir = dir;
            }
            else {
                g_free(pipeline.name);
                pipeline.name = NULL;
                g_free(dir);
            }
            g_free(name);
        }
        else {
            g_error("Failed to convert 'name' string to valid "
                    "directory name: %s", err->message);
        }
    }
    if (pipeline.name == NULL) {
        g_printerr("Failed to create socket directory for instance");
        exit(-1);
    }

    // set up TermSource to catch termination signals
    gboolean finished = FALSE;
    gboolean term_handler(gint *signum) {
        finished = TRUE;
        return FALSE;
    }
    TermSource *term_source = term_source_new();
    term_source_set_callback(term_source, term_handler);
    term_source_attach(term_source, NULL);

    gboolean done(gpointer data) {
        if (finished && g_main_loop_is_running(pipeline.loop)) {
            g_main_loop_quit(pipeline.loop);
        }
        return TRUE;
    };
    g_timeout_add(100, done, NULL);

    // Set up comm channel
    GIOChannel *comm_channel =
        open_comm_channel(pipeline.sockdir, CONTROL_SOCK_NAME);
    if (comm_channel == NULL) {
        g_error("Failed to open communication channel");
    }
    g_io_add_watch(comm_channel, G_IO_IN | G_IO_ERR, handle_comm, &pipeline);
    pipeline.sockfd = g_io_channel_unix_get_fd(comm_channel);

    // Create pipeline if possible
    if (pipeline.src != NULL) {
        create_pipeline(&pipeline);
        if (pipeline.is_pipeline != NULL && pipeline.flow == WCBE_FLOW_ON) {
            start_pipeline(&pipeline, FALSE);
        }
    }

    // Set up initial configuration
    if (conffile != NULL) {
        GError *err = NULL;
        gchar *contents;
        gboolean rc = g_file_get_contents(conffile, &contents, NULL, &err);
        if (rc) {
            gsize desc_len = strlen(contents);
            wcbe_config_create_msg_t *msg =
                g_malloc(sizeof(wcbe_config_create_msg_t) + desc_len);
            wcbe_msg_t *cbemsg = &(msg->hdr);
            g_strlcpy(msg->name, "init", WCBE_MAX_NAME_LEN);
            msg->desc_len = desc_len;
            g_strlcpy(msg->desc, contents, desc_len);
            msg->hdr.type = WCBE_CONFIG_CREATE_MSG;
            config_create_command(&pipeline, &cbemsg);
            msg->hdr.type = WCBE_CONFIG_ACTIVATE_MSG;
            config_activate_command(&pipeline, &cbemsg);
            g_free(contents);
        }
        else {
            g_warning("Failed to read contents of initial config file %s: %s",
                      conffile, err->message);
        }
    }

    // Go
    if (!finished) {
        g_main_loop_run(pipeline.loop);
    }

    // Clean up pipeline
    destroy_pipeline(&pipeline);
    void pkill(gpointer data, gpointer user_data) {
        ls_pipeline_t *lsp = data;
        lsp->state = PIPELINE_STATE_STOPPED;
        pipeline_kill(data);
    }
    g_slist_foreach(pipeline.old_ls_pipelines, pkill, NULL);
    cleanup_old_pipeline(&pipeline);
    g_hash_table_destroy(pipeline.configs);
    g_main_loop_unref(pipeline.loop);
    rm_path(pipeline.sockdir, "");
    g_free(pipeline.sockdir);

    return 0;
}

static void
rm_path(const gchar *path, const gchar *dir)
{
    gchar *fullpath = g_build_filename(dir, path, NULL);
    if (g_file_test(fullpath, G_FILE_TEST_IS_DIR)) {
        GError *err = NULL;
        GDir *pdir = g_dir_open(fullpath, 0, &err);
        if (pdir != NULL) {
            const gchar *name = g_dir_read_name(pdir);
            while (name != NULL) {
                rm_path(name, fullpath);
                name = g_dir_read_name(pdir);
            }
            g_dir_close(pdir);
        }
        else {
            g_error("Failed to open directory '%s': %s",
                    fullpath, err->message);
        }
    }
    gint rc = g_remove(fullpath);
    if (rc != 0) {
        g_error("Failed to remove '%s': %s", fullpath,
                g_strerror(errno));
    }
    g_free(fullpath);
}

static void
parse_options(gint argc, gchar *argv[], pipeline_t *pipeline, gchar **conffile)
{
    gboolean src_cb(const gchar *option_name, const gchar *value,
                    gpointer data, GError **error) {
        pipeline->src = parse_src(value);
        return pipeline->src != NULL;
    }
    gboolean name_cb(const gchar *option_name, const gchar *value,
                     gpointer data, GError **error) {
        pipeline->name = parse_name(value);
        return pipeline->name != NULL;
    }
    gboolean framedump_cb(const gchar *option_name, const gchar *value,
                          gpointer data, GError **error) {
        gboolean result = parse_framedump(value, &(pipeline->framedump));
        if (!result) {
            g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_FAILED,
                        "Failed to parse 'framedump' value; "
                        "must be unique prefix of 'binary' or 'ascii'.");
        }
        return result;
    }
    gboolean flow_cb(const char *option_name, const gchar *value,
                     gpointer data, GError **error) {
        gboolean result = parse_flow(value, &(pipeline->flow));
        if (!result) {
            g_set_error(error, G_OPTION_ERROR, G_OPTION_ERROR_FAILED,
                        "Failed to parse 'flow' value; "
                        "must be unique prefix of 'on' or 'off'.");
        }
        return result;
    }

    const GOptionEntry entries[] = {
        {"name", 'n', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_CALLBACK,
         name_cb, "WIDAR backend instance name", "NAME"},
        {"src", 's', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_CALLBACK,
         src_cb, "WIDAR frame source URI", "URI"},
        {"live-timestamps", 'l', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_NONE,
         &(pipeline->live_ts),
         "Rewrite frames from file source to use live timestamps", NULL},
        {"rate", 'r', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_INT,
         &(pipeline->fps), "File frame source frame rate (frames per second)",
         "RATE"},
        {"sockdir", 'k', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_FILENAME,
         &(pipeline->sockdir), "Socket directory", "DIR"},
        {"framedump", 'd', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_CALLBACK,
         framedump_cb, "Dump raw frames", "<none|binary|ascii>"},
        {"flow", 'f', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_CALLBACK,
         flow_cb, "Data flow state", "<on|off>"},
        {"conf", 'c', G_OPTION_FLAG_OPTIONAL_ARG, G_OPTION_ARG_FILENAME,
         conffile, "Initial configuration file", "FILE"},
        {NULL}
    };

    GError *error = NULL;
    GOptionContext *context =
        g_option_context_new("- WIDAR correlator backend input stage");
    g_option_context_set_summary(
        context,
        "Input stage for compute nodes of WIDAR correlator "
        "backend system.");
    g_option_context_set_description(
        context,
        "Please report feature requests, bugs, etc. to\n"
        "mpokorny@nrao.edu.");
    g_option_context_set_help_enabled(context, TRUE);
    g_option_context_set_ignore_unknown_options(context, FALSE);
    g_option_context_add_main_entries(context, entries, NULL);
    g_option_context_add_group(context, gst_init_get_option_group());
    gboolean ctxt_parse =
        g_option_context_parse(context, &argc, &argv, &error);
    if (ctxt_parse) {
        if (pipeline->flow == WCBE_FLOW_ON && pipeline->src == NULL) {
            g_set_error(&error, G_OPTION_ERROR, G_OPTION_ERROR_FAILED,
                        "Invalid command line options: "
                        "'flow=on' requires a 'src' option value.");
            ctxt_parse = FALSE;
        }
    }
    if (!ctxt_parse) {
        if (error != NULL) {
            g_printerr("%s\n", error->message);
            g_error_free(error);
        }
        exit(-1);
    }
    g_option_context_free(context);
}

static gchar *
parse_src(const gchar *value)
{
    // TODO: validate 'value'
    return g_strdup(value);
}

static gchar *
parse_name(const gchar *value)
{
    gchar *result = NULL;
    // We don't check for name conflicts here, since we're not ready
    // to open the named socket.
    if (strchr(value, G_DIR_SEPARATOR) == NULL) {
        result = g_strdup(value);
    }
    return result;
}

static gboolean
parse_framedump(const gchar *value, wcbe_dump_format_t *format)
{
    gboolean result;
    if (g_str_has_prefix("binary", value)) {
        *format = WCBE_DUMP_BINARY;
        result = TRUE;
    }
    else if (g_str_has_prefix("ascii", value)) {
        *format = WCBE_DUMP_ASCII;
        result = TRUE;
    }
    else if (g_str_has_prefix("none", value)) {
        *format = WCBE_DUMP_NONE;
        result = TRUE;
    }
    else {
        result = FALSE;
    }
    return result;
}

static gboolean
parse_flow(const gchar *value, wcbe_flow_t *flow)
{
    gboolean result = FALSE;
    if (value[0] == 'o') {
        if (g_str_has_prefix("on", value)) {
            *flow = WCBE_FLOW_ON;
            result = TRUE;
        }
        else if (g_str_has_prefix("off", value)) {
            *flow = WCBE_FLOW_OFF;
            result = TRUE;
        }
    }
    return result;
}

static GIOChannel *
open_comm_channel(const gchar *sockdir, const gchar *name)
{
    GIOChannel *result = NULL;
    int sockfd = open_sock(sockdir, name);
    if (sockfd != -1) {
        result = g_io_channel_unix_new(sockfd);
        g_io_channel_set_close_on_unref(result, TRUE);
        GError *err = NULL;
        g_io_channel_set_encoding(result, NULL, &err);
        if (err != NULL) {
            g_error("Failed to set NULL encoding on communication channel: %s",
                    err->message);
        }
        g_io_channel_set_buffered(result, FALSE);
    }
    return result;
}

static gint
open_sock(const gchar *dir, const gchar *name)
{
    gint result = -1;
    gchar *path = g_build_filename(dir, name, NULL);
    gint sockfd = socket(PF_UNIX, SOCK_DGRAM, 0);
    if (sockfd != -1) {
        struct sockaddr_un addr = {.sun_family = AF_UNIX};
        gint rc = g_strlcpy(addr.sun_path, path, sizeof(addr.sun_path));
        if (rc < sizeof(addr.sun_path)) {
            g_assert(rc > 0);
            rc = bind(sockfd, (struct sockaddr *)&addr, sizeof(addr));
            if (rc != -1) {
                result = sockfd;
            }
            else {
                close(sockfd);
                g_warning("Failed to bind name of Unix domain "
                          "socket to '%s': %s", addr.sun_path,
                          g_strerror(errno));
            }
        }
        else {
            close(sockfd);
            g_warning("Socket file name (%s) too long", path);
        }
    }
    else {
        g_error("Failed to create Unix domain socket: %s",
                g_strerror(errno));
    }
    g_free(path);

    return result;
}

static void
create_pipeline(pipeline_t *pipeline)
{
    g_return_if_fail(pipeline->src != NULL);

    cleanup_old_pipeline(pipeline);

    GstBus *bus = NULL;
    GstElement *src = NULL;
    GstElement *framedumper = NULL;
    GstElement *decoder = NULL;
    GstElement *sorter = NULL;
    GstElement *sink = NULL;
    gboolean pipeline_ready = FALSE;

    // pipeline
    pipeline->is_pipeline = gst_pipeline_new("input_pipeline");
    if (pipeline->is_pipeline == NULL) {
        g_printerr("Failed to create pipeline");
        goto cleanup_and_return;
    }

    // bus
    bus = gst_pipeline_get_bus(GST_PIPELINE(pipeline->is_pipeline));
    if (bus == NULL) {
        g_printerr("Failed to get pipeline message bus");
        goto cleanup_and_return;
    }
    gst_bus_add_watch(bus, bus_cb, pipeline);

    // source element
    src = gst_element_factory_make("wcbe_src", "src");
    if (src == NULL) {
        g_printerr("Failed to create 'wcbe_src' element\n");
        goto cleanup_and_return;
    }
    if (!gst_uri_handler_set_uri(GST_URI_HANDLER(src), pipeline->src)) {
        g_printerr("Failed to set pipeline source\n");
        goto cleanup_and_return;
    }
    g_object_set(G_OBJECT(src), "live-timestamps", pipeline->live_ts, NULL);
    g_object_set(G_OBJECT(src), "fps", pipeline->fps, NULL);

    // framedumper element
    framedumper = gst_element_factory_make("wcbe_framedumper", "framedumper");
    if (framedumper == NULL) {
        g_printerr("Failed to create 'wcbe_framedumper' element\n");
        goto cleanup_and_return;
    }
    if (pipeline->framedump != WCBE_DUMP_NONE) {
        WcbeFramedumpFormat fdformat =
            ((pipeline->framedump == WCBE_DUMP_BINARY)
             ? WCBE_FRAMEDUMP_BINARY : WCBE_FRAMEDUMP_ASCII);
        g_object_set(G_OBJECT(framedumper), "format", fdformat, NULL);
        g_object_set(G_OBJECT(framedumper), "do-output", TRUE, NULL);
    }
    else {
        g_object_set(G_OBJECT(framedumper), "do-output", FALSE, NULL);
    }

    // decoder element
    decoder = gst_element_factory_make("wcbe_decoder", "decoder");
    if (decoder == NULL) {
        g_printerr("Failed to create 'wcbe_decode' element\n");
        goto cleanup_and_return;
    }

    // sorter element
    sorter = gst_element_factory_make("wcbe_sorter", "sorter");
    if (sorter == NULL) {
        g_printerr("Failed to create 'wcbe_sorter' element\n");
        goto cleanup_and_return;
    }
    connect_all_signals(pipeline->is_pipeline, sorter);

    // sink element
    sink = gst_element_factory_make("wcbe_lagset_sink", "sink");
    if (sink == NULL) {
        g_printerr("Failed to create 'wcbe_lagset_sink' element\n");
        goto cleanup_and_return;
    }
    g_debug("setting sink sockdir to %s", pipeline->sockdir);
    g_object_set(G_OBJECT(sink), "sockdir", pipeline->sockdir, NULL);
    connect_all_signals(pipeline->is_pipeline, sink);

    // assemble pipeline
    gst_bin_add_many(GST_BIN(pipeline->is_pipeline), src, framedumper, decoder,
                     sorter, sink, NULL);
    gboolean link_result =
        gst_element_link_many(src, framedumper, decoder, sorter, sink, NULL);
    src = framedumper = decoder = sorter = sink = NULL;
    if (!link_result) {
        g_printerr("Failed to link pipeline elements\n");
        goto cleanup_and_return;
    }

    // create comm channel for child lagset pipelines
    pipeline->lsp_channel =
        open_comm_channel(pipeline->sockdir, INPUT_STAGE_SOCK_NAME);
    if (pipeline->lsp_channel == NULL) {
        g_error("Failed to open communication channel for child "
                "lagset pipelines");
    }
    pipeline->lsp_source =
        g_io_add_watch(pipeline->lsp_channel, G_IO_IN | G_IO_ERR,
                       handle_child_comm, pipeline);
    g_assert(pipeline->lsp_source != DEFAULT_PIPELINE_LSP_SOURCE);

    pipeline_ready = TRUE;

 cleanup_and_return:
    if (bus != NULL) {
        gst_object_unref(GST_OBJECT(bus));
    }
    if (sink != NULL) {
        gst_object_unref(GST_OBJECT(sink));
    }
    if (sorter != NULL) {
        gst_object_unref(GST_OBJECT(sorter));
    }
    if (decoder != NULL) {
        gst_object_unref(GST_OBJECT(decoder));
    }
    if (framedumper != NULL) {
        gst_object_unref(GST_OBJECT(framedumper));
    }
    if (src != NULL) {
        gst_object_unref(GST_OBJECT(src));
    }
    if (!pipeline_ready && pipeline->is_pipeline != NULL) {
        gst_object_unref(GST_OBJECT(pipeline->is_pipeline));
        pipeline->is_pipeline = NULL;
    }
}

static void
destroy_pipeline(pipeline_t *pipeline)
{
    pipeline->config = DEFAULT_PIPELINE_CONFIG;
    g_hash_table_remove_all(pipeline->configs);
    if (pipeline->is_pipeline != NULL) {
        if (pipeline->flow == WCBE_FLOW_ON) {
            stop_pipeline(pipeline, g_main_loop_is_running(pipeline->loop));
        }
    }
    pipeline->flow = DEFAULT_PIPELINE_FLOW;
    if (pipeline->src != DEFAULT_PIPELINE_SRC) {
        g_free(pipeline->src);
        pipeline->src = DEFAULT_PIPELINE_SRC;
    }
    pipeline->old_is_pipeline = pipeline->is_pipeline;
    pipeline->is_pipeline = DEFAULT_PIPELINE_PIPELINE;
}

static void
cleanup_old_pipeline(pipeline_t *pipeline)
{
    if (pipeline->old_is_pipeline != DEFAULT_PIPELINE_PIPELINE) {
        gst_object_unref(GST_OBJECT(pipeline->old_is_pipeline));
        pipeline->old_is_pipeline = DEFAULT_PIPELINE_PIPELINE;
    }
    if (pipeline->lsp_source != DEFAULT_PIPELINE_LSP_SOURCE) {
        g_source_remove(pipeline->lsp_source);
        pipeline->lsp_source = DEFAULT_PIPELINE_LSP_SOURCE;
    }
    if (pipeline->lsp_channel != DEFAULT_PIPELINE_LSP_CHANNEL) {
        GError *err = NULL;
        g_io_channel_shutdown(pipeline->lsp_channel, TRUE, &err);
        if (err != NULL) {
            g_debug("Failed to shut down lagset pipelines IO channel: %s",
                    err->message);
        }
        g_io_channel_unref(pipeline->lsp_channel);
        pipeline->lsp_channel = DEFAULT_PIPELINE_LSP_CHANNEL;
        gchar *sockname =
            g_build_filename(pipeline->sockdir, INPUT_STAGE_SOCK_NAME, NULL);
        gint rc = g_remove(sockname);
        if (rc != 0) {
            g_warning("Failed to unlink socket directory entry '%s': %s",
                      sockname, g_strerror(errno));
        }
        g_free(sockname);
    }
}

static gboolean
bus_cb(GstBus *bus, GstMessage *msg, gpointer data)
{
    pipeline_t *pipeline = data;

    switch (GST_MESSAGE_TYPE(msg)) {
      case GST_MESSAGE_EOS:
        destroy_pipeline(pipeline);
        break;

      case GST_MESSAGE_ERROR: {
        gchar *debug = NULL;
        GError *err = NULL;
        gst_message_parse_error(msg, &err, &debug);
        g_printerr("ERROR: %s\n", err->message);
        if (debug != NULL) {
            g_printerr("Debug details: %s\n", debug);
            g_free(debug);
        }
        g_error_free(err);
        destroy_pipeline(pipeline);
        break;
      }
      case GST_MESSAGE_APPLICATION: {
        const gchar *name = gst_structure_get_name(msg->structure);
        if (strcmp(name, "pipeline-state") == 0) {
            handle_pipeline_state(pipeline, msg->structure);
        }
        break;
      }
      default:
        break;
    }
    return TRUE;
}

ls_pipeline_t *
find_ls_pipeline(GSList *list, const gchar *name)
{
    ls_pipeline_t *result = NULL;
    while (result == NULL && list != NULL) {
        ls_pipeline_t *lsp = list->data;
        if (strcmp(lsp->name, name) == 0) {
            result = lsp;
        }
        list = g_slist_next(list);
    }
    return result;
}

static void
handle_pipeline_state(pipeline_t *pipeline, GstStructure *structure)
{
    const gchar *src = gst_structure_get_string(structure, "src");
    if (src != NULL) {
        gchar **src_parts = g_strsplit(src, G_DIR_SEPARATOR_S, 0);
        gchar *cfg_name = src_parts[0];
        if (cfg_name != NULL) {
            ls_pipeline_t *lsp = NULL;
            configuration_t *config =
                g_hash_table_lookup(pipeline->configs, cfg_name);
            if (config != NULL) {
                lsp = find_ls_pipeline(config->ls_pipelines, src);
            }
            if (lsp == NULL) {
                lsp = find_ls_pipeline(pipeline->old_ls_pipelines, src);
            }
            if (lsp != NULL) {
                gint state_int;
                if (gst_structure_get_enum(structure,
                                           "value",
                                           PIPELINE_STATE_ENUM_TYPE,
                                           &state_int)) {
                    g_debug("got 'pipeline-state' message '%s'",
                            g_enum_get_value(
                                g_type_class_peek_static(
                                    PIPELINE_STATE_ENUM_TYPE),
                                state_int)->value_nick);
                    lsp->state = state_int;
                    switch (lsp->state) {
                      case PIPELINE_STATE_STOPPED:
                        if (lsp->exit_timeout != 0){
                            g_source_remove(lsp->exit_timeout);
                        }
                        lsp->exit_timeout =
                            g_timeout_add(5000, pipeline_kill, lsp);
                        if (lsp->config != NULL
                            && lsp->config->state == CONFIG_STATE_READY) {
                            lsp->config->state = CONFIG_STATE_NOT_READY;
                        }
                        break;
                      case PIPELINE_STATE_RUNNING:
                        if (lsp->config != NULL) {
                            gboolean ready = TRUE;
                            GSList *item = lsp->config->ls_pipelines;
                            g_assert(item != NULL);
                            while (item != NULL && ready) {
                                ready = (((ls_pipeline_t *)(item->data))->state
                                         == PIPELINE_STATE_RUNNING);
                                item = g_slist_next(item);
                            }
                            if (ready) {
                                lsp->config->state = CONFIG_STATE_READY;
                            }
                        }
                        break;
                      default:
                        g_assert_not_reached();
                        break;
                    }
                }
                else {
                    g_debug("Received 'pipeline-state' message without "
                            "'value' field");
                }
            }
            else {
                g_debug("Received 'pipeline-state' message from unexpected "
                        "src '%s'", src);
            }
        }
        else {
            g_debug("Received 'pipeline-state' message with invalid "
                    "'src' value");
        }
        g_strfreev(src_parts);
    }
    else {
        g_debug("Received 'pipeline-state' message without 'src' field");
    }
}

/* handle_comm
 *
 * Read a datagram from the input comm socket.
 *
 * We are using a GIOChannel simply as an easy means to integrate the
 * socket polling into the main loop. We definitely do not read the
 * datagram via the GIOChannel because it simply reads bytes, not
 * datagrams. Reading bytes could be a problem if something unexpected
 * arrived on the socket---perhaps unlikely, but it's safer this way.
 */
static gboolean
handle_comm(GIOChannel *channel, GIOCondition condition, gpointer data)
{
    pipeline_t *pipeline = data;
    switch (condition) {
    case G_IO_IN: {
      gssize num_bytes;
      wcbe_msg_t in_msg;
      struct sockaddr_un addr;
      struct iovec iovec[] =
          {{.iov_base = &in_msg, .iov_len = sizeof(in_msg)}};
      struct msghdr msghdr = {
          .msg_name = addr.sun_path, .msg_namelen = sizeof(addr.sun_path),
          .msg_control = NULL, .msg_controllen = 0,
          .msg_iov = iovec, .msg_iovlen = 1};
      gint sockfd = g_io_channel_unix_get_fd(channel);
      num_bytes = recvmsg(sockfd, &msghdr, MSG_PEEK);
      if (num_bytes == sizeof(in_msg)
          && in_msg.direction == WCBE_REQUEST
          && in_msg.type < WCBE_NUM_MSG_TYPES
          && in_msg.intent < WCBE_NUM_MSG_INTENTS) {
          wcbe_msg_t *msg = NULL;
          void read_fixed_size_msg(gssize size) {
              msg = g_malloc0(size);
              iovec[0].iov_base = msg;
              iovec[0].iov_len = size;
              num_bytes = recvmsg(sockfd, &msghdr, 0);
              if (num_bytes != size) {
                  g_free(msg);
                  msg = NULL;
              }
          }
          switch (in_msg.type) {
            case WCBE_CONFIG_CREATE_MSG: {
              msg = g_malloc0(sizeof(wcbe_config_create_msg_t));
              iovec[0].iov_base = msg;
              iovec[0].iov_len = sizeof(wcbe_config_create_msg_t);
              num_bytes = recvmsg(sockfd, &msghdr, MSG_PEEK);
              if (num_bytes == sizeof(wcbe_config_create_msg_t)) {
                  wcbe_config_create_msg_t *ccmsg =
                      (wcbe_config_create_msg_t *)msg;
                  gsize msgsize =
                      sizeof(wcbe_config_create_msg_t) + ccmsg->desc_len;
                  msg = g_realloc(msg, msgsize);
                  iovec[0].iov_base = msg;
                  iovec[0].iov_len = msgsize;
                  num_bytes = recvmsg(sockfd, &msghdr, 0);
                  if (num_bytes != msgsize) {
                      g_free(msg);
                      msg = NULL;
                  }
              }
              break;
            }
            case WCBE_CONFIG_ACTIVATE_MSG:
              read_fixed_size_msg(sizeof(wcbe_config_activate_msg_t));
              break;
            case WCBE_CONFIG_DESTROY_MSG:
              read_fixed_size_msg(sizeof(wcbe_config_destroy_msg_t));
              break;
            case WCBE_CONFIG_LIST_MSG:
              read_fixed_size_msg(sizeof(wcbe_config_list_msg_t));
              break;
            case WCBE_SRC_MSG:
              read_fixed_size_msg(sizeof(wcbe_src_msg_t));
              break;
            case WCBE_LIVE_TS_MSG:
              read_fixed_size_msg(sizeof(wcbe_live_ts_msg_t));
              break;
            case WCBE_FRAMEDUMP_MSG:
              read_fixed_size_msg(sizeof(wcbe_framedump_msg_t));
              break;
            case WCBE_FLOW_MSG:
              read_fixed_size_msg(sizeof(wcbe_flow_msg_t));
              break;
            case WCBE_RATE_MSG:
              read_fixed_size_msg(sizeof(wcbe_rate_msg_t));
              break;
            case WCBE_NAME_MSG:
              read_fixed_size_msg(sizeof(wcbe_name_msg_t));
              break;
            case WCBE_QUIT_MSG:
              read_fixed_size_msg(sizeof(wcbe_quit_msg_t));
              break;
            default:
              break;
          }
          if (msg != NULL) {
              request_fn_t fn = request_handlers[msg->type][msg->intent];
              if (fn != NULL) {
                  gssize msgsize = fn(pipeline, &msg);
                  if (msgsize != -1) {
                      iovec[0].iov_base = msg;
                      iovec[0].iov_len = msgsize;
                  }
              }
              msg->direction = WCBE_RESPONSE;
              num_bytes = sendmsg(sockfd, &msghdr, 0);
              if (num_bytes < iovec[0].iov_len) {
                  g_warning("Input comm channel error on write: %s",
                            g_strerror(errno));
              }
          }
      }
      break;
    }
    case G_IO_ERR:
      break;
    default:
      g_assert_not_reached();
      break;
    }
    return TRUE;
}

static gboolean
handle_child_comm(GIOChannel *source, GIOCondition condition, gpointer data)
{
    pipeline_t *pipeline = data;
    guint str_len;
    struct iovec iovec[2] =
        {{.iov_base = &str_len, .iov_len = sizeof(str_len)}};
    struct msghdr msghdr =
        {.msg_name = NULL, .msg_namelen = 0,
         .msg_control = NULL, .msg_controllen = 0,
         .msg_iov = iovec, .msg_iovlen = 1};
    gint fd = g_io_channel_unix_get_fd(source);
    gssize nbytes = recvmsg(fd, &msghdr, MSG_PEEK | MSG_TRUNC);
    if (G_LIKELY(nbytes > 0)) {
        iovec[1].iov_base = g_malloc(str_len);
        iovec[1].iov_len = str_len;
        msghdr.msg_iovlen = 2;
        nbytes = recvmsg(fd, &msghdr, 0);
        if (G_LIKELY(nbytes == iovec[0].iov_len + iovec[1].iov_len)) {
            ((gchar *)(iovec[1].iov_base))[iovec[1].iov_len - 1]
                = '\0';
            GstStructure *structure =
                gst_structure_from_string(iovec[1].iov_base, NULL);
            if (structure != NULL) {
                GstElement *is_pipeline =
                    (pipeline->is_pipeline ? pipeline->is_pipeline
                     : pipeline->old_is_pipeline);
                g_assert(is_pipeline != NULL);
                GstMessage *msg = gst_message_new_application(
                    GST_OBJECT(is_pipeline), structure);
                gboolean res = gst_element_post_message(
                    GST_ELEMENT(is_pipeline), msg);
                if (!res) {
                    GST_DEBUG("Failed to post message to bus");
                }
            }
            else {
                GST_DEBUG("Failed to parse application message '%s'",
                          iovec[1].iov_base);
            }
        }
        else {
            GST_DEBUG("Failed to read expected number of bytes in "
                      "application message");
        }
        g_free(iovec[1].iov_base);
    }
    else {
        GST_DEBUG("Failed to read expected number of bytes in "
                  "application message header");
    }

    return TRUE;
}

static void
start_pipeline(pipeline_t *pipeline, gboolean in_main_loop)
{
    g_assert(pipeline->is_pipeline != NULL);
    change_pipeline_state(pipeline, in_main_loop, GST_STATE_PAUSED);
    pipeline->flow = WCBE_FLOW_ON;
    while (pipeline->pending_configs != NULL) {
        GSList *pending = pipeline->pending_configs;
        pipeline->pending_configs =
            g_slist_remove_link(pipeline->pending_configs, pending);
        pending_config_t *pconf = pending->data;
        switch (pconf->action) {
          case PENDING_CREATE:
            g_debug("create pending '%s'", pconf->config->name);
            if (!create_config(pipeline, pconf->config)) {
                GSList *same_conf =
                    find_pending_config(pipeline->pending_configs,
                                        pconf->config->name);
                while (same_conf != NULL) {
                    pipeline->pending_configs =
                        g_slist_delete_link(pipeline->pending_configs,
                                            same_conf);
                    same_conf = find_pending_config(pipeline->pending_configs,
                                                    pconf->config->name);
                }
                cleanup_config(pconf->config);
            }
            break;
          case PENDING_ACTIVATE:
            g_debug("activate pending '%s'", pconf->config->name);
            if (!activate_config(pipeline, pconf->config)) {
                g_timeout_add(DELAYED_ACTIVATE_INTERVAL,
                              delayed_activate_config, pconf->config);
            }
            break;
          case PENDING_DESTROY:
            g_debug("destroy pending '%s'", pconf->config->name);
            destroy_config(pipeline, pconf->config);
            break;
          default:
            g_assert_not_reached();
            break;
        }
        g_free(pconf);
        g_slist_free_1(pending);
    }
    if (pipeline->config != NULL) {
        change_pipeline_state(pipeline, in_main_loop, GST_STATE_PLAYING);
    }
}

static void
stop_pipeline(pipeline_t *pipeline, gboolean in_main_loop)
{
    change_pipeline_state(pipeline, in_main_loop, GST_STATE_NULL);
    pipeline->flow = WCBE_FLOW_OFF;
}

static void
change_pipeline_state(pipeline_t *pipeline, gboolean in_main_loop,
                      GstState state)
{
    GstStateChangeReturn ret =
        gst_element_set_state(pipeline->is_pipeline, state);
    if (ret == GST_STATE_CHANGE_FAILURE) {
        const gchar *st = (state == GST_STATE_PLAYING ? "start" : "stop");
        g_warning("Failed to %s pipeline", st);
        if (!in_main_loop) {
            GstBus *bus =
                gst_pipeline_get_bus(GST_PIPELINE(pipeline->is_pipeline));
            GstMessage *msg = gst_bus_poll(bus, GST_MESSAGE_ERROR, 0);
            if (msg != NULL) {
                GError *error = NULL;
                gst_message_parse_error(msg, &error, NULL);
                g_printerr("ERROR: %s\n", error->message);
                g_error_free(error);
                gst_message_unref(msg);
            }
            gst_object_unref(GST_OBJECT(bus));
        }
    }
}

static void
destroy_configuration(gpointer data)
{
    configuration_t *config = data;
    g_assert(config != config->pipeline->config);
    gboolean in_main_loop = g_main_loop_is_running(config->pipeline->loop);

    config->destroy = TRUE;
    config->state = CONFIG_STATE_FINALIZING;
    if (config->ls_pipelines != NULL) {
        GSList *living_pipelines = NULL;
        GSList *list = config->ls_pipelines;
        while (list != NULL) {
            ls_pipeline_t *lsp = list->data;
            if (lsp->state != PIPELINE_STATE_DEFUNCT) {
                living_pipelines = g_slist_prepend(living_pipelines, lsp);
                list = g_slist_next(list);
            }
            else {
                free_ls_pipeline(lsp);
                list = g_slist_delete_link(list, list);
            }
        }
        config->ls_pipelines = living_pipelines;

        void mark_destroyed(gpointer data, gpointer user_data) {
            ls_pipeline_t *lsp = data;
            lsp->destroy = TRUE;
            lsp->config = NULL;
        }
        g_slist_foreach(config->ls_pipelines, mark_destroyed, NULL);

        config->pipeline->old_ls_pipelines = g_slist_concat(
            config->pipeline->old_ls_pipelines, config->ls_pipelines);
    }

    // destroy config in input pipeline
    if (in_main_loop && config->pipeline->flow == WCBE_FLOW_ON) {
        GstEvent *destroy_config_event = gst_event_new_custom(
            WCBE_EVENT_DESTROY_CONFIG,
            gst_structure_new("config-destroy",
                              "name", G_TYPE_STRING, config->name,
                              NULL));
        if (!gst_element_send_event(GST_ELEMENT(config->pipeline->is_pipeline),
                                    destroy_config_event)) {
            g_info("Failed to send DESTROY_CONFIG event for config '%s' "
                   "to input stage pipeline", config->name);
        }
    }

    // send EOS event and also schedule a process kill signal to every
    // pipeline in ls_pipelines
    void send_eos(gpointer data, gpointer user_data) {
        ls_pipeline_t *lsp = data;
        gchar *sockname =
            g_build_filename(config->pipeline->sockdir, lsp->name, NULL);
        wcbe_stream_event_t stream_event;
        stream_event.object.type = WCBE_STREAM_EVENT;
        stream_event.event = GST_EVENT_EOS;
        struct sockaddr_un addr = {.sun_family = AF_UNIX};
        gint rc =
            g_strlcpy(addr.sun_path, sockname, sizeof(addr.sun_path));
        g_assert(0 < rc && rc < sizeof(addr.sun_path));
        struct iovec iovec[] =
            {{.iov_base = &stream_event, .iov_len = sizeof(stream_event)}};
        struct msghdr msghdr = {.msg_name = &addr,
                                .msg_namelen = sizeof(addr),
                                .msg_control = NULL, .msg_controllen = 0,
                                .msg_iov = iovec, .msg_iovlen = 1};
        gssize bytes_written =
            sendmsg(config->pipeline->sockfd, &msghdr, 0);
        if (bytes_written != iovec[0].iov_len) {
            const gchar *errstr;
            if (errno != 0) {
                errstr = g_strerror(errno);
            }
            else {
                errstr = "'errno' not set";
            }
            g_debug("Failed to write stream object to socket: %s", errstr);
        }
        g_free(sockname);
        lsp->exit_timeout =
            g_timeout_add(30000, pipeline_kill, lsp);
    }
    g_slist_foreach(config->ls_pipelines, send_eos, NULL);
    config->pipeline->old_configs =
        g_slist_prepend(config->pipeline->old_configs, config);
    cleanup_config(config);
}

static gssize
config_create_command(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_CREATE_MSG, -1);

    wcbe_config_create_msg_t *create_msg = (wcbe_config_create_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, create_msg->name);
    if (config == NULL) {
        gboolean free_config = FALSE;
        config = g_new(configuration_t, 1);
        config->name = g_strdup(create_msg->name);
        config->desc = g_strdup(create_msg->desc);
        config->ls_pipelines = NULL;
        config->dir = g_build_filename(pipeline->sockdir, config->name, NULL);
        config->destroy = FALSE;
        config->pipeline = pipeline;
        gint rc = mkdir(config->dir, S_IRWXU);
        if (rc == 0) {
            if (pipeline->is_pipeline != NULL
                && pipeline->flow == WCBE_FLOW_ON) {
                free_config = !create_config(pipeline, config);
            }
            else {
                insert_pending_conf(pipeline, PENDING_CREATE, config);
            }
        }
        else {
            g_info("Failed to create configuration directory '%s': %s",
                   config->dir, g_strerror(errno));
            free_config = TRUE;
        }
        if (free_config) {
            cleanup_config(config);
        }
    }
    return -1;
}

static gssize
config_activate_command(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_ACTIVATE_MSG, -1);

    wcbe_config_activate_msg_t *activate_msg =
        (wcbe_config_activate_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, activate_msg->name);
    if (config != NULL) {
        if (pipeline->is_pipeline != NULL && pipeline->flow == WCBE_FLOW_ON) {
            if (!config->destroy
                && !activate_config(pipeline, config)) {
                g_timeout_add(DELAYED_ACTIVATE_INTERVAL,
                              delayed_activate_config, config);
            }
        }
        else {
            insert_pending_conf(pipeline, PENDING_ACTIVATE, config);
        }
    }
    return -1;
}

static gssize
config_destroy_command(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_DESTROY_MSG, -1);

    wcbe_config_destroy_msg_t *destroy_msg = (wcbe_config_destroy_msg_t *)*msg;
    configuration_t *config = find_config(pipeline, destroy_msg->name);
    if (config != NULL) {
        if (pipeline->is_pipeline != NULL && pipeline->flow == WCBE_FLOW_ON) {
            destroy_config(pipeline, config);
        }
        else {
            insert_pending_conf(pipeline, PENDING_DESTROY, config);
        }
    }
    return -1;
}

static gssize
config_list_query(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_CONFIG_LIST_MSG, -1);

    wcbe_config_list_msg_t **config_list_msg = (wcbe_config_list_msg_t **)msg;
    GPtrArray *configs = g_ptr_array_new();
    (*config_list_msg)->list_len = 1;
    // configs in active pipeline
    void config_name(gpointer key, gpointer value, gpointer user_data) {
        const gchar *name = key;
        const configuration_t *cfg = value;
        const gchar *fmt;
        if (cfg->destroy) {
            fmt = "[%s]";
        }
        else if (cfg == pipeline->config) {
            fmt = "<%s>";
        }
        else {
            fmt = "(%s)";
        }
        gchar *tagged_name = g_strdup_printf(fmt, name);
        g_ptr_array_add(configs, tagged_name);
        (*config_list_msg)->list_len += strlen(tagged_name) + 1;
    }
    g_hash_table_foreach(pipeline->configs, config_name, NULL);
    // pending configs
    void pending_config_name(gpointer data, gpointer user_data) {
        pending_config_t *pconf = data;
        const gchar *fmt;
        switch (pconf->action) {
          case PENDING_CREATE:
            fmt = "*%s*";
            break;
          case PENDING_ACTIVATE:
            fmt = "#%s#";
            break;
          case PENDING_DESTROY:
            fmt = "@%s@";
            break;
          default:
            g_assert_not_reached();
            break;
        }
        gchar *tagged_name = g_strdup_printf(fmt, pconf->config->name);
        g_ptr_array_add(configs, tagged_name);
        (*config_list_msg)->list_len += strlen(tagged_name) + 1;
    }
    g_slist_foreach(pipeline->pending_configs, pending_config_name, NULL);

    gssize result =
        sizeof(wcbe_config_list_msg_t) + (*config_list_msg)->list_len;
    *config_list_msg = g_realloc(*config_list_msg, result);
    guint list_end = 0;
    guint i;
    for (i = 0; i < configs->len; ++i) {
        gchar *config = g_ptr_array_index(configs, i);
        guint len = strlen(config) + 1;
        memcpy(&((*config_list_msg)->list[list_end]), config, len);
        list_end += len;
        g_free(config);
    }
    (*config_list_msg)->list[list_end] = '\0';
    g_ptr_array_free(configs, TRUE);
    return result;
}

static gssize
src_command(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_SRC_MSG, -1);

    wcbe_src_msg_t *src_msg = (wcbe_src_msg_t *)*msg;
    src_msg->name[sizeof(src_msg->name) - 1] = '\0';

    if (pipeline->src == NULL
        || strcmp(src_msg->name, pipeline->src) != 0) {
        if (pipeline->src != NULL) {
            g_free(pipeline->src);
            pipeline->src = NULL;
        }
        wcbe_flow_t flow = pipeline->flow;
        if (pipeline->is_pipeline != DEFAULT_PIPELINE_PIPELINE) {
            destroy_pipeline(pipeline);
        }
        pipeline->src = g_strdup(src_msg->name);
        if (pipeline->src[0] != '\0') {
            create_pipeline(pipeline);
            if (pipeline->is_pipeline != DEFAULT_PIPELINE_PIPELINE
                && flow == WCBE_FLOW_ON) {
                start_pipeline(pipeline, TRUE);
            }
        }
    }
    return -1;
}

static gssize
src_query(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_SRC_MSG, -1);

    wcbe_src_msg_t *src_msg = (wcbe_src_msg_t *)*msg;
    const gchar *name = (pipeline->src ? pipeline->src : "");
    g_strlcpy(src_msg->name, name, sizeof(src_msg->name));
    return -1;
}

static gssize
live_ts_command(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_LIVE_TS_MSG, -1);

    wcbe_live_ts_msg_t *live_ts_msg = (wcbe_live_ts_msg_t *)*msg;
    if (live_ts_msg->state != pipeline->live_ts) {
        if (pipeline->is_pipeline != NULL) {
            GstElement *src =
                gst_bin_get_by_name(GST_BIN(pipeline->is_pipeline), "src");
            g_assert(src != NULL);
            g_object_set(G_OBJECT(src), "live-timestamps", live_ts_msg->state,
                         NULL);
            gst_object_unref(GST_OBJECT(src));
        }
        pipeline->live_ts = live_ts_msg->state;
    }
    return -1;
}

static gssize
live_ts_query(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_LIVE_TS_MSG, -1);

    wcbe_live_ts_msg_t *live_ts_msg = (wcbe_live_ts_msg_t *)*msg;
    live_ts_msg->state = pipeline->live_ts;
    return -1;
}

static gssize
framedump_command(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_FRAMEDUMP_MSG, -1);

    wcbe_framedump_msg_t *framedump_msg = (wcbe_framedump_msg_t *)*msg;
    if (framedump_msg->format != pipeline->framedump) {
        if (pipeline->is_pipeline != NULL) {
            GstElement *framedumper = gst_bin_get_by_name(
                GST_BIN(pipeline->is_pipeline), "framedumper");
            g_assert(framedumper != NULL);
            if (framedump_msg->format != WCBE_DUMP_NONE) {
                WcbeFramedumpFormat fdformat =
                    ((framedump_msg->format == WCBE_DUMP_BINARY)
                     ? WCBE_FRAMEDUMP_BINARY
                     : WCBE_FRAMEDUMP_ASCII);
                g_object_set(G_OBJECT(framedumper), "do-output", FALSE, NULL);
                g_object_set(G_OBJECT(framedumper), "format", fdformat, NULL);
                g_object_set(G_OBJECT(framedumper), "do-output", TRUE, NULL);
            }
            else {
                g_object_set(G_OBJECT(framedumper), "do-output", FALSE, NULL);
            }
            gst_object_unref(GST_OBJECT(framedumper));
        }
        pipeline->framedump = framedump_msg->format;
    }
    return -1;
}

static gssize
framedump_query(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_FRAMEDUMP_MSG, -1);

    wcbe_framedump_msg_t *framedump_msg = (wcbe_framedump_msg_t *)*msg;
    framedump_msg->format = pipeline->framedump;
    return -1;
}

static gssize
flow_command(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_FLOW_MSG, -1);

    wcbe_flow_msg_t *flow_msg = (wcbe_flow_msg_t *)*msg;
    if (flow_msg->state != pipeline->flow) {
        if (flow_msg->state == WCBE_FLOW_ON) {
            if (pipeline->is_pipeline != NULL) {
                start_pipeline(pipeline, TRUE);
            }
        }
        else { // flow_msg->state == WCBE_FLOW_OFF
            if (pipeline->is_pipeline != NULL) {
                stop_pipeline(pipeline, TRUE);
            }
        }
        // start/stop_pipeline update pipeline->flow automatically
    }
    return -1;
}

static gssize
flow_query(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_FLOW_MSG, -1);

    wcbe_flow_msg_t *flow_msg = (wcbe_flow_msg_t *)*msg;
    flow_msg->state = pipeline->flow;
    return -1;
}

static gssize
rate_command(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_RATE_MSG, -1);

    wcbe_rate_msg_t *rate_msg = (wcbe_rate_msg_t *)*msg;
    if (rate_msg->fps != pipeline->fps) {
        if (pipeline->is_pipeline != NULL) {
            GstElement *src =
                gst_bin_get_by_name(GST_BIN(pipeline->is_pipeline), "src");
            g_assert(src != NULL);
            g_object_set(G_OBJECT(src), "fps", rate_msg->fps, NULL);
            gst_object_unref(GST_OBJECT(src));
        }
        pipeline->fps = rate_msg->fps;
    }
    return -1;
}

static gssize
rate_query(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_RATE_MSG, -1);

    wcbe_rate_msg_t *rate_msg = (wcbe_rate_msg_t *)*msg;
    rate_msg->fps = pipeline->fps;
    return -1;
}

static gssize
name_query(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_NAME_MSG, -1);

    wcbe_name_msg_t *name_msg = (wcbe_name_msg_t *)*msg;
    g_assert(pipeline->name != NULL);
    g_strlcpy(name_msg->value, pipeline->name, sizeof(name_msg->value));
    return -1;
}

static gssize
quit_command(pipeline_t *pipeline, wcbe_msg_t **msg)
{
    g_return_val_if_fail((*msg)->type == WCBE_QUIT_MSG, -1);
    g_main_loop_quit(pipeline->loop);
    return -1;
}

static void
create_ls_pipelines(pipeline_t *pipeline, configuration_t *config)
{
    xmlDocPtr doc = xmlParseMemory(config->desc, strlen(config->desc));
    if (doc != NULL) {
        xmlNodePtr root = xmlDocGetRootElement(doc);
        if (root != NULL
            && xmlStrEqual(root->name, BAD_CAST "config")) {
            xmlXPathInit();
            xmlXPathContextPtr context = xmlXPathNewContext(doc);
            g_assert(context != NULL);
            context->node = root;
            xmlXPathObjectPtr xml_pipelines = xmlXPathEvalExpression(
                BAD_CAST "//productPipeline", context);
            if (xml_pipelines != NULL) {
                guint i;
                for (i = 0; i < xml_pipelines->nodesetval->nodeNr; ++i) {
                    xmlNodePtr xml_pipeline =
                        xml_pipelines->nodesetval->nodeTab[i];
                    ls_pipeline_t *lsp = g_new0(ls_pipeline_t, 1);
                    lsp->state = PIPELINE_STATE_UNKNOWN;
                    lsp->pipeline = pipeline;
                    lsp->destroy = FALSE;
                    lsp->config = config;
                    if (spawn_ls_pipeline(config->name, xml_pipeline, lsp)) {
                        config->ls_pipelines =
                            g_slist_prepend(config->ls_pipelines, lsp);
                    }
                    else {
                        g_free(lsp);
                    }
                }
                xmlXPathFreeObject(xml_pipelines);
            }
            else {
                g_debug("Failed to find any products in "
                        "configuration '%s'", config->name);
            }
            xmlXPathFreeContext(context);
        }
        else {
            g_debug("Configuration root element is missing or is "
                    "not a 'config' element");
        }
        xmlFreeDoc(doc);
    }
    else {
        g_debug("Failed to parse configuration");
    }
}

static gboolean
spawn_ls_pipeline(const gchar *cfg_name, xmlNodePtr xml_pipeline,
                  ls_pipeline_t *lsp)
{
    gboolean result = FALSE;
    gboolean free_lsp = TRUE;
    xmlChar *pipeline_id = xmlGetProp(xml_pipeline, BAD_CAST "id");
    if (pipeline_id != NULL) {
        xmlChar *pipeline_desc = xmlGetProp(xml_pipeline, BAD_CAST "desc");
        if (pipeline_desc != NULL) {
            lsp->name =
                g_strjoin(G_DIR_SEPARATOR_S, cfg_name, pipeline_id, NULL);
            gchar *cmd =
                g_strdup_printf("lagset_pipeline --input %s/%s %s",
                                lsp->pipeline->sockdir,
                                lsp->name,
                                pipeline_desc);
            lsp->argv = NULL;
            gint argc;
            GError *err = NULL;
            gboolean rc = g_shell_parse_argv(cmd, &argc, &(lsp->argv), &err);
            if (rc) {
                rc = g_spawn_async(NULL, lsp->argv, NULL,
                                   G_SPAWN_DO_NOT_REAP_CHILD
                                   | G_SPAWN_SEARCH_PATH,
                                   NULL, NULL, &(lsp->pid), &err);
                if (rc) {
                    result = TRUE;
                    g_child_watch_add(lsp->pid, pipeline_exit_notify, lsp);
                    lsp->spawn_count++;
                    free_lsp = FALSE;
                }
                else {
                    g_debug("Failed to spawn lagset pipeline process: %s",
                            err->message);
                }
            }
            else {
                g_info("Failed to parse lagset pipeline command line '%s': %s",
                       cmd, err->message);
            }
            if (free_lsp && lsp->argv != NULL) {
                g_strfreev(lsp->argv);
            }
            g_free(cmd);
            xmlFree(pipeline_desc);
        }
        else {
            g_debug("Lagset pipeline config element has no 'desc' attribute");
        }
        xmlFree(pipeline_id);
    }
    else {
        g_debug("Lagset pipeline config element has no 'id' attribute");
    }
    if (free_lsp) {
        if (lsp->name != NULL) {
            g_free(lsp->name);
        }
        g_free(lsp);
    }
    return result;
}

static void
pipeline_exit_notify(GPid pid, gint status, gpointer data)
{
    ls_pipeline_t *lsp = data;
    gboolean respawned = FALSE;

    g_spawn_close_pid(pid);

    if (lsp->exit_timeout != 0){
        g_source_remove(lsp->exit_timeout);
        lsp->exit_timeout = 0;
    }

    if (!lsp->destroy) {
        if (WIFEXITED(status)) {
            g_info("Pipeline process exited unexpectedly with status %u",
                   WEXITSTATUS(status));
        }
        else if (WIFSIGNALED(status)) {
            g_info("Pipeline process exited with signal %u",
                   WTERMSIG(status));
        }
        if (lsp->spawn_count < MAX_SPAWN_ATTEMPTS) {
            GError *err = NULL;
            gint rc = g_spawn_async(NULL, lsp->argv, NULL,
                                    G_SPAWN_DO_NOT_REAP_CHILD
                                    | G_SPAWN_SEARCH_PATH,
                                    NULL, NULL, &(lsp->pid), &err);
            if (rc) {
                g_child_watch_add(lsp->pid, pipeline_exit_notify, lsp);
                lsp->spawn_count++;
                lsp->state = PIPELINE_STATE_UNKNOWN;
                respawned = TRUE;
            }
            else {
                g_debug("Failed to spawn lagset pipeline process: %s",
                        err->message);
            }
        }
        else {
            g_info("Pipeline program '%s' failed %u times...will not respawn",
                   lsp->argv[0], lsp->spawn_count);
        }
    }

    if (!respawned) {
        configuration_t *config = lsp->config;
        lsp->state = PIPELINE_STATE_DEFUNCT;
        GSList *listitem = NULL;
        if (config == NULL) {
            GSList **list = &(lsp->pipeline->old_ls_pipelines);
            listitem = g_slist_find(*list, lsp);
            g_assert(listitem != NULL);
            *list = g_slist_delete_link(*list, listitem);
            free_ls_pipeline(lsp);
        }
        else {
            if (config->state != CONFIG_STATE_FINALIZING) {
                config->state = CONFIG_STATE_NOT_READY;
                config->destroy = TRUE;
            }
            else {
                free_ls_pipeline(lsp);
            }
        }
    }
}

static gboolean
pipeline_kill(gpointer data)
{
    ls_pipeline_t *lsp = data;
    lsp->destroy = TRUE;
    gint rc = kill(lsp->pid, SIGKILL);
    if (rc != 0 && errno != ESRCH) {
        // ignore errno == ESRCH since the process could have exited
        // between the timeout and the 'kill' invocation
        g_warning("Failed to kill lagset pipeline process %d: %s",
                  lsp->pid, g_strerror(errno));
    }
    return FALSE;
}

static void
cleanup_config(configuration_t *config)
{
    // remove config directory
    rm_path(config->dir, "");

    g_free(config->name);
    g_free(config->desc);
    g_free(config->dir);
    g_free(config);
}

static configuration_t *
find_config(pipeline_t *pipeline, const gchar *name)
{
    configuration_t *result;
    result = g_hash_table_lookup(pipeline->configs, name);
    if (result == NULL) {
        if (pipeline->pending_configs != NULL) {
            g_assert(pipeline->is_pipeline == NULL
                     || pipeline->flow == WCBE_FLOW_OFF);
            GSList *pconf =
                find_pending_config(pipeline->pending_configs, name);
            if (pconf != NULL) {
                result = ((pending_config_t *)(pconf->data))->config;
            }
        }
    }
    return result;
}

static void
insert_pending_conf(pipeline_t *pipeline, guint action,
                    configuration_t *config)
{
    g_assert(pipeline->is_pipeline == NULL || pipeline->flow == WCBE_FLOW_OFF);
    pending_config_t *pconf = g_new(pending_config_t, 1);
    pconf->action = action;
    pconf->config = config;
    pipeline->pending_configs =
        g_slist_append(pipeline->pending_configs, pconf);
}

static gboolean
create_config(pipeline_t *pipeline, configuration_t *config)
{
    g_assert(pipeline->is_pipeline != NULL && pipeline->flow == WCBE_FLOW_ON);

    gboolean result;
    GstEvent *create_config_event =
        gst_event_new_custom(
            WCBE_EVENT_CREATE_CONFIG,
            gst_structure_new("config-create",
                              "config", G_TYPE_STRING, config->desc,
                              "name", G_TYPE_STRING, config->name,
                              NULL));
    g_assert(create_config_event != NULL);
    if (gst_element_send_event(GST_ELEMENT(pipeline->is_pipeline),
                               create_config_event)) {
        config->state = CONFIG_STATE_INITIALIZING;
        create_ls_pipelines(pipeline, config);
        g_hash_table_insert(pipeline->configs, config->name, config);
        result = TRUE;
    }
    else {
        g_info("Failed to send CREATE_CONFIG event for "
               "config '%s' to input stage pipeline",
               config->name);
        result = FALSE;
    }
    return result;
}

static gboolean
delayed_activate_config(gpointer data)
{
    gboolean result = FALSE;
    configuration_t *config = data;
    if (!config->destroy) {
        result = !activate_config(config->pipeline, config);
    }
    return result;
}

static gboolean
activate_config(pipeline_t *pipeline, configuration_t *config)
{
    g_return_val_if_fail(!config->destroy, FALSE);

    g_assert(pipeline->is_pipeline != NULL && pipeline->flow == WCBE_FLOW_ON);
    g_assert(g_hash_table_lookup(pipeline->configs, config->name) != NULL);

    gboolean result = FALSE;
    if (config->state == CONFIG_STATE_READY && config != pipeline->config) {
        GstEvent *activate_config_event = gst_event_new_custom(
            WCBE_EVENT_ACTIVATE_CONFIG,
            gst_structure_new("config-activate",
                              "name", G_TYPE_STRING, config->name,
                              NULL));
        if (gst_element_send_event(GST_ELEMENT(pipeline->is_pipeline),
                                   activate_config_event)) {
            configuration_t *old_config = pipeline->config;
            pipeline->config = config;
            if (old_config != NULL && old_config->destroy) {
                g_hash_table_remove(pipeline->configs, old_config->name);
            }
            if (old_config == NULL) {
                change_pipeline_state(pipeline,
                                      g_main_loop_is_running(pipeline->loop),
                                      GST_STATE_PLAYING);
            }
            result = TRUE;
        }
        else {
            g_info("Failed to send ACTIVATE_CONFIG event for "
                   "config '%s' to input stage pipeline",
                   config->name);
        }
    }

    return result;
}

static void
destroy_config(pipeline_t *pipeline, configuration_t *config)
{
    g_assert(pipeline->is_pipeline != NULL && pipeline->flow == WCBE_FLOW_ON);
    g_assert(g_hash_table_lookup(pipeline->configs, config->name) != NULL);

    config->destroy = TRUE;
    if ((config->state == CONFIG_STATE_READY
         || config->state == CONFIG_STATE_NOT_READY)
        && config != pipeline->config) {
        g_hash_table_remove(pipeline->configs, config->name);
    }
}

static GSList *
find_pending_config(GSList *pending_configs, const gchar *name)
{
    gint cmp_names(gconstpointer a, gconstpointer b) {
        const pending_config_t *pconf1 = a;
        const pending_config_t *pconf2 = b;
        return strcmp(pconf1->config->name, pconf2->config->name);
    }
    configuration_t dummy_conf = {.name = (gchar *)name};
    pending_config_t dummy_pconf = {.config = &dummy_conf};
    return g_slist_find_custom(pending_configs, &dummy_pconf, cmp_names);
}

static void
free_ls_pipeline(ls_pipeline_t *lsp)
{
    g_strfreev(lsp->argv);
    g_free(lsp->name);
    g_free(lsp);
}

static void
connect_all_signals(GstElement *pipeline, GstElement *element)
{
    g_signal_connect(G_OBJECT(element), "config-activated",
                     G_CALLBACK(config_activated_handler), NULL);
    g_signal_connect(G_OBJECT(element), "config-created",
                     G_CALLBACK(config_created_handler), NULL);
    g_signal_connect(G_OBJECT(element), "config-destroyed",
                     G_CALLBACK(config_destroyed_handler), NULL);
    g_signal_connect(G_OBJECT(element), "config-deactivated",
                     G_CALLBACK(config_deactivated_handler), NULL);
}

static void
config_activated_handler(GstElement *element, const gchar *name,
                         gboolean result, gpointer data)
{
    g_debug("Got 'config-activated' %s %s from %s",
            name, result ? "TRUE" : "FALSE", GST_ELEMENT_NAME(element));
}

static void
config_created_handler(GstElement *element, const gchar *name,
                       gboolean result, gpointer data)
{
    g_debug("Got 'config-created' %s %s from %s",
            name, result ? "TRUE" : "FALSE", GST_ELEMENT_NAME(element));
}

static void
config_destroyed_handler(GstElement *element, const gchar *name,
                         gboolean result, gpointer data)
{
    g_debug("Got 'config-destroyed' %s %s from %s",
            name, result ? "TRUE" : "FALSE", GST_ELEMENT_NAME(element));
}

static void
config_deactivated_handler(GstElement *element, const gchar *name,
                           gpointer data)
{
    g_debug("Got 'config-deactivated' %s from %s",
            name, GST_ELEMENT_NAME(element));
}
