#! /usr/bin/env python

from optparse import OptionParser
import os.path
import os
import struct
import sys
import socket
import stat
import tempfile
import atexit
import readline
from copy import deepcopy
from tempfile import mkstemp
from time import time, sleep
from Enum import Enum

def mktempsock(sockdir):
    sock = socket.socket(socket.AF_UNIX, socket.SOCK_DGRAM, 0)

    # To get a reasonable name for our socket, I think that the use of
    # mktemp followed by bind would be sufficient; but mktemp is
    # deprecated, so we use mkstemp instead. Unfortunately, mkstemp
    # also produces an open file, which is useless and must be closed
    # and unlinked immediately after opening. So there is a race
    # condition here, but I don't think that it's harmful (I hope).

    (tempfd, sockname) = mkstemp(dir=sockdir)
    os.close(tempfd)
    os.unlink(sockname)
    sock.bind(sockname) # raises socket.error in case name is in use
    return (sock, sockname)

class DumpFormat (object):
    __metaclass__ = Enum
    __enums = {'none':0, 'binary':1, 'ascii':2}

class Flow (object):
    __metaclass__ = Enum
    __enums = {'off':0, 'on':1}

class LiveTs (object):
    __metaclass__ = Enum
    __enums = {'false':0, 'true':1}

class MsgType (object):
    __metaclass__ = Enum
    __enums = {'config_create':0, 'config_activate':1, 'config_destroy':2,
               'config_list':3, 'src':4, 'live_ts':5, 'framedump':6,
               'flow':7, 'rate':8, 'name':9, 'quit':10}

class MsgDirection (object):
    __metaclass__ = Enum
    __enums = {'request':0, 'response':1}

class MsgIntent (object):
    __metaclass__ = Enum
    __enums = {'command':0, 'query':1}

def str_trunc(str):
    null = str.find('\0')
    if null == -1:
        result = str
    else:
        result = str[:null]
    return result

class CbeMsg (object):

    format_str = ''
    type = ''

    def __init__(self, direction, intent, id):
        self.fields = {}
        self.fields['intent'] = intent
        self.fields['id'] = id
        self.fields['direction'] = direction
        return

    intent = property(lambda (self):self.fields['intent'])
    id = property(lambda (self):self.fields['id'])
    direction = property(lambda (self):self.fields['direction'])

    def msg_classes():
        return (ConfigCreateMsg, ConfigActivateMsg, ConfigDestroyMsg,
                ConfigListMsg, SrcMsg, LiveTsMsg, FramedumpMsg,
                FlowMsg, RateMsg, NameMsg, QuitMsg)
    msg_classes = staticmethod(msg_classes)

    def class_lookup(code):
        klass = [k for k in CbeMsg.msg_classes() if int(k.type) == code]
        if len(klass) != 1:
            raise IndexError()
        return klass[0]
    class_lookup = staticmethod(class_lookup)

    def pack(self):
        (fmt, args) = self.pack_args()
        return struct.pack('iiiI' + fmt, self.type,
                           self.direction, self.intent, self.id, *args)

    def unpack(str):
        size = len(str) - struct.calcsize('iiiI')
        (type, dir, intent, id, msg) = struct.unpack('iiiI%ds' % size, str)
        klass = CbeMsg.class_lookup(type)
        fields = (dir, intent, id) + klass.unpack(msg)
        return klass(*klass.convert_fields(*fields))
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id):
        return (MsgDirection(direction), MsgIntent(intent), id)
    convert_fields = staticmethod(convert_fields)

class StrMsg (CbeMsg):

    format_str = "256s"

    def __init__(self, intent, id=0, name='', direction=MsgDirection.request):
        super(StrMsg, self).__init__(direction, intent, id)
        null = name.find('\0')
        self.fields['name'] = str_trunc(name)
        return

    name = property(lambda (self):self.fields['name'])

    def pack_args(self):
        return (self.format_str, (self.name,))

    def unpack(str):
        return struct.unpack(StrMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, name):
        return (MsgIntent(intent), id, name, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class NameMsg (StrMsg):
    type = MsgType.name

class SrcMsg (StrMsg):
    type = MsgType.src

class ConfigCreateMsg (CbeMsg):

    type = MsgType.config_create
    format_str = '256sI%ds'

    def __init__(self, name, desc, intent=MsgIntent.command, id=0,
                 direction=MsgDirection.request):
        super(ConfigCreateMsg, self).__init__(direction, intent, id)
        self.fields['name'] = str_trunc(name)
        self.fields['desc'] = desc
        return

    name = property(lambda (self):self.fields['name'])
    desc = property(lambda (self):self.fields['desc'])

    def pack_args(self):
        desc_len = len(self.desc) + 1
        return (self.format_str % desc_len, (self.name, desc_len, self.desc))

    def unpack(str):
        size = len(str) - struct.calcsize(ConfigCreateMsg.format_str[:-3])
        return struct.unpack(ConfigCreateMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, name, desc_len, desc):
        return (name, desc, MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class ConfigActivateMsg (StrMsg):
    type = MsgType.config_activate

class ConfigDestroyMsg (StrMsg):
    type = MsgType.config_destroy

class ConfigListMsg (CbeMsg):

    format_str = 'I%ds'
    type = MsgType.config_list

    def __init__(self, configs=(), intent=MsgIntent.query, id=0,
                 direction=MsgDirection.request):
        super(ConfigListMsg, self).__init__(direction, intent, id)
        self.fields['configs'] = deepcopy(configs)
        return

    configs = property(lambda (self):deepcopy(self.fields['configs']))

    def pack_args(self):
        fmt = ''
        for c in self.fields['configs']:
            fmt += '%ds' % (len(c) + 1)
        fmt += 'c'
        fields = self.fields['configs'] + ('\0',)
        result = struct.pack(fmt, *fields)
        return (self.format_str % len(result), (len(result), result))

    def unpack(str):
        size = len(str) - struct.calcsize(ConfigListMsg.format_str[:-3])
        return struct.unpack(ConfigListMsg.format_str % size, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, configs_len, configs_str):
        configs = []
        null = configs_str.find('\0')
        while null != 0:
            configs.append(configs_str[:null])
            configs_str = configs_str[null + 1:]
            null = configs_str.find('\0')
        return (tuple(configs), MsgIntent(intent), id, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class LiveTsMsg (CbeMsg):

    format_str = "i"
    type = MsgType.live_ts

    def __init__(self, intent, id=0, state=LiveTs.false,
                 direction=MsgDirection.request):
        super(LiveTsMsg, self).__init__(direction, intent, id)
        self.fields['state'] = state
        return

    state = property(lambda (self):self.fields['state'])

    def pack_args(self):
        return (self.format_str, (self.state,))

    def unpack(str):
        return struct.unpack(LiveTsMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, state):
        return (MsgIntent(intent), id, LiveTs(state), MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class FramedumpMsg (CbeMsg):

    format_str = "i"
    type = MsgType.framedump

    def __init__(self, intent, id=0, format=DumpFormat.none,
                 direction=MsgDirection.request):
        super(FramedumpMsg, self).__init__(direction, intent, id)
        self.fields['format'] = format
        return

    format = property(lambda (self):self.fields['format'])

    def pack_args(self):
        return (self.format_str, (self.format,))

    def unpack(str):
        return struct.unpack(FramedumpMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, format):
        return (MsgIntent(intent), id, DumpFormat(format),
                MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class FlowMsg (CbeMsg):

    format_str = "i"
    type = MsgType.flow

    def __init__(self, intent, id=0, state=Flow.off,
                 direction=MsgDirection.request):
        super(FlowMsg, self).__init__(direction, intent, id)
        self.fields['state'] = state
        return

    state = property(lambda (self):self.fields['state'])

    def pack_args(self):
        return (self.format_str, (self.state,))

    def unpack(str):
        return struct.unpack(FlowMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, state):
        return (MsgIntent(intent), id, Flow(state), MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class RateMsg (CbeMsg):

    format_str = "I"
    type = MsgType.rate

    def __init__(self, intent, id=0, rate=0,
                 direction=MsgDirection.request):
        super(RateMsg, self).__init__(direction, intent, id)
        self.fields['rate'] = rate
        return

    rate = property(lambda (self):self.fields['rate'])

    def pack_args(self):
        return (self.format_str, (self.rate,))

    def unpack(str):
        return struct.unpack(RateMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, rate):
        return (MsgIntent(intent), id, rate, MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class QuitMsg (CbeMsg):

    format_str = "I"
    type = MsgType.quit

    def __init__(self, id=0, intent=MsgIntent.command,
                 direction=MsgDirection.request):
        super(QuitMsg, self).__init__(direction, intent, id)
        return

    def pack_args(self):
        return (self.format_str, (0,))

    def unpack(str):
        return struct.unpack(QuitMsg.format_str, str)
    unpack = staticmethod(unpack)

    def convert_fields(direction, intent, id, reserved):
        return (id, MsgIntent(intent), MsgDirection(direction))
    convert_fields = staticmethod(convert_fields)

class BEConn:

    msgid = 0

    def __init__(self, sockdir):
        self.name = None
        self.besockname = None
        self.sockdir = sockdir
        return

    def has_instance(self):
        return (self.name is not None)

    def clear_instance(self):
        self.name = None
        self.besockname = None
        return

    def set_instance(self, name):
        result = False
        self.name = name
        self.besockname = os.path.join(self.sockdir, name, 'ctlsock')
        try:
            mode = os.stat(self.besockname)[stat.ST_MODE]
        except OSError:
            os.spawnlp(os.P_NOWAIT, 'gstcbe', 'gstcbe',
                       '--name', name, '--sockdir', self.sockdir)
            t0 = time()
            while time() - t0 < 5 and not result:
                sleep(0.5)
                try:
                    mode = os.stat(self.besockname)[stat.ST_MODE]
                    result = True
                except OSError:
                    pass
            if not result:
                raise OSError
        else:
            if stat.S_ISSOCK(mode):
                result = True
            else:
                print "'%s' exists, but is not a socket" % self.besockname
                self.name = None
                self.besockname = None
        return result

def selword(words, text, state):
    try:
        result = [w for w in words if w.startswith(text)][state]
    except:
        result = None
    return result

def completer(text, state):

    result = None
    begidx = readline.get_begidx()
    if begidx == 0:
        result = selword(('get', 'set', 'quit', 'kill', 'create', 'destroy'),
                         text, state)
    elif begidx == 4:
        attributes = ('config', 'flow', 'framedump', 'live_ts',
                      'name', 'src', 'rate')
        if readline.get_line_buffer().startswith('get'):
            attributes += ("configs", 'all',)
        result = selword(attributes, text, state)
    else:
        line = readline.get_line_buffer()
        if line.startswith('set flow'):
            result = selword(tuple(Flow.strs()), text, state)
        elif line.startswith('set framedump'):
            result = selword(tuple(DumpFormat.strs()), text, state)
        elif line.startswith('set live_ts'):
            result = selword(tuple(LiveTs.strs()), text, state)
    return result

def parse_input(text, beconn):
    words = text.split(' ', 2)
    if len(words) > 0:
        msg = None
        kill_be = False
        if words[0] == 'get':
            if len(words) > 1:
                if words[1] == 'config' or words[1] == 'configs':
                    msg = ConfigListMsg(intent=MsgIntent.query,
                                        id=beconn.msgid)
                elif words[1] == 'src':
                    msg = SrcMsg(intent=MsgIntent.query, id=beconn.msgid)
                elif words[1] == 'live_ts':
                    msg = LiveTsMsg(intent=MsgIntent.query, id=beconn.msgid)
                elif words[1] == 'flow':
                    msg = FlowMsg(intent=MsgIntent.query, id=beconn.msgid)
                elif words[1] == 'framedump':
                    msg = FramedumpMsg(intent=MsgIntent.query, id=beconn.msgid)
                elif words[1] == 'name':
                    msg = NameMsg(intent=MsgIntent.query, id=beconn.msgid)
                elif words[1] == 'rate':
                    msg = RateMsg(intent=MsgIntent.query, id=beconn.msgid)
                elif words[1] == 'all':
                    for attr in ('name', 'src', 'configs', 'flow', 'framedump',
                                 'rate', 'live_ts'):
                        parse_input('get %s' % attr, beconn)
            else:
                print "Argument of 'get' command is missing"
        elif words[0] == 'set':
            if len(words) == 3:
                if words[1] == 'config':
                    msg = ConfigActivateMsg(intent=MsgIntent.command,
                                            id=beconn.msgid, name=words[2])
                elif words[1] == 'src':
                    msg = SrcMsg(intent=MsgIntent.command, id=beconn.msgid,
                                 name=words[2])
                elif words[1] == 'live_ts':
                    try:
                        msg = LiveTsMsg(intent=MsgIntent.command,
                                        id=beconn.msgid,
                                        state=LiveTs(words[2]))
                    except:
                        pass
                elif words[1] == 'flow':
                    try:
                        msg = FlowMsg(intent=MsgIntent.command,
                                      id=beconn.msgid,
                                      state=Flow(words[2]))
                    except:
                        pass
                elif words[1] == 'framedump':
                    try:
                        msg = FramedumpMsg(intent=MsgIntent.command,
                                           id=beconn.msgid,
                                           format=DumpFormat(words[2]))
                    except:
                        pass
                elif words[1] == 'rate':
                    try:
                        msg = RateMsg(intent=MsgIntent.command,
                                      id=beconn.msgid,
                                      rate=int(words[2]))
                    except:
                        pass
                elif words[1] == 'name':
                    if beconn.set_instance(words[2]):
                        msg = NameMsg(intent=MsgIntent.query, id=beconn.msgid)
            else:
                print "Argument of 'set' command is missing"
        elif words[0] == 'kill':
            msg = QuitMsg(beconn.msgid)
            kill_be = True
        elif words[0] == 'create':
            if len(words) == 3:
                conffile = open(words[2])
                msg = ConfigCreateMsg(intent=MsgIntent.command,
                                      id=beconn.msgid,
                                      name=words[1], desc=conffile.read())
                conffile.close()
            else:
                print "Aargument of 'create' command is missing"
        elif words[0] == 'destroy':
            if len(words) > 1:
                msg = ConfigDestroyMsg(intent=MsgIntent.command,
                                       id=beconn.msgid, name=words[1])
            else:
                print "Argument of 'destroy' command is missing"
        elif words[0] != '' and words[0] != 'quit':
            print "Unrecognized command"
        if msg is not None:
            if beconn.has_instance():
                beconn.msgid += 1
                intent = msg.intent
                beconn.sock.sendto(msg.pack(), beconn.besockname)
                (str, addr) = beconn.sock.recvfrom(300)
                msg = CbeMsg.unpack(str)
                if intent == MsgIntent.query:
                    if msg is not None:
                        if words[1] == 'config':
                            config = [c for c in msg.configs if c[0] in '<[']
                            if len(config) == 1:
                                config = config[0][1:-1]
                                print "Active configuration name: '%s'" \
                                      % config
                            elif len(config) == 0:
                                print "No active configuration"
                            else:
                                print "Pipeline indicates multiple active " \
                                      "configurations: %s" % msg.configs
                        elif words[1] == 'configs':
                            if len(msg.configs) > 0:
                                print 'Active configuration:'
                                active = [c for c in msg.configs
                                          if c[0] in '<[']
                                if len(active) == 0:
                                    print '\t(none)'
                                elif len(active) == 1:
                                    print '\t%s' % active[0][1:-1],
                                    if active[0][0] == '<':
                                        print ''
                                    else:
                                        print '(retired)'
                                else:
                                    print '\tERROR: multiple active ' \
                                          'configurations'
                                print 'Inactive configurations:'
                                show_confs(msg.configs, '(')
                                pending = [c for c in msg.configs
                                           if c[0] in '*#@']
                                if len(pending) > 0:
                                    print 'Pending configuration actions:'
                                    for p in pending:
                                        if p[0] == '*':
                                            print '\tcreate %s' % p[1:-1]
                                        elif p[0] == '#':
                                            print '\tactivate %s' % p[1:-1]
                                        else:
                                            print '\tdestroy %s' % p[1:-1]
                            else:
                                print 'No configurations'
                        elif words[1] == 'src':
                            print "Backend source name: '%s'" % msg.name
                        elif words[1] == 'live_ts':
                            print msg.state
                        elif words[1] == 'flow':
                            print msg.state
                        elif words[1] == 'framedump':
                            print msg.format
                        elif words[1] == 'rate':
                            print "File data frame rate: %d" % msg.rate
                        elif words[1] == 'name':
                            print "Backend instance name: '%s'" % msg.name
                elif kill_be:
                    beconn.clear_instance()
            else:
                print "Backend instance name not set..."
                print "Use 'set name <name>' to select or create a "\
                      "backend instance"
    return

def show_confs(configs, tag):
    configs = [cnf[1:-1] for cnf in configs if cnf[:1] == tag]
    if len(configs) > 0:
        for c in configs:
            print '\t%s' % c
    else:
        print '\t(none)'
    return

def main():

    parser = OptionParser(description='''Console for interactive
control of WIDAR correlator backend system on compute nodes.
''')
    parser.add_option('--name', '-n', action='store', type='string',
                      dest='name', help='Backend instance name')
    parser.add_option('--sockdir', '-s', action='store', type='string',
                      dest='sockdir', help='Directory for sockets')
    (options, args) = parser.parse_args()

    if options.sockdir is not None:
        sockdir = options.sockdir
    else:
        if os.environ.has_key('TMPDIR'):
            sockdir = os.environ['TMPDIR']
        elif os.environ.has_key('TMP'):
            sockdir = os.environ['TMP']
        elif os.environ.has_key('TEMP'):
            sockdir = os.environ['TEMP']
        else:
            sockdir = '/tmp'

    def sock_cleanup():
        if sockname is not None:
            os.unlink(sockname)
        if sock is not None:
            sock.close()
    atexit.register(sock_cleanup)
    (sock, sockname) = mktempsock(sockdir)

    beconn = BEConn(sockdir)
    beconn.sock = sock
    if options.name is not None and options.name != '':
        parse_input('set name %s' % options.name, beconn)

    readline.set_history_length(100)
    readline.set_completer(completer)
    readline.parse_and_bind('tab: complete')
    histfile = None
    def histfile_cleanup():
        if histfile is not None:
            readline.write_history_file(histfile)
    atexit.register(histfile_cleanup)
    try:
        histfile = os.path.join(os.environ['HOME'], '.cbehist')
        readline.read_history_file(histfile)
    except IOError:
        print 'Failed to open history file; command history will not be saved.'
        pass

    inp = ''
    while inp != 'quit':
        if beconn.has_instance():
            prompt = '%s> ' % beconn.name
        else:
            prompt = '> '
        try:
            inp = raw_input(prompt)
            parse_input(inp.strip(), beconn)
            try:
                os.waitpid(-1, os.WNOHANG) # TODO: improve this
            except:
                pass
        except EOFError:
            print
            inp = 'quit'

if __name__ == '__main__':
    main()
