/*
 * dump_bin_frame.c
 *
 * Functions for dumping lag frames in binary format.
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <glib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <time.h>
#include <sys/time.h>

#include <dump_bin_frame.h>

// this module is not MT-safe

static void init(void);

static gboolean initialized = FALSE;
static const char *iso_format = "%Y%m%dT%H%M%S";
#define ISO_FORMAT_LEN 15
static char timestamp[ISO_FORMAT_LEN + 1];
static GHashTable *output_files = NULL;
static size_t max_file_size = 0;

static FILE *get_output_file(const lf_data_frame_word_t *df);

/* dump_bin_frame
 *
 * Write a lag frame in binary format to a file.
 */
size_t
dump_bin_frame(const lf_data_frame_word_t *df)
{
    if (!initialized) {
        init();
    }
    FILE *f = get_output_file(df);
    size_t num_words;
    switch (LF_DATA_FRAME_FTYPE(df)) {
    case LF_LTA_DATA_FRAME_FTYPE:
        num_words = LF_DATA_FRAME_NUM_WORDS(LF_LTA_DATA_FRAME_NUM_LAGS(df));
        break;
    case LF_CC_DATA_FRAME_FTYPE:
        num_words = LF_DATA_FRAME_NUM_WORDS(LF_CC_DATA_FRAME_NUM_LAGS(df));
        break;
    default:
        // only output the header fields, note that end sync and
        // checksum words cannot be included
        num_words = LF_DATA_FRAME_NUM_WORDS(0) - 2;
        break;
    }
    size_t rc = fwrite(df, sizeof(lf_data_frame_word_t), num_words, f);
    if (rc != num_words) {
        g_warning("Failed to write complete lag frame to file: %s",
                  strerror(errno));
    }
    long fsize = ftell(f);
    g_assert(fsize != -1);
    if (fsize > max_file_size) {
        max_file_size = fsize;
    }
    return max_file_size;
}

/* dump_bin_frame_close_all
 *
 * Close all files to which lag frames are being written.
 */
void
dump_bin_frame_close_all(void)
{
    if (initialized) {
        gboolean closef(gpointer key, gpointer val, gpointer reserved) {
            FILE *f = val;
            if (f != NULL) {
                fclose(f);
            }
            return TRUE;
        }
        g_hash_table_foreach_remove(output_files, closef, NULL);
        max_file_size = 0;
        initialized = FALSE;
    }
}

/* init
 *
 * Prepare the open file hash table.
 */
static void
init(void)
{
    if (output_files == NULL) {
        output_files = g_hash_table_new(NULL, NULL);
    }
    time_t t = time(NULL);
    if (G_LIKELY(((time_t)-1) != t)) {
        struct tm *tm = localtime(&t);
        if (G_LIKELY(tm != NULL)) {
            size_t rc = strftime(timestamp, sizeof(timestamp), iso_format, tm);
            if (G_UNLIKELY(rc != ISO_FORMAT_LEN)) {
                g_error("Failed to convert timestamp to ISO format");
            }
        }
        else {
            g_error("Failed to convert system time to local: %s",
                    strerror(errno));
        }
    }
    else {
        g_error("Failed to get system time: %s", strerror(errno));
    }
    initialized = TRUE;
}

/* get_output_file
 *
 * Get the file for a given cc number.
 */
static FILE *
get_output_file(const lf_data_frame_word_t *df)
{
    unsigned key;
    switch (LF_DATA_FRAME_FTYPE(df)) {
    case LF_LTA_DATA_FRAME_FTYPE:
        // key = board s/n | cchip id | ftype
        key = (g_ntohl(df[8]) & 0xFFFF0000) | (g_ntohl(df[1]) & 0x00001F87);
        break;
    case LF_CC_DATA_FRAME_FTYPE:
        // key = ccc# | ftype
        key = (g_ntohl(df[1]) & 0x00000007);
        break;
    default:
        key = df[0]; // start sync word does not match any other key
        break;
    }

    FILE *result = g_hash_table_lookup(output_files, GUINT_TO_POINTER(key));
    if (result == NULL) {
        char name[40];
        int cchip;
        int rc;
        switch (LF_DATA_FRAME_FTYPE(df)) {
        case LF_LTA_DATA_FRAME_FTYPE:
            cchip = LF_LTA_DATA_FRAME_CCHIP_ID(df);
            rc = g_snprintf(name, sizeof(name), "CC_c%ur%u_%04X_%s.blf",
                            cchip & 0x7, (cchip >> 3) & 0x7,
                            LF_LTA_DATA_FRAME_BOARD_SN(df), timestamp);
            break;
        case LF_CC_DATA_FRAME_FTYPE:
            rc = g_snprintf(name, sizeof(name), "CCC_%s.blf", timestamp);
            break;
        default:
            rc = g_strlcpy(name, "unknowns.blf", sizeof(name));
            break;
        }
        g_assert(rc < sizeof(name));

        result = fopen(name, "w+");
        if (result == NULL) {
            // abort, since we want to avoid repeatedly trying to open
            // the same file
            g_error("Failed to open output file %s: %s", name,
                    strerror(errno));
        }
        g_hash_table_insert(output_files, GUINT_TO_POINTER(key), result);
    }
    return result;
}
