#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>

#include <wcbe_framedumper.h>
#include <wcbe.h>
#include <dump_frame.h>
#include <dump_bin_frame.h>

/* GST_DEBUG_CATEGORY_STATIC(wcbe_framedumper_debug); */
/* #define GST_CAT_DEFAULT wcbe_framedumper_debug */

/* Filter signals and args */
enum
{
    /* FILL ME */
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_FORMAT,
    ARG_DO_OUTPUT
};

#define DEFAULT_FORMAT WCBE_FRAMEDUMP_BINARY

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_RAW_MIME)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_RAW_MIME)
    );

GST_BOILERPLATE(WcbeFramedumper, wcbe_framedumper, GstBaseTransform,
                GST_TYPE_BASE_TRANSFORM);

static void wcbe_framedumper_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void wcbe_framedumper_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static gboolean wcbe_framedumper_start(GstBaseTransform *trans);
static gboolean wcbe_framedumper_stop(GstBaseTransform *trans);
static GstFlowReturn wcbe_framedumper_transform_ip(GstBaseTransform *trans,
                                                   GstBuffer *buffer);
static void set_format(WcbeFramedumper *framedumper,
                       WcbeFramedumpFormat format);

static void
wcbe_framedumper_base_init(gpointer gclass)
{
    static GstElementDetails element_details = {
        "WIDAR frame dump",
        "Transform/WIDAR",
        "Dump WIDAR binary frames to files",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_framedumper_class_init(WcbeFramedumperClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GstBaseTransformClass *gstbasetrans_class =
        GST_BASE_TRANSFORM_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_framedumper_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_framedumper_get_property);

    g_object_class_install_property(
        gobject_class, ARG_DO_OUTPUT,
        g_param_spec_boolean("do-output", "Dump frames",
                             "Produce binary frame dump files?",
                             FALSE, G_PARAM_READWRITE));
    g_object_class_install_property(
        gobject_class, ARG_FORMAT,
        g_param_spec_enum("format", "Format", "Output format switch",
                          WCBE_TYPE_FRAMEDUMP_FORMAT, DEFAULT_FORMAT,
                          G_PARAM_READWRITE));

    gstbasetrans_class->start = GST_DEBUG_FUNCPTR(wcbe_framedumper_start);
    gstbasetrans_class->stop = GST_DEBUG_FUNCPTR(wcbe_framedumper_stop);
    gstbasetrans_class->transform_ip =
        GST_DEBUG_FUNCPTR(wcbe_framedumper_transform_ip);
    gstbasetrans_class->passthrough_on_same_caps = TRUE;
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_framedumper_init(WcbeFramedumper *framedumper,
                      WcbeFramedumperClass *gclass)
{
    set_format(framedumper, DEFAULT_FORMAT);
    framedumper->do_output = FALSE;
}

static void
wcbe_framedumper_set_property(GObject *object, guint prop_id,
                              const GValue *value, GParamSpec *pspec)
{
    WcbeFramedumper *framedumper = WCBE_FRAMEDUMPER(object);

    switch (prop_id) {
      case ARG_FORMAT:
        set_format(framedumper, g_value_get_enum(value));
        break;
      case ARG_DO_OUTPUT:
        framedumper->do_output = g_value_get_boolean(value);
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_framedumper_get_property(GObject *object, guint prop_id,
                              GValue *value, GParamSpec *pspec)
{
    WcbeFramedumper *framedumper = WCBE_FRAMEDUMPER(object);

    switch (prop_id) {
      case ARG_FORMAT:
        g_value_set_enum(value, framedumper->format);
        break;
      case ARG_DO_OUTPUT:
        g_value_set_boolean(value, framedumper->do_output);
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/* GstBaseSink vmethod implementations */

static gboolean
wcbe_framedumper_start(GstBaseTransform *trans)
{
    return TRUE;
}

static gboolean
wcbe_framedumper_stop(GstBaseTransform *trans)
{
    dump_bin_frame_close_all();
    dump_frame_close_all();
    return TRUE;
}

static GstFlowReturn
wcbe_framedumper_transform_ip(GstBaseTransform *trans, GstBuffer *buf)
{
    WcbeFramedumper *framedumper = WCBE_FRAMEDUMPER(trans);
    if (framedumper->do_output) {
        gsize (*output_fn)(const lf_data_frame_word_t *);
        output_fn = g_atomic_pointer_get(&(framedumper->output_fn));
        output_fn((lf_data_frame_word_t *)GST_BUFFER_DATA(buf));
    }
    return GST_FLOW_OK;
}

GType
wcbe_framedump_format_get_type(void)
{
    static GType framedump_format_type = 0;
    if (framedump_format_type == 0) {
        static GEnumValue format_types[] = {
            {WCBE_FRAMEDUMP_ASCII, "ascii", "ASCII format"},
            {WCBE_FRAMEDUMP_BINARY, "binary", "Binary format"},
            {0, NULL, NULL}
        };
        framedump_format_type =
            g_enum_register_static("WcbeFramedumpFormat", format_types);
    }
    return framedump_format_type;
}

static void
set_format(WcbeFramedumper *framedumper, WcbeFramedumpFormat format)
{
    framedumper->format = format;
    switch (framedumper->format) {
      case WCBE_FRAMEDUMP_BINARY:
        g_atomic_pointer_set(&(framedumper->output_fn), dump_bin_frame);
        break;
      case WCBE_FRAMEDUMP_ASCII:
        g_atomic_pointer_set(&(framedumper->output_fn), dump_frame);
        break;
      default:
        g_assert_not_reached();
        break;
    }
}
