/*
 * wcbe_lag_frame.c
 *
 * Lag frame utility functions.
 *
 */
#if HAVE_CONFIG_H
# include <config.h>
#endif

#include <stdio.h>
#include <string.h>
#include <netinet/in.h>
#include <glib.h>

#include <wcbe_lag_frame.h>

#if SIZEOF_UINT32_T != SIZEOF_LF_DATA_FRAME_WORD_T
# error "lag_frame module implementation doesn't support data frame word size"
#endif

// bit masks to apply to header fields of data frames to make key
// values
static const lf_data_frame_word_t cc_frame_mask[] = {
    0x000001FF, // W1 (of corrrelator chip frame)
    0xFFFFFFFF, // W2
    0xFFFFFFFF, // W3
    0x0         // unused
};
static const lf_data_frame_word_t lta_frame_mask[] = {
    0xFFFF1FFF, // W1 (of LTA frame)
    0xFFFFFFFF, // W2
    0x0000FFFF, // W3
    0xFFFFFFFF  // W8
};

static int8_t encode_num_lags(unsigned num_lags) ATTRIBUTE((__const__));
static int decode_nlags(uint8_t nlags) ATTRIBUTE((__const__));
static int to_cc_lag_frame(lf_cc_lag_frame_t *lag_frame,
                           const lf_data_frame_word_t data_frame[])
    ATTRIBUTE((__nonnull__));
static int to_lta_lag_frame(lf_lta_lag_frame_t *lag_frame,
                            const lf_data_frame_word_t data_frame[])
    ATTRIBUTE((__nonnull__));

/* lf_to_cc_lag_frame
 *
 * Convert a correlator chip data frame to a lf_cc_lag_frame_t. Data
 * frame is expected to be in network byte-order.
 *
 * *Arguments*
 *
 * data_frame : data frame array (as produced by correlator chip)
 *
 * lag_frame : pointer to lf_cc_lag_frame_t structure, with a 'lags'
 * array of at least 'max_lags' elements
 *
 * max_lags : maximum number of lags that can be stored in 'lag_frame'
 *
 * *Return values*
 *
 * LF_SUCCESS : 'data_frame' successfully converted to 'lag_frame',
 * with verified checksum
 *
 * LF_E_TOO_SMALL : the number of lags in 'data_frame' exceeds
 * 'max_lags'
 *
 * LF_E_CHECKSUM : checksum of 'data_frame' could not be verified
 *
 * LF_E_START_SYNC : start sync word not present
 *
 * LF_E_END_SYNC : end sync word not present
 *
 * LF_E_FTYPE : CMMD field value not equal to LF_CC_DATA_FRAME_FTYPE
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_CHECKSUM : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_START_SYNC, LF_E_END_SYNC, LF_E_FTYPE, LF_E_TOO_SMALL : no
 * side-effects
 */
int
lf_to_cc_lag_frame(lf_cc_lag_frame_t *lag_frame,
                   unsigned max_lags,
                   const lf_data_frame_word_t data_frame[])
{
    int result;
    uint16_t num_lags;

    /* check start sync word */
    if (G_UNLIKELY(data_frame[0] != LF_START_SYNC_WORD)) {
        result = LF_E_START_SYNC;
        goto done;
    }

    /* get number of lags and compare with 'max_lags' */
    num_lags = LF_BITFIELD_EXTRACT(g_ntohl(data_frame[1]), 10, 9);
    if (G_UNLIKELY(num_lags > max_lags)) {
        result = LF_E_TOO_SMALL;
        goto done;
    }

    /* check end sync word */
    if (G_UNLIKELY(data_frame[LF_DATA_FRAME_NUM_WORDS(num_lags) - 2]
                   != LF_END_SYNC_WORD)) {
        result = LF_E_END_SYNC;
        goto done;
    }

    /* check station ids */
    lf_data_frame_word_t word = g_ntohl(data_frame[2]);
    if (G_UNLIKELY(LF_BITFIELD_EXTRACT(word, 8, 0) == 0
                   || LF_BITFIELD_EXTRACT(word, 8, 16) == 0)) {
        result = LF_E_STATION_ZERO;
        goto done;
    }

    result = to_cc_lag_frame(lag_frame, data_frame);

 done:
    return result;
}

/* lf_to_lta_lag_frame
 *
 * Convert an LTA data frame to a lf_lta_lag_frame_t. Data frame is
 * expected to be in network byte-order.
 *
 * *Arguments*
 *
 * data_frame : data frame array (as produced by LTA)
 *
 * lag_frame : pointer to lf_lta_lag_frame_t structure, with a 'lags'
 * array of at least 'max_lags' elements
 *
 * max_lags : maximum number of lags that can be stored in 'lag_frame'
 *
 * *Return values*
 *
 * LF_SUCCESS : 'data_frame' successfully converted to 'lag_frame',
 * with verified checksum
 *
 * LF_E_TOO_SMALL : the number of lags in 'data_frame' exceeds
 * 'max_lags'
 *
 * LF_E_CHECKSUM : checksum of 'data_frame' could not be verified
 *
 * LF_E_START_SYNC : start sync word not present
 *
 * LF_E_END_SYNC : end sync word not present
 *
 * LF_E_FTYPE : FType field value not equal to LF_LTA_DATA_FRAME_FTYPE
 *
 * LF_E_NLAGS : 'nlags' field value is invalid
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_CHECKSUM : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_START_SYNC, LF_E_END_SYNC, LF_E_FTYPE, LF_E_NLAGS,
 * LF_E_TOO_SMALL : no side-effects
 */
int
lf_to_lta_lag_frame(lf_lta_lag_frame_t *lag_frame,
                    unsigned max_lags,
                    const lf_data_frame_word_t data_frame[])
{
    int result;
    int nlags;
    uint8_t num_lags;

    /* check start sync word */
    if (G_UNLIKELY(data_frame[0] != LF_START_SYNC_WORD)) {
        result = LF_E_START_SYNC;
        goto done;
    }

    /* get nlags and convert to number of lags */
    nlags = decode_nlags(LF_BITFIELD_EXTRACT(g_ntohl(data_frame[1]), 3, 13));
    if (G_UNLIKELY(nlags < 0)) {
        result = LF_E_NLAGS;
        goto done;
    }
    num_lags = nlags;

    /* compare 'num_lags' with 'max_lags' */
    if (G_UNLIKELY(num_lags > max_lags)) {
        result = LF_E_TOO_SMALL;
        goto done;
    }

    /* check end sync word */
    if (G_UNLIKELY(data_frame[LF_DATA_FRAME_NUM_WORDS(num_lags) - 2]
                   != LF_END_SYNC_WORD)) {
        result = LF_E_END_SYNC;
        goto done;
    }

    /* check station ids */
    lf_data_frame_word_t word = g_ntohl(data_frame[2]);
    if (G_UNLIKELY(LF_BITFIELD_EXTRACT(word, 8, 0) == 0
                   || LF_BITFIELD_EXTRACT(word, 8, 16) == 0)) {
        result = LF_E_STATION_ZERO;
        goto done;
    }

    result = to_lta_lag_frame(lag_frame, data_frame);

 done:
    return result;
}

/* lf_to_cc_data_frame
 *
 * Convert a lf_cc_lag_frame_t to a correlator chip data frame (in
 * network byte-order.)
 *
 * *Arguments*
 *
 * lag_frame : pointer to lf_cc_lag_frame_t structure
 *
 * data_frame : pointer to data frame with space for at least
 * 'max_lags' complex lags
 *
 * max_lags : maximum number of lags that can be stored in
 * 'data_frame'
 *
 * *Return values*
 *
 * LF_SUCCESS : 'lag_frame' successfully converted to 'data_frame'
 *
 * LF_E_TOO_SMALL : the number of lags in 'lag_frame' exceeds
 * 'max_lags'
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'data_frame' is filled with values converted from
 * 'lag_frame' together with sync words and checksum word
 *
 * LF_E_TOO_SMALL : no side-effects
 */
int
lf_to_cc_data_frame(lf_data_frame_word_t *data_frame,
                    unsigned max_lags,
                    const lf_cc_lag_frame_t *lag_frame)
{
    int result;
    lf_data_frame_word_t checksum;
    unsigned i;

    /* compare lag_frame->num_lags with 'max_lags' */
    if (lag_frame->lf.num_lags > max_lags) {
        result = LF_E_TOO_SMALL;
        goto done;
    }

    /* check station ids */
    if (LF_CC_FRAME_SID_X(lag_frame) == 0
        || LF_CC_FRAME_SID_Y(lag_frame) == 0) {
        result = LF_E_STATION_ZERO;
        goto done;
    }

    /*
     * Extract fields from 'lag_frame' to 'data_frame'
     */
    /* word 0 */
    *data_frame = g_ntohl(LF_START_SYNC_WORD);
    checksum = *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 1 */
    *data_frame = lag_frame->lf.key[0]
        | LF_BITFIELD_PART(lag_frame->lf.status, 7, 25)
        | LF_BITFIELD_PART(lag_frame->lf.num_lags, 10, 9);
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 2 */
    *data_frame = lag_frame->lf.key[1];
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 3 */
    *data_frame = lag_frame->lf.key[2];
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 4 */
    *data_frame = lag_frame->lf.timestamp.seconds;
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 5 */
    *data_frame = LF_BITFIELD_PART(lag_frame->lf.timestamp.clock_cycles, 29, 0)
        | LF_BITFIELD_PART(lag_frame->lf.timestamp.epoch, 3, 29);
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 6 */
    *data_frame = lag_frame->lf.data_valid_center;
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 7 */
    *data_frame = lag_frame->lf.data_valid_edge;
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 8 */
    *data_frame = lag_frame->data_bias;
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* lag words */
    for (i = 0; i < lag_frame->lf.num_lags; ++i) {
        *data_frame = LF_LAG_REAL(lag_frame->lf.lags[i]);
        checksum += *data_frame;
        *data_frame = g_htonl(*data_frame);
        ++data_frame;
        *data_frame = LF_LAG_IMAG(lag_frame->lf.lags[i]);
        checksum += *data_frame;
        *data_frame = g_htonl(*data_frame);
        ++data_frame;
    }

    /* end sync word */
    *data_frame = g_ntohl(LF_END_SYNC_WORD);
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* write checksum */
    *data_frame = g_htonl(checksum);
    result = LF_SUCCESS;

 done:
    return result;
}

/* lf_to_lta_data_frame
 *
 * Convert a lf_lta_lag_frame_t to an LTA data frame array (in network
 * byte-order.)
 *
 * *Arguments*
 *
 * lag_frame : pointer to lf_lta_lag_frame_t structure
 *
 * data_frame : pointer to data frame with space for at least
 * 'max_lags' complex lags
 *
 * max_lags : maximum number of lags that can be stored in
 * 'data_frame'
 *
 * *Return values*
 *
 * LF_SUCCESS : 'lag_frame' successfully converted to 'data_frame'
 *
 * LF_E_TOO_SMALL : the number of lags in 'lag_frame' exceeds
 * 'max_lags'
 *
 * LF_E_NLAGS : the value of 'lag_frame->num_lags' field is invalid
 * (i.e, it cannot be encoded)
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'data_frame' is filled with values converted from
 * 'lag_frame' together with sync words and checksum word
 *
 * LF_E_TOO_SMALL, LF_E_NLAGS : no side-effects
 */
int
lf_to_lta_data_frame(lf_data_frame_word_t *data_frame,
                     unsigned max_lags,
                     const lf_lta_lag_frame_t *lag_frame)
{
    int result;
    int8_t nlags;
    lf_data_frame_word_t checksum;
    unsigned i;

    /* compare lag_frame->num_lags with 'max_lags' */
    if (lag_frame->lf.num_lags > max_lags) {
        result = LF_E_TOO_SMALL;
        goto done;
    }

    /* encode lag_frame->num_lags */
    nlags = encode_num_lags(lag_frame->lf.num_lags);
    if (nlags < 0) {
        result = LF_E_NLAGS;
        goto done;
    }

    /* check station ids */
    if (LF_LTA_FRAME_SID_X(lag_frame) == 0
        || LF_LTA_FRAME_SID_X(lag_frame) == 0) {
        result = LF_E_STATION_ZERO;
        goto done;
    }

    /*
     * Extract fields from 'lag_frame' to 'data_frame'
     */
    /* word 0 */
    *data_frame = g_ntohl(LF_START_SYNC_WORD);
    checksum = *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 1 */
    *data_frame = lag_frame->lf.key[0]
        | LF_BITFIELD_PART(nlags, 3, 13);
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 2 */
    *data_frame = lag_frame->lf.key[1];
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 3 */
    *data_frame = lag_frame->lf.key[2]
        | LF_BITFIELD_PART(lag_frame->lf.status, 7, 25)
        | LF_BITFIELD_PART(lag_frame->frame_count, 9, 16);
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 4 */
    *data_frame = lag_frame->lf.timestamp.seconds;
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 5 */
    *data_frame = LF_BITFIELD_PART(lag_frame->lf.timestamp.clock_cycles, 29, 0)
        | LF_BITFIELD_PART(lag_frame->lf.timestamp.epoch, 3, 29);
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 6 */
    *data_frame = lag_frame->lf.data_valid_center;
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 7 */
    *data_frame = lag_frame->lf.data_valid_edge;
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* word 8 */
    *data_frame = lag_frame->lf.key[3];
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* lag words */
    for (i = 0; i < lag_frame->lf.num_lags; ++i) {
        *data_frame = LF_LAG_REAL(lag_frame->lf.lags[i]);
        checksum += *data_frame;
        *data_frame = g_htonl(*data_frame);
        ++data_frame;
        *data_frame = LF_LAG_IMAG(lag_frame->lf.lags[i]);
        checksum += *data_frame;
        *data_frame = g_htonl(*data_frame);
        ++data_frame;
    }

    /* end sync word */
    *data_frame = g_ntohl(LF_END_SYNC_WORD);
    checksum += *data_frame;
    *data_frame = g_htonl(*data_frame);
    ++data_frame;

    /* write checksum */
    *data_frame = g_htonl(checksum);
    result = LF_SUCCESS;

 done:
    return result;
}

/* encode_num_lags
 *
 * Encode a number of lags into the form required by the LTA data
 * fraem 'nlags' field.
 *
 * *Arguments*
 *
 * num_lags - number of lags
 *
 * *Return values*
 *
 * -1 : an error, invalid 'num_lags' value
 *
 * any value except -1 : the integer value (which may need to be
 * truncated) representing the 'num_lags' value in a data frame (_not_
 * a cc data frame)
 */
static int8_t
encode_num_lags(unsigned num_lags)
{
    int8_t result;

    switch (num_lags) {
    case 64:
        result = 1;
        break;
    case 128:
        result = 0;
        break;
    default:
        result = -1;
        break;
    }
    return result;
}

/* decode_num_lags
 *
 * Decode the 'nlags' field of a data frame into a number of lags.
 *
 * *Arguments*
 *
 * nlags - value of nlags field
 *
 * *Return values*
 *
 * -1 : an error, invalid 'nlags' value
 *
 * any value except -1 : the decoded integer value given by the
 * 'nlags' value of a data frame (_not_ a correlator chip data frame)
 */
static int
decode_nlags(uint8_t nlags)
{
    int result;

    switch (nlags) {
    case 0:
        result = 128;
        break;
    case 1:
        result = 64;
        break;
    default:
        result = -1;
        break;
    }
    return result;
}

/* to_cc_lag_frame
 *
 * Convert a correlator chip data frame to a lf_cc_lag_frame_t. Data
 * frame is expected to be in network byte-order. Sync words in the
 * data frame are not checked.
 *
 * *Arguments*
 *
 * data_frame : data frame array (as produced by correlator chip)
 *
 * lag_frame : pointer to lf_cc_lag_frame_t structure (assume that at least
 * at least 'num_lags' elements can be written to the 'lags' array)
 *
 * *Return values*
 *
 * LF_SUCCESS : 'data_frame' successfully converted to 'lag_frame',
 * with verified checksum
 *
 * LF_E_CHECKSUM : checksum of 'data_frame' could not be verified
 *
 * LF_E_FTYPE : CMMD field value not equal to LF_CC_DATA_FRAME_FTYPE
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_CHECKSUM : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_FTYPE : no side-effects
 */
static int
to_cc_lag_frame(lf_cc_lag_frame_t *lag_frame,
                const lf_data_frame_word_t data_frame[])
{
    int result;
    lf_data_frame_word_t word;
    const lf_data_frame_word_t *df;
    lf_data_frame_word_t checksum;
    unsigned i;

    /* check CMMD value */
    if (G_UNLIKELY(LF_BITFIELD_EXTRACT(g_ntohl(data_frame[1]), 3, 0)
                   != LF_CC_DATA_FRAME_FTYPE)) {
        result = LF_E_FTYPE;
        goto done;
    }

    // We need to set the fourth element of lag_frame->lf.key to zero, or
    // the hashing and comparison of keys across both types of lag
    // frames won't work (or would need to be complicated with extra
    // tests).
    lag_frame->lf.key[3] = 0;

    /*
     * Extract fields from 'data_frame' to 'lag_frame'
     */
    /* word 1 */
    df = &data_frame[1];
    word = g_ntohl(*df);
    checksum = g_ntohl(data_frame[0]) + word;
    lag_frame->lf.status = LF_BITFIELD_EXTRACT(word, 7, 25);
    lag_frame->lf.num_lags = LF_BITFIELD_EXTRACT(word, 10, 9);
    lag_frame->lf.key[0] = word & cc_frame_mask[0];

    /* word 2 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.key[1] = word & cc_frame_mask[1];

    /* word 3 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.key[2] = word & cc_frame_mask[2];

    /* word 4 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.timestamp.seconds = word;

    /* word 5 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.timestamp.clock_cycles = LF_BITFIELD_EXTRACT(word, 29, 0);
    lag_frame->lf.timestamp.epoch = LF_BITFIELD_EXTRACT(word, 3, 29);

    /* word 6 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.data_valid_center = word;

    /* word 7 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.data_valid_edge = word;

    /* word 8 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->data_bias = word;

    /* lag words */
    for (i = 0; i < lag_frame->lf.num_lags; ++i) {
        ++df;
        word = g_ntohl(*df);
        checksum += word;
        LF_LAG_REAL(lag_frame->lf.lags[i]) = word;
        ++df;
        word = g_ntohl(*df);
        checksum += word;
        LF_LAG_IMAG(lag_frame->lf.lags[i]) = word;
    }

    /* end sync word */
    ++df;
    word = g_ntohl(*df);
    checksum += word;

    /* verify checksum */
    ++df;
    word = g_ntohl(*df);
    if (word == checksum) {
        result = LF_SUCCESS;
    }
    else {
        result = LF_E_CHECKSUM;
    }
 done:
    return result;
}

/* to_lta_lag_frame
 *
 * Convert an LTA data frame to a lf_lta_lag_frame_t. Data frame is
 * expected to be in network byte-order. Sync words in the data frame
 * are not checked.
 *
 * *Arguments*
 *
 * data_frame : data frame array (as produced by LTA)
 *
 * lag_frame : pointer to lf_lta_lag_frame_t structure (assume that at
 * least at least 'num_lags' elements can be written to the 'lags'
 * array)
 *
 * *Return values*
 *
 * LF_SUCCESS : 'data_frame' successfully converted to 'lag_frame',
 * with verified checksum
 *
 * LF_E_CHECKSUM : checksum of 'data_frame' could not be verified
 *
 * LF_E_FTYPE : FType field value not equal to LF_LTA_DATA_FRAME_FTYPE
 *
 * LF_E_NLAGS : 'nlags' field value is invalid
 *
 * *Side effects consequent to the following return values*
 *
 * LF_SUCCESS : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_CHECKSUM : 'lag_frame' is filled with values converted from
 * 'data_frame'
 *
 * LF_E_FTYPE, LF_E_NLAGS: no side-effects
 */
static int
to_lta_lag_frame(lf_lta_lag_frame_t *lag_frame,
                 const lf_data_frame_word_t data_frame[])
{
    int result;
    int nlags;
    lf_data_frame_word_t word;
    const lf_data_frame_word_t *df;
    lf_data_frame_word_t checksum;
    unsigned i;

    /* get nlags and convert to number of lags */
    word = g_ntohl(data_frame[1]);
    nlags = decode_nlags(LF_BITFIELD_EXTRACT(word, 3, 13));
    if (G_UNLIKELY(nlags < 0)) {
        result = LF_E_NLAGS;
        goto done;
    }

    /* check FType value */
    if (G_UNLIKELY(LF_BITFIELD_EXTRACT(word, 3, 0)
                   != LF_LTA_DATA_FRAME_FTYPE)) {
        result = LF_E_FTYPE;
        goto done;
    }

    /*
     * Extract fields from 'data_frame' to 'lag_frame'
     */
    /* word 1 */
    df = &data_frame[1];
    word = g_ntohl(*df);
    checksum = g_ntohl(data_frame[0]) + word;
    lag_frame->lf.num_lags = nlags;
    lag_frame->lf.key[0] = word & lta_frame_mask[0];

    /* word 2 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.key[1] = word & lta_frame_mask[1];

    /* word 3 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.status = LF_BITFIELD_EXTRACT(word, 7, 25);
    lag_frame->frame_count = LF_BITFIELD_EXTRACT(word, 9, 16);
    lag_frame->lf.key[2] = word & lta_frame_mask[2];

    /* word 4 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.timestamp.seconds = word;

    /* word 5 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.timestamp.clock_cycles = LF_BITFIELD_EXTRACT(word, 29, 0);
    lag_frame->lf.timestamp.epoch = LF_BITFIELD_EXTRACT(word, 3, 29);

    /* word 6 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.data_valid_center = word;

    /* word 7 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.data_valid_edge = word;

    /* word 8 */
    ++df;
    word = g_ntohl(*df);
    checksum += word;
    lag_frame->lf.key[3] = word & lta_frame_mask[3];

    /* lag words */
    for (i = 0; i < lag_frame->lf.num_lags; ++i) {
        ++df;
        word = g_ntohl(*df);
        checksum += word;
        LF_LAG_REAL(lag_frame->lf.lags[i]) = word;
        ++df;
        word = g_ntohl(*df);
        checksum += word;
        LF_LAG_IMAG(lag_frame->lf.lags[i]) = word;
    }

    /* end sync word */
    ++df;
    word = g_ntohl(*df);
    checksum += word;

    /* verify checksum */
    ++df;
    word = g_ntohl(*df);
    if (word == checksum) {
        result = LF_SUCCESS;
    }
    else {
        result = LF_E_CHECKSUM;
    }
 done:
    return result;
}

/* lf_lta_lag_frame_equal
 *
 * Test two 'lf_lta_lag_frame_t's for equality. Note that a simple memcmp
 * of the space isn't sufficient because of unused bits in bitfields
 * of the structure.
 */
int
lf_lta_lag_frame_equal(const lf_lta_lag_frame_t *frame1,
                       const lf_lta_lag_frame_t *frame2)
{
    return (frame1->lf.key[0] == frame2->lf.key[0]
            && frame1->lf.key[1] == frame2->lf.key[1]
            && frame1->lf.key[2] == frame2->lf.key[2]
            && frame1->lf.key[3] == frame2->lf.key[3]
            && frame1->lf.num_lags == frame2->lf.num_lags
            && frame1->lf.status == frame2->lf.status
            && frame1->frame_count == frame2->frame_count
            && frame1->lf.timestamp.seconds == frame2->lf.timestamp.seconds
            && frame1->lf.timestamp.clock_cycles ==
            frame2->lf.timestamp.clock_cycles
            && frame1->lf.timestamp.epoch == frame2->lf.timestamp.epoch
            && frame1->lf.data_valid_center == frame2->lf.data_valid_center
            && frame1->lf.data_valid_edge == frame2->lf.data_valid_edge
            && memcmp(frame1->lf.lags, frame2->lf.lags,
                      frame1->lf.num_lags * sizeof(lf_lag_t)) == 0);
}

/* lf_cc_lag_frame_equal
 *
 * Test two 'lf_cc_lag_frame_t's for equality. Note that a simple
 * memcmp of the space isn't sufficient because of unused bits in
 * bitfields of the structure.
 */
int
lf_cc_lag_frame_equal(const lf_cc_lag_frame_t *frame1,
                      const lf_cc_lag_frame_t *frame2)
{
    return (frame1->lf.key[0] == frame2->lf.key[0]
            && frame1->lf.key[1] == frame2->lf.key[1]
            && frame1->lf.key[2] == frame2->lf.key[2]
            && frame1->lf.status == frame2->lf.status
            && frame1->lf.num_lags == frame2->lf.num_lags
            && frame1->lf.timestamp.seconds == frame2->lf.timestamp.seconds
            && frame1->lf.timestamp.clock_cycles ==
            frame2->lf.timestamp.clock_cycles
            && frame1->lf.timestamp.epoch == frame2->lf.timestamp.epoch
            && frame1->lf.data_valid_center == frame2->lf.data_valid_center
            && frame1->lf.data_valid_edge == frame2->lf.data_valid_edge
            && frame1->data_bias == frame2->data_bias
            && memcmp(frame1->lf.lags, frame2->lf.lags,
                      frame1->lf.num_lags * sizeof(lf_lag_t)) == 0);
}

/* lf_strerror
 *
 * Returns string corresponding to given error code from this module.
 * The string should only be used until the next call to this function
 * (which therefore is not thread-safe).
 */
const char *
lf_strerror(int errnum)
{
    const char *result;
    static const char success[] = "success";
    static const char too_small[] = "buffer too small";
    static const char checksum[] = "checksum verification failed";
    static const char start_sync[] = "invalid lag frame start sync word";
    static const char end_sync[] = "invalid lag frame end sync word";
    static const char ftype[] = "invalid lag frame FTYPE value";
    static const char nlags[] = "invalid lag frame NLAGS value";
    static const char stn_zero[] = "invalid station id value (0)";
    static const char unknown[] = "unknown error (%d)";
    static char *str = NULL;

    switch (errnum) {
    case LF_SUCCESS:
        result = success;
        break;
    case LF_E_TOO_SMALL:
        result = too_small;
        break;
    case LF_E_CHECKSUM:
        result = checksum;
        break;
    case LF_E_START_SYNC:
        result = start_sync;
        break;
    case LF_E_END_SYNC:
        result = end_sync;
        break;
    case LF_E_FTYPE:
        result = ftype;
        break;
    case LF_E_NLAGS:
        result = nlags;
        break;
    case LF_E_STATION_ZERO:
        result = stn_zero;
        break;
    default:
        if (str != NULL) {
            g_free(str);
        }
        str = g_strdup_printf(unknown, errnum);
        result = str;
        break;
    }
    return result;
}
