#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/epoll.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>

#include <wcbe_lagset_sink.h>
#include <wcbe.h>
#include <user_marshal.h>

typedef struct pipelinesock {
    struct sockaddr_un addr;
} pipelinesock_t;

GST_DEBUG_CATEGORY_STATIC(wcbe_lagset_sink_debug);
#define GST_CAT_DEFAULT wcbe_lagset_sink_debug

/* Filter signals and args */
enum
{
    SIGNAL_CREATE_CFG,
    SIGNAL_ACTIVATE_CFG,
    SIGNAL_DEACTIVATE_CFG,
    SIGNAL_DESTROY_CFG,
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_SOCKDIR,
    ARG_CONFIGS,
    ARG_ACTIVE_CONFIG
};

static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_LAGSET_MIME));

static guint wcbe_lagset_sink_signals[LAST_SIGNAL] = {0};

GST_BOILERPLATE(WcbeLagsetSink, wcbe_lagset_sink, GstBaseSink,
                GST_TYPE_BASE_SINK);

static void wcbe_lagset_sink_set_property(
    GObject * object, guint prop_id, const GValue * value, GParamSpec * pspec);
static void wcbe_lagset_sink_get_property(
    GObject * object, guint prop_id, GValue * value, GParamSpec * pspec);
static void wcbe_lagset_sink_finalize(GObject *object);
static gboolean wcbe_lagset_sink_event(GstBaseSink *sink, GstEvent *event);

static GstFlowReturn wcbe_lagset_sink_render(
    GstBaseSink *sink, GstBuffer *buf);
static gboolean wcbe_lagset_sink_start(GstBaseSink *sink);
static gboolean wcbe_lagset_sink_stop(GstBaseSink *sink);
static void send_stream_object(gint sock, struct msghdr *msghdr);
static void forward_event(WcbeLagsetSink *sink, GstEventType event_type);
static wcbe_lagset_sink_cfg_t *create_config(const gchar *sockdir,
                                             const gchar *config,
                                             const gchar *name);
static void pipeline_config(wcbe_lagset_sink_cfg_t *cfg,
                            xmlXPathContextPtr context, xmlNodePtr pipeline);
static void deactivate_config(WcbeLagsetSink *sink);
static void destroy_config(gpointer data);
static void destroy_config_silently(wcbe_lagset_sink_cfg_t *cfg);
static gint open_sock(WcbeLagsetSink *sink);
static void close_sock(WcbeLagsetSink *sink);

static void
wcbe_lagset_sink_base_init(gpointer gclass)
{
    static GstElementDetails element_details = {
        "WIDAR lagset sink",
        "Sink/WIDAR",
        "Write WIDAR lag sets to Unix domain sockets",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_lagset_sink_class_init(WcbeLagsetSinkClass *klass)
{
    GObjectClass *gobject_class  = (GObjectClass *)klass;
    GstBaseSinkClass *gstbasesink_class = (GstBaseSinkClass *)klass;

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_finalize);

    g_object_class_install_property(
        gobject_class, ARG_SOCKDIR,
        g_param_spec_string("sockdir", "Socket dir",
                            "Directory for lagset sockets",
                            NULL, G_PARAM_READWRITE));
    GParamSpec *config_spec =
        g_param_spec_string("config", "Configuration name",
                            "Configuration name", NULL,
                            G_PARAM_READABLE);
    g_object_class_install_property(
        gobject_class, ARG_CONFIGS,
        g_param_spec_value_array("configs", "Available configurations",
                                 "Get list of all available configurations",
                                 config_spec, G_PARAM_READABLE));
    g_object_class_install_property(
        gobject_class, ARG_ACTIVE_CONFIG,
        g_param_spec_string("active-config", "Active configuration",
                            "Get name of active configuration",
                            NULL, G_PARAM_READABLE));

    gstbasesink_class->render =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_render);
    gstbasesink_class->start =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_start);
    gstbasesink_class->stop =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_stop);
    gstbasesink_class->event =
        GST_DEBUG_FUNCPTR(wcbe_lagset_sink_event);

    wcbe_lagset_sink_signals[SIGNAL_CREATE_CFG] =
        g_signal_new("config-created", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    wcbe_lagset_sink_signals[SIGNAL_ACTIVATE_CFG] =
        g_signal_new("config-activated", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    wcbe_lagset_sink_signals[SIGNAL_DEACTIVATE_CFG] =
        g_signal_new("config-deactivated", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_marshal_VOID__STRING,
                     G_TYPE_NONE, 1, G_TYPE_STRING);

    wcbe_lagset_sink_signals[SIGNAL_DESTROY_CFG] =
        g_signal_new("config-destroyed", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_lagset_sink_init(WcbeLagsetSink *lsink,
                      WcbeLagsetSinkClass *gclass)
{
    GstBaseSink *basesink = GST_BASE_SINK(lsink);
    basesink->sync = FALSE;
    lsink->sockdir = NULL;
    lsink->configurations =
        g_hash_table_new_full(g_str_hash, g_str_equal, NULL, destroy_config);
    lsink->config = NULL;
    lsink->sockfd = -1;
}

static void
wcbe_lagset_sink_finalize(GObject *object)
{
    WcbeLagsetSink *lsink = WCBE_LAGSET_SINK(object);

    deactivate_config(lsink);
    if (lsink->configurations != NULL) {
        g_hash_table_destroy(lsink->configurations);
    }
    close_sock(lsink);
    if (lsink->sockdir != NULL) {
        g_free(lsink->sockdir);
    }
    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static gboolean
wcbe_lagset_sink_start(GstBaseSink *sink)
{
    WcbeLagsetSink *lsink = WCBE_LAGSET_SINK(sink);
    lsink->config = NULL;
    return TRUE;
}

static gboolean
wcbe_lagset_sink_stop(GstBaseSink *sink)
{
    WcbeLagsetSink *lsink = WCBE_LAGSET_SINK(sink);
    deactivate_config(lsink);
    return TRUE;
}

static void
wcbe_lagset_sink_set_property(GObject *object, guint prop_id,
                              const GValue *value, GParamSpec *pspec)
{
    WcbeLagsetSink *sink = WCBE_LAGSET_SINK(object);

    switch (prop_id) {
      case ARG_SOCKDIR: {
          const gchar *sockdir = g_value_get_string(value);
          if (sink->sockdir == NULL || strcmp(sink->sockdir, sockdir) != 0) {
              if (sink->sockdir != NULL) {
                  // Can't transfer configurations (sockets) to new
                  // directory, so just destroy them.
                  deactivate_config(sink);
                  g_hash_table_remove_all(sink->configurations);
                  close_sock(sink);
                  g_free(sink->sockdir);
              }
              sink->sockdir = g_strdup(sockdir);
              gint rc = open_sock(sink);
              if (rc != 0) {
                  g_free(sink->sockdir);
                  sink->sockdir = NULL;
              }
          }
          break;
      }
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_lagset_sink_get_property(GObject *object, guint prop_id,
                              GValue *value, GParamSpec *pspec)
{
    WcbeLagsetSink *sink = WCBE_LAGSET_SINK(object);

    switch (prop_id) {
      case ARG_SOCKDIR:
        g_value_set_string(value, sink->sockdir ? sink->sockdir : "");
        break;
      case ARG_ACTIVE_CONFIG:
        g_value_set_string(value, sink->config ? sink->config->name : "");
        break;
      case ARG_CONFIGS: {
          GValueArray *varray =
              g_value_array_new(g_hash_table_size(sink->configurations));
          GValue str = {0,};
          g_value_init(&str, G_TYPE_STRING);
          void get_name(gpointer key, gpointer value, gpointer user_data) {
              const gchar *name = key;
              const wcbe_lagset_sink_cfg_t *config = value;
              const gchar *fmt;
              if (config->destroy) {
                  fmt = "[%s]";
              }
              else if (config == sink->config) {
                  fmt = "<%s>";
              }
              else {
                  fmt = "(%s)";
              }
              gchar *tagged_name = g_strdup_printf(fmt, name);
              g_value_set_string(&str, tagged_name);
              g_value_array_append(varray, &str);
              g_free(tagged_name);
          }
          g_hash_table_foreach(sink->configurations, get_name, NULL);
          g_value_take_boxed(value, varray);
          break;
      }
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/* GstBaseSink vmethod implementations */

/* render function
 * this function does the actual processing
 */

static GstFlowReturn
wcbe_lagset_sink_render(GstBaseSink *sink, GstBuffer *buf)
{
    GstFlowReturn result;
    WcbeLagsetSink *lsink;

    lsink = WCBE_LAGSET_SINK(sink);
    wcbe_stream_lagset_t stream_lagset;
    stream_lagset.object.type = WCBE_STREAM_LAGSET;
    lagset_t *lagset = (lagset_t *)GST_BUFFER_DATA(buf);
    pipelinesock_t *pipelinesock =
        g_hash_table_lookup(lsink->config->destinations,
                            GINT_TO_POINTER(lagset->id));
    if (G_LIKELY(pipelinesock != NULL)) {
        // Use iovec for message---should involve less data copying.
        struct iovec iovec[2];
        wcbe_stream_lagset_init_iovec(
            &stream_lagset, lagset, GST_BUFFER_SIZE(buf), iovec);
        g_strlcpy(stream_lagset.product_id,
                  g_quark_to_string(lagset->id),
                  sizeof(stream_lagset.product_id));
        struct msghdr msghdr = {.msg_name = &(pipelinesock->addr),
                                .msg_namelen = sizeof(struct sockaddr_un),
                                .msg_control = NULL, .msg_controllen = 0,
                                .msg_iov = iovec,
                                .msg_iovlen = G_N_ELEMENTS(iovec)};
        result = GST_FLOW_OK;
        send_stream_object(lsink->sockfd, &msghdr);
    }
    else {
        GST_INFO("Lagset with unexpected 'id' received...and dropped");
        result = GST_FLOW_ERROR;
    }
    return result;
}

static gboolean
wcbe_lagset_sink_event(GstBaseSink *sink, GstEvent *event)
{
    gboolean result;
    WcbeLagsetSink *lsink = WCBE_LAGSET_SINK(sink);
    switch (GST_EVENT_TYPE(event)) {
      case GST_EVENT_EOS:
      case GST_EVENT_FLUSH_STOP:
      case GST_EVENT_FLUSH_START:
        forward_event(lsink, GST_EVENT_TYPE(event));
        result = TRUE;
        break;

      case WCBE_EVENT_CREATE_CONFIG:
        g_assert(gst_structure_has_name(event->structure, "config-create"));
        g_assert(gst_structure_has_field(event->structure, "name"));
        g_assert(gst_structure_has_field(event->structure, "config"));
        wcbe_lagset_sink_create_config(
            lsink, gst_structure_get_string(event->structure, "config"),
            gst_structure_get_string(event->structure, "name"));
        result = TRUE;
        break;

      case WCBE_EVENT_ACTIVATE_CONFIG:
        g_assert(gst_structure_has_name(event->structure, "config-activate"));
        g_assert(gst_structure_has_field(event->structure, "name"));
        wcbe_lagset_sink_activate_config(
            lsink, gst_structure_get_string(event->structure, "name"));
        result = TRUE;
        break;

      case WCBE_EVENT_DESTROY_CONFIG:
        g_assert(gst_structure_has_name(event->structure, "config-destroy"));
        g_assert(gst_structure_has_field(event->structure, "name"));
        wcbe_lagset_sink_destroy_config(
            lsink, gst_structure_get_string(event->structure, "name"));
        result = TRUE;
        break;

      default:
        result = TRUE;
    }
    return result;
}

static void
send_stream_object(gint sock, struct msghdr *msghdr)
{
    wcbe_stream_object_t *stream_object = msghdr->msg_iov[0].iov_base;
    gsize msg_size(void) {
        gsize result = 0;
        guint i;
        for (i = 0; i < msghdr->msg_iovlen; ++i) {
            result += msghdr->msg_iov[i].iov_len;
        }
        return result;
    }
    gsize expected_size;
    switch (stream_object->type) {
      case WCBE_STREAM_LAGSET: {
          lagset_t *lagset = msghdr->msg_iov[msghdr->msg_iovlen - 1].iov_base;
          expected_size = offsetof(wcbe_stream_lagset_t, lagset)
              + SIZEOF_LAGSET_T(lagset->num_frames, lagset->num_lags,
                                lagset->num_bins, lagset->nbits);
          break;
      }
      case WCBE_STREAM_EVENT:
        expected_size = sizeof(wcbe_stream_event_t);
        break;
      default:
        g_assert_not_reached();
        break;
    }
    g_assert(expected_size == msg_size());
    errno = 0;
    gssize bytes_written = sendmsg(sock, msghdr, 0);
    if (G_UNLIKELY(bytes_written != expected_size)) {
        const gchar *errstr;
        if (errno != 0) {
            errstr = g_strerror(errno);
        }
        else {
            errstr = "'errno' not set";
        }
        GST_WARNING("Failed to write stream object to socket %s: %s",
                    ((struct sockaddr_un *)(msghdr->msg_name))->sun_path,
                    errstr);
    }
}

static void
forward_event(WcbeLagsetSink *sink, GstEventType event_type)
{
    wcbe_stream_event_t stream_event;
    stream_event.object.type = WCBE_STREAM_EVENT;
    stream_event.event = event_type;
    struct iovec iovec[] =
        {{.iov_base = &stream_event, .iov_len = sizeof(stream_event)}};
    struct msghdr msghdr = {.msg_name = NULL, .msg_namelen = 0,
                            .msg_control = NULL, .msg_controllen = 0,
                            .msg_iov = iovec, .msg_iovlen = 1};
    guint i;
    for (i = 0; i < sink->config->socks->len; ++i) {
        pipelinesock_t *pipelinesock =
            &g_array_index(sink->config->socks, pipelinesock_t, i);
        msghdr.msg_name = &(pipelinesock->addr);
        msghdr.msg_namelen = sizeof(struct sockaddr_un);
        send_stream_object(sink->sockfd, &msghdr);
    }
}

static wcbe_lagset_sink_cfg_t *
create_config(const gchar *sockdir, const gchar *config, const gchar *name)
{
    g_return_val_if_fail(sockdir != NULL, NULL);
    g_return_val_if_fail(config != NULL, NULL);
    g_return_val_if_fail(name != NULL, NULL);

    gboolean success = FALSE;
    wcbe_lagset_sink_cfg_t *result = NULL;

    xmlDocPtr doc = xmlParseMemory(config, strlen(config));
    if (doc != NULL) {
        xmlNodePtr root = xmlDocGetRootElement(doc);
        if (root != NULL
            && xmlStrEqual(root->name, BAD_CAST "config")) {
            result = g_new0(wcbe_lagset_sink_cfg_t, 1);
            result->destroy = FALSE;
            result->socks = g_array_new(FALSE, FALSE, sizeof(pipelinesock_t));
            result->destinations =
                g_hash_table_new(g_direct_hash, g_direct_equal);
            result->sockdir = g_build_filename(sockdir, name, NULL);
            result->name = g_strdup(name);
            xmlXPathInit();
            xmlXPathContextPtr context = xmlXPathNewContext(doc);
            g_assert(context != NULL);
            context->node = root;
            xmlXPathObjectPtr pipelines = xmlXPathEvalExpression(
                BAD_CAST "//productPipeline", context);
            if (pipelines != NULL) {
                guint i;
                for (i = 0; i < pipelines->nodesetval->nodeNr; ++i) {
                    xmlNodePtr pipeline = pipelines->nodesetval->nodeTab[i];
                    context->node = pipeline;
                    pipeline_config(result, context, pipeline);
                }
                xmlXPathFreeObject(pipelines);
                success = TRUE;
            }
            else {
                GST_INFO("Failed to find any products in "
                         "configuration '%s'", result->name);
            }
            xmlXPathFreeContext(context);
        }
        else {
            GST_INFO("Configuration root element is missing or is "
                     "not a 'config' element");
        }
        xmlFreeDoc(doc);
    }
    else {
        GST_INFO("Failed to parse configuration");
    }
    if (!success) {
        destroy_config_silently(result);
        result = NULL;
    }
    return result;
}

static void
pipeline_config(wcbe_lagset_sink_cfg_t *cfg, xmlXPathContextPtr context,
                xmlNodePtr pipeline)
{
    xmlChar *pipeline_id = xmlGetProp(pipeline, BAD_CAST "id");
    g_assert(pipeline_id != NULL);
    pipelinesock_t sck;
    sck.addr.sun_family = AF_UNIX;
    gint rc = g_snprintf(sck.addr.sun_path, sizeof(sck.addr.sun_path), "%s/%s",
                         cfg->sockdir, pipeline_id);
    g_assert(rc < sizeof(sck.addr.sun_path));
    g_array_append_val(cfg->socks, sck);
    g_debug("config pipeline %s", sck.addr.sun_path);
    pipelinesock_t *pipelinesock =
        &g_array_index(cfg->socks, pipelinesock_t, cfg->socks->len - 1);

    xmlNodePtr child = pipeline->children;
    while (child != NULL) {
        if (xmlStrEqual(child->name, BAD_CAST "inputs")) {
            xmlChar *inputs_ref = xmlGetProp(child, BAD_CAST "ref");
            if (inputs_ref != NULL) {
                context->node = child;
                xmlXPathObjectPtr product_object =
                    xmlXPathEval(inputs_ref, context);
                if (product_object != NULL) {
                    guint i;
                    for (i = 0; i < product_object->nodesetval->nodeNr; ++i) {
                        xmlNodePtr node =
                            product_object->nodesetval->nodeTab[i];
                        xmlChar *product_id = xmlGetProp(node, BAD_CAST "id");
                        GQuark productq = g_quark_from_string(product_id);
                        g_hash_table_insert(
                            cfg->destinations, GINT_TO_POINTER(productq),
                            pipelinesock);
                        xmlFree(product_id);
                    }
                    xmlXPathFreeObject(product_object);
                }
                else {
                    GST_WARNING("Failed to dereference XPath expression '%s'",
                                inputs_ref);
                }
                xmlFree(inputs_ref);
            }
        }
        child = child->next;
    }

    xmlFree(pipeline_id);
}

static void
deactivate_config(WcbeLagsetSink *sink)
{
    if (sink->config != NULL) {
        forward_event(sink, GST_EVENT_FLUSH_START);
        forward_event(sink, GST_EVENT_FLUSH_STOP);
        g_signal_emit(sink,
                      wcbe_lagset_sink_signals[SIGNAL_DEACTIVATE_CFG],
                      0, sink->config->name);
        if (sink->config->destroy) {
            g_hash_table_remove(sink->configurations, sink->config->name);
        }
        sink->config = NULL;
    }
}

static void
destroy_config(gpointer data)
{
    wcbe_lagset_sink_cfg_t *cfg = data;

    g_return_if_fail(cfg != NULL);
    gchar *name = g_strdup(cfg->name);
    destroy_config_silently(cfg);
    g_signal_emit(cfg->sink, wcbe_lagset_sink_signals[SIGNAL_DESTROY_CFG],
                  0, name, TRUE);
    g_free(name);
}

static void
destroy_config_silently(wcbe_lagset_sink_cfg_t *cfg)
{
    g_return_if_fail(cfg != NULL);

    if (cfg->name != NULL) {
        g_free(cfg->name);
    }
    if (cfg->sockdir != NULL) {
        g_free(cfg->sockdir);
    }
    if (cfg->destinations != NULL) {
        g_hash_table_destroy(cfg->destinations);
    }
    if (cfg->socks != NULL) {
        g_array_free(cfg->socks, TRUE);
    }
    g_free(cfg);
}

static gint
open_sock(WcbeLagsetSink *sink)
{
    gint result = -1;
    sink->sockfd = socket(PF_UNIX, SOCK_DGRAM, 0);
    if (sink->sockfd != -1) {
        struct sockaddr_un addr = {.sun_family = AF_UNIX};
        gchar *path =
            g_build_filename(sink->sockdir, "lagsets", NULL);
        gint rc = g_strlcpy(addr.sun_path, path, sizeof(addr.sun_path));
        if (rc < sizeof(addr.sun_path)) {
            g_assert(rc > 0);
            rc = bind(sink->sockfd, (struct sockaddr *)&addr, sizeof(addr));
            if (rc != -1) {
                rc = fcntl(sink->sockfd, F_SETFL, O_NONBLOCK);
                if (rc == 0) {
                    result = 0;
                }
                else {
                    GST_ERROR("Failed to set socket to non-blocking mode: %s",
                              g_strerror(errno));
                }
            }
            else {
                GST_WARNING("Failed to bind socket to address '%s': %s",
                            addr.sun_path, g_strerror(errno));
            }
        }
        else {
            GST_WARNING("Socket name too long (%d >= %d)",
                        rc, sizeof(addr.sun_path));
        }
        g_free(path);
    }
    return result;
}

static void
close_sock(WcbeLagsetSink *sink)
{
    gint rc;
    if (sink->sockfd != -1) {
        rc = close(sink->sockfd);
        if (rc != 0) {
            GST_WARNING("Failed to close socket: %s", g_strerror(errno));
        }
        sink->sockfd = -1;
    }
    if (sink->sockdir != NULL) {
        gchar *path =
            g_build_filename(sink->sockdir, "lagsets", NULL);
        rc = unlink(path);
        if (rc != 0 && errno != ENOENT) {
            GST_WARNING("Failed to unlink socket directory name '%s': %s",
                        path, g_strerror(errno));
        }
        g_free(path);
    }
}

gboolean
wcbe_lagset_sink_create_config(WcbeLagsetSink *sink, const gchar *config,
                               const gchar *name)
{
    g_return_val_if_fail(sink != NULL, FALSE);
    g_return_val_if_fail(config != NULL, FALSE);
    g_return_val_if_fail(name != NULL, FALSE);

    gboolean result = FALSE;
    if (sink->sockdir != NULL) {
        if (g_hash_table_lookup(sink->configurations, name) == NULL) {
            wcbe_lagset_sink_cfg_t *cfg =
                create_config(sink->sockdir, config, name);
            if (cfg != NULL) {
                cfg->sink = sink;
                g_hash_table_insert(sink->configurations, cfg->name, cfg);
                result = TRUE;
            }
            else {
                GST_INFO("Configuration '%s' failed", name);
            }
        }
        else {
            GST_INFO("Rejected duplicate configuration name '%s'", name);
        }
    }
    else {
        GST_INFO("Configuration '%s' failed: 'sockdir' value is NULL", name);
    }
    g_signal_emit(sink, wcbe_lagset_sink_signals[SIGNAL_CREATE_CFG],
                  0, name, result);
    return result;
}

gboolean
wcbe_lagset_sink_activate_config(WcbeLagsetSink *sink, const gchar *cfg_name)
{
    g_return_val_if_fail(sink != NULL, FALSE);
    g_return_val_if_fail(cfg_name != NULL, FALSE);

    gboolean result;
    wcbe_lagset_sink_cfg_t *config =
        g_hash_table_lookup(sink->configurations, cfg_name);
    if (config != NULL) {
        result = TRUE;
        if (config != sink->config) {
            deactivate_config(sink);
            sink->config = config;
        }
    }
    else {
        result = FALSE;
    }
    g_signal_emit(sink, wcbe_lagset_sink_signals[SIGNAL_ACTIVATE_CFG],
                  0, cfg_name, result);
    return TRUE;
}

void
wcbe_lagset_sink_destroy_config(WcbeLagsetSink *sink, const gchar *cfg_name)
{
    g_return_if_fail(sink != NULL);
    g_return_if_fail(cfg_name != NULL);

    if (sink->config != NULL
        && strcmp(sink->config->name, cfg_name) == 0) {
        sink->config->destroy = TRUE;
    }
    else {
        if (!g_hash_table_remove(sink->configurations, cfg_name)) {
            g_signal_emit(sink,
                          wcbe_lagset_sink_signals[SIGNAL_DESTROY_CFG],
                          0, cfg_name, FALSE);
        }
    }
}
