#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/un.h>
#include <sys/epoll.h>
#include <errno.h>
#include <unistd.h>
#include <fcntl.h>

#include <wcbe.h>
#include <wcbe_lagset_src.h>

// The control socket stuff initially was copied from GStreamer's
// 'gstfdsrc' module.
//
//  * Copyright (C) 1999,2000 Erik Walthinsen <omega@cse.ogi.edu>
//  *                    2000 Wim Taymans <wtay@chello.be>
//  *                    2005 Philippe Khalaf <burger@speedy.org>
//
// Licensed under the LGPL v2.

/* the select call is also performed on the control sockets, that way
 * we can send special commands to unblock the select call */
#define CONTROL_STOP            'S'     /* stop the select call */
#define CONTROL_SOCKETS(src)   src->control_sock
#define WRITE_SOCKET(src)      src->control_sock[1]
#define READ_SOCKET(src)       src->control_sock[0]

#define SEND_COMMAND(src, command)          \
G_STMT_START {                              \
  unsigned char c; c = command;             \
  write (WRITE_SOCKET(src), &c, 1);         \
} G_STMT_END

#define READ_COMMAND(src, command, res)        \
G_STMT_START {                                 \
  res = read(READ_SOCKET(src), &command, 1);   \
} G_STMT_END

GST_DEBUG_CATEGORY_STATIC(wcbe_lagset_src_debug);
#define GST_CAT_DEFAULT wcbe_lagset_src_debug

/* Filter signals and args */
enum
{
    /* FILL ME */
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_INPUT,
    ARG_NUM_LAGSETS
};

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_LAGSET_MIME)
    );

GST_BOILERPLATE(WcbeLagsetSrc, wcbe_lagset_src, GstPushSrc,
                GST_TYPE_PUSH_SRC);

static void wcbe_lagset_src_finalize(GObject *object);
static void wcbe_lagset_src_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void wcbe_lagset_src_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static gboolean wcbe_lagset_src_start(GstBaseSrc *src);
static gboolean wcbe_lagset_src_stop(GstBaseSrc *src);
static GstFlowReturn wcbe_lagset_src_create(GstPushSrc *src, GstBuffer **buf);
static gboolean wcbe_lagset_src_unlock(GstBaseSrc *src);
#ifdef BASE_SRC_HAS_UNLOCK_STOP
static gboolean wcbe_lagset_src_unlock_stop(GstBaseSrc *src);
#endif

static gssize handle_msg_on(gint fd, GstPushSrc *pushsrc, GstBuffer **outbuf);
static gint open_sock(const gchar *path);

static void
wcbe_lagset_src_base_init(gpointer gclass)
{
    static GstElementDetails element_details = {
        "WIDAR lagset source",
        "Source/WIDAR",
        "WIDAR lagset source",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_lagset_src_class_init(WcbeLagsetSrcClass * klass)
{
    GObjectClass *gobject_class = G_OBJECT_CLASS(klass);
    GstBaseSrcClass *gstbasesrc_class = GST_BASE_SRC_CLASS(klass);
    GstPushSrcClass *gstpushsrc_class = GST_PUSH_SRC_CLASS(klass);

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_src_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_lagset_src_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_lagset_src_finalize);

    g_object_class_install_property(
        gobject_class, ARG_INPUT,
        g_param_spec_string("input", "Lag set socket",
                            "Lag set source socket",
                            NULL, G_PARAM_READWRITE));

    g_object_class_install_property(
        gobject_class, ARG_NUM_LAGSETS,
        g_param_spec_uint64("num-lagsets",
                            "Number of lag sets sent",
                            "Get number of lag sets sent",
                            0, G_MAXUINT64,0, G_PARAM_READABLE));

    gstbasesrc_class->start = GST_DEBUG_FUNCPTR(wcbe_lagset_src_start);
    gstbasesrc_class->stop = GST_DEBUG_FUNCPTR(wcbe_lagset_src_stop);
    gstbasesrc_class->unlock = GST_DEBUG_FUNCPTR(wcbe_lagset_src_unlock);
#ifdef BASE_SRC_HAS_UNLOCK_STOP
    gstbasesrc_class->unlock_stop =
        GST_DEBUG_FUNCPTR(wcbe_lagset_src_unlock_stop);
#endif

    gstpushsrc_class->create = GST_DEBUG_FUNCPTR(wcbe_lagset_src_create);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_lagset_src_init(WcbeLagsetSrc *src, WcbeLagsetSrcClass *gclass)
{
    src->sockname = NULL;
    src->sockfd = -1;
    READ_SOCKET(src) = -1;
    WRITE_SOCKET(src) = -1;
    src->epollfd = -1;
    src->caps = NULL;
    src->num_lagsets = 0;
    gst_base_src_set_live(GST_BASE_SRC(src), TRUE);
}

static void
wcbe_lagset_src_finalize(GObject *object)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(object);

    if (src->sockname != NULL) {
        g_free(src->sockname);
    }
    if (src->sockfd != -1) {
        close(src->sockfd);
    }
    if (src->epollfd != -1) {
        close(src->epollfd);
    }
    if (READ_SOCKET(src) > 0) {
        close(READ_SOCKET(src));
    }
    if (WRITE_SOCKET(src) > 0) {
        close(WRITE_SOCKET(src));
    }
    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
wcbe_lagset_src_set_property(GObject *object, guint prop_id,
                             const GValue *value, GParamSpec *pspec)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(object);

    switch (prop_id) {
      case ARG_INPUT:
        src->sockname = g_strdup(g_value_get_string(value));
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_lagset_src_get_property(GObject *object, guint prop_id,
                             GValue *value, GParamSpec *pspec)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(object);

    switch (prop_id) {
      case ARG_INPUT:
        g_value_set_string(value, src->sockname);
        break;
      case ARG_NUM_LAGSETS:
#if SIZEOF_UINT64_T == SIZEOF_UNSIGNED
        g_value_set_uint64(value,
                           (guint64)g_atomic_int_get(&(src->num_lagsets)));
#else
        // The following is exposed to a race condition w.r.t. the
        // pipeline thread updating the num_frames value --- we'll
        // live with it, a mutex lock seems like overkill.
        g_value_set_uint64(value, src->num_lagsets);
#endif
        break;
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/* GstBaseSrc vmethod implementations */

static gboolean
wcbe_lagset_src_start(GstBaseSrc *gstbasesrc)
{
    gboolean result = TRUE;
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(gstbasesrc);
    src->sockfd = open_sock(src->sockname);
    if (src->sockfd != -1) {
        src->epollfd = epoll_create(2);
        if (src->epollfd != -1) {
            struct epoll_event event = {.events = EPOLLIN};
            // Open the control socket pair
            gint control_sock[2];
            gint rc = socketpair(PF_UNIX, SOCK_STREAM, 0, control_sock);
            if (rc == 0) {
                READ_SOCKET(src) = control_sock[0];
                WRITE_SOCKET(src) = control_sock[1];
                rc = fcntl(READ_SOCKET(src), F_SETFL, O_NONBLOCK);
                if (rc != -1) {
                    rc = fcntl(WRITE_SOCKET(src), F_SETFL, O_NONBLOCK);
                }
                if (rc != -1) {
                    event.data.fd = READ_SOCKET(src);
                    rc = epoll_ctl(src->epollfd, EPOLL_CTL_ADD,
                                   READ_SOCKET(src), &event);
                    if (rc != 0) {
                        result = FALSE;
                        GST_ERROR("Failed to add control socket to epoll "
                                  "descriptor: %s", g_strerror(errno));
                    }
                }
                else {
                    GST_ERROR("Failed to set control socket to non-blocking "
                              "mode: %s", g_strerror(errno));
                }
            }
            else {
                result = FALSE;
                GST_ERROR("Failed to create control socket pair: %s",
                          g_strerror(errno));
            }
            if (result) {
                event.data.fd = src->sockfd;
                rc = epoll_ctl(src->epollfd, EPOLL_CTL_ADD, src->sockfd,
                               &event);
                if (rc != 0) {
                    result = FALSE;
                    GST_ERROR("Failed to add lagset source socket to epoll "
                              "descriptor: %s", g_strerror(errno));
                }
            }
        }
        else {
            result = FALSE;
            GST_ERROR("Failed to create epoll descriptor: %s",
                      g_strerror(errno));
        }
    }
    else {
        result = FALSE;
        GST_WARNING("Failed to open input socket '%s': %s", src->sockname,
                    g_strerror(errno));
    }

    if (result) {
        src->num_lagsets = 0;
        src->caps = gst_pad_get_caps(GST_BASE_SRC_PAD(gstbasesrc));
    }
    GST_INFO("wcbe_lagset_src_start result: %d", result);
    return result;
}

static gboolean
wcbe_lagset_src_stop(GstBaseSrc *gstbasesrc)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(gstbasesrc);
    gint rc;

    if (src->epollfd > 0) {
        rc = close(src->epollfd);
        if (rc != 0) {
            GST_WARNING("Failed to close epoll file descriptor: %s",
                        g_strerror(errno));
        }
        src->epollfd = -1;
    }

    if (src->sockfd > 0) {
        rc = close(src->sockfd);
        if (rc != 0) {
            GST_WARNING("Failed to close lagset source socket: %s",
                        g_strerror(errno));
        }
        rc = unlink(src->sockname);
        if (rc != 0) {
            GST_WARNING("Failed to unlink lagset socket directory entry: %s",
                        g_strerror(errno));
        }
    }
    src->sockfd = -1;

    if (READ_SOCKET(src) > 0) {
        rc = close(READ_SOCKET(src));
        if (rc != 0) {
            GST_WARNING("Failed to close control (read) socket: %s",
                        g_strerror(errno));
        }
    }
    READ_SOCKET(src) = -1;

    if (WRITE_SOCKET(src) > 0) {
        rc = close(WRITE_SOCKET(src));
        if (rc != 0) {
            GST_WARNING("Failed to close control (write) socket: %s",
                        g_strerror(errno));
        }
    }
    WRITE_SOCKET(src) = -1;

    if (src->caps != NULL) {
        gst_caps_unref(src->caps);
        src->caps = NULL;
    }

    GST_INFO("Pushed %"G_GUINT64_FORMAT" lagsets to source pad",
             src->num_lagsets);
    return TRUE;
}

static gboolean
wcbe_lagset_src_unlock(GstBaseSrc *basesrc)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(basesrc);
    SEND_COMMAND(src, CONTROL_STOP);
    return TRUE;
}

#ifdef BASE_SRC_HAS_UNLOCK_STOP
static gboolean
wcbe_lagset_src_unlock_stop(GstBaseSrc *basesrc)
{
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(basesrc);

    /* read all stop commands */
    int res = 0;
    while (res >= 0) {
        gchar command;
        READ_COMMAND(src, command, res);
    }
    return TRUE;
}
#endif

static GstFlowReturn
wcbe_lagset_src_create(GstPushSrc *pushsrc, GstBuffer **outbuf)
{
    // Use GST_FLOW_RESEND value to indicate internally that no data
    // buffer has yet been created.
    GstFlowReturn result = GST_FLOW_RESEND;
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(pushsrc);
    static struct epoll_event ev[2];
    static guint ev_idx = 0;
    static guint n_ev = 0;

    errno = EAGAIN;
    while (result == GST_FLOW_RESEND
           && (errno == EAGAIN || errno == EWOULDBLOCK || n_ev == 0)) {
        if (n_ev == 0) {
            n_ev = epoll_wait(src->epollfd, ev, G_N_ELEMENTS(ev), -1);
            if (n_ev >= 0) {
                ev_idx = 0;
            }
            else {
                result = GST_FLOW_ERROR;
                GST_ERROR("Failed epoll_wait: %s", g_strerror(errno));
            }
        }
        if (n_ev > 0) {
            if (G_LIKELY(ev[ev_idx].data.fd != READ_SOCKET(src))) {
                gssize rc = handle_msg_on(ev[ev_idx].data.fd, pushsrc, outbuf);
                if (rc > 0) {
                    ev_idx++;
                    if (*outbuf != NULL) {
                        result = GST_FLOW_OK;
                    }
                }
                else if (rc == 0) {
                    if (ev_idx != G_N_ELEMENTS(ev) - 1) {
                        memcpy(&ev[ev_idx], &ev[ev_idx - 1],
                               sizeof(struct epoll_event));
                    }
                    n_ev--;
                }
                else {
                    GST_WARNING("Failed to receive message on socket: %s",
                                g_strerror(errno));
                    result = GST_FLOW_ERROR;
                }
                if (ev_idx >= n_ev) {
                    ev_idx = 0;
                }
            }
            else {
                GST_DEBUG("stopped");
                result = GST_FLOW_WRONG_STATE;
            }
        }
    }
    return result;
}

static gint
open_sock(const gchar *path)
{
    gint result = -1;
    gint sockfd = socket(PF_UNIX, SOCK_DGRAM, 0);
    if (sockfd != -1) {
        struct sockaddr_un addr = {.sun_family = AF_UNIX};
        gint rc = g_strlcpy(addr.sun_path, path, sizeof(addr.sun_path));
        if (rc < sizeof(addr.sun_path)) {
            g_assert(rc > 0);
            rc = bind(sockfd, (struct sockaddr *)&addr, sizeof(addr));
            if (rc != -1) {
                rc = fcntl(sockfd, F_SETFL, O_NONBLOCK);
                if (rc != -1) {
                    result = sockfd;
                }
                else {
                    GST_ERROR("Failed to set socket to non-blocking "
                              "read mode: %s", g_strerror(errno));
                }
            }
            else {
                GST_WARNING("Failed to bind Unix domain socket: %s",
                            g_strerror(errno));
            }
        }
        else {
            GST_WARNING("Unix domain socket path name too long");
        }
        if (result == -1) {
            close(sockfd);
        }
    }
    else {
        GST_ERROR("Failed to open Unix domain socket: %s",
                  g_strerror(errno));
    }
    return result;
}

static gssize
handle_msg_on(gint fd, GstPushSrc *pushsrc, GstBuffer **outbuf)
{
    gssize result = -1;
    WcbeLagsetSrc *src = WCBE_LAGSET_SRC(pushsrc);
    gchar hdr[MAX(sizeof(wcbe_stream_event_t), sizeof(wcbe_stream_lagset_t))];
    struct iovec iovec[2] = {{.iov_base = hdr, .iov_len = sizeof(hdr)}};
    struct msghdr msghdr = {
        .msg_name = NULL, .msg_namelen = 0,
        .msg_control = NULL, .msg_controllen = 0,
        .msg_iov = iovec, .msg_iovlen = 1};
    errno = 0;
    result = recvmsg(fd, &msghdr, MSG_PEEK | MSG_TRUNC);
    if (G_LIKELY(result > 0)) {
        wcbe_stream_object_t *stream_object = (wcbe_stream_object_t *)hdr;
        switch (stream_object->type) {
          case WCBE_STREAM_LAGSET: {
            wcbe_stream_lagset_t *stream_lagset =
                (wcbe_stream_lagset_t *)stream_object;
            lagset_t *lagset = &(stream_lagset->lagset);
            gsize lagset_size =
                SIZEOF_LAGSET_T(lagset->num_frames, lagset->num_lags,
                                lagset->num_bins, lagset->nbits);
            GstFlowReturn ret = gst_pad_alloc_buffer(
                GST_BASE_SRC_PAD(GST_BASE_SRC(pushsrc)), 0, lagset_size,
                src->caps, outbuf);
            if (G_LIKELY(ret == GST_FLOW_OK)) {
                msghdr.msg_iovlen = G_N_ELEMENTS(iovec);
                lagset = (lagset_t *)GST_BUFFER_DATA(*outbuf);
                wcbe_stream_lagset_init_iovec(
                    stream_lagset, lagset, lagset_size, iovec);
                gssize expected_size =
                    offsetof(wcbe_stream_lagset_t, lagset) + lagset_size;
                errno = 0;
                result = recvmsg(fd, &msghdr, 0);
                if (G_LIKELY(result == expected_size)) {
                    GST_BUFFER_SIZE(*outbuf) = lagset_size;
                    GST_BUFFER_TIMESTAMP(*outbuf) = GST_CLOCK_TIME_NONE;
                    GST_BUFFER_OFFSET(*outbuf) = src->num_lagsets;
                    // Rewrite lagset 'id' quark, since it was
                    // generated in another process.
                    lagset->id =
                        g_quark_from_string(stream_lagset->product_id);
#if SIZEOF_UINT64_T == SIZEOF_UNSIGNED
                    g_atomic_int_inc(&(src->num_lagsets));
#else
                    src->num_lagsets++;
#endif
                }
                else {
                    GST_DEBUG("Failed to read lagset: %s", g_strerror(errno));
                    result = -1;
                    gst_buffer_unref(*outbuf);
                    *outbuf = NULL;
                }
            }
            else {
                GST_DEBUG("Failed to allocate buffer on source pad");
            }
            break;
          }
          case WCBE_STREAM_EVENT: {
            errno = 0;
            result = recvmsg(fd, &msghdr, 0);
            if (G_LIKELY(result == sizeof(wcbe_stream_event_t))) {
                wcbe_stream_event_t *stream_event = (wcbe_stream_event_t *)hdr;
                GstEvent *event = NULL;
                switch (stream_event->event) {
                  case GST_EVENT_EOS:
                    event = gst_event_new_eos();
                    break;
                  case GST_EVENT_FLUSH_START:
                    event = gst_event_new_flush_start();
                    break;
                  case GST_EVENT_FLUSH_STOP:
                    event = gst_event_new_flush_start();
                    break;
                  default:
                    GST_DEBUG("Unexpected event on input: %d",
                              (gint)stream_event->event);
                    break;
                }
                if (G_LIKELY(event != NULL)) {
                    gst_pad_push_event(GST_BASE_SRC_PAD(GST_BASE_SRC(pushsrc)),
                                       event);
                    result = 0;
                }
            }
            else {
                result = -1;
            }
            break;
          }
          default:
            GST_DEBUG("Received unexpected object type: %d",
                      stream_object->type);
            break;
        }
    }
    else if (result == 0 || errno == EAGAIN || errno == EWOULDBLOCK) {
        result = 0;
    }
    else {
        result = -1;
    }
    return result;
}
