#ifdef HAVE_CONFIG_H
#  include <config.h>
#endif

#include <gst/gst.h>
#include <libxml/parser.h>
#include <libxml/tree.h>
#include <libxml/xpath.h>

#include <wcbe_sorter.h>
#include <wcbe.h>
#include <user_marshal.h>

#define MAX_TIMEOUTS 3
// The following hard codes the WIDAR clock rate, but I want to avoid
// floating point arithmetic, and integer arithmetic overflow.
#if WIDAR_CLOCK_RATE == 128000000
#  define WIDAR_TIME_TO_NS(t) (((t) * 1000) / 128)
#else
# error "Unsupported WIDAR_CLOCK_RATE value; please change WIDAR_TIME_TO_NS macro"
#endif

// define the following macros to facilitate adding debug statements
// to mutex operations
#define MUTEX_LOCK(mtx) do { \
    g_mutex_lock(mtx); \
} while (0)
#define MUTEX_UNLOCK(mtx) do { \
    g_mutex_unlock(mtx); \
} while (0)

// Hash keys depend on ftype, ccc, bbid[xy], sbid[xy], sid[xy],
// cchip_id, recirc_blk_y, and board_id only; thus, we apply a mask to
// the keys to retain only these fields.
static const lf_data_frame_word_t lta_hash_mask[] = {
    0x00001FFF, 0xFFFFFFFF, 0x0000FF00, 0xFFFF0000
};
static const lf_data_frame_word_t cc_hash_mask[] = {
    0x000001FF, 0xFFFFFFFF, 0x0000FF00, 0x00000000
};

typedef struct output_queue {
    GQuark product_id;
    GQueue *queue;
    GMutex *qmutex;
    GTimer *timer;
    gdouble timeout;
    gint timeouts_remaining;
    guint64 hw_dump;
    guint num_frames; // frames per lagset
    guint num_lags;   // lags per lagset
    guint num_bins;   // bins (multiple of lagsets)
    guint nbits;      // number of correlation bits
} output_queue_t;

typedef struct queue_element {
    GstBuffer *buffer;
    guint unfilled_frames;
    guint refcount;
} queue_element_t;

typedef struct frame_dest {
    output_queue_t *oqueue;
    guint header_idx;
    guint lags_idx;
} frame_dest_t;

typedef struct lc_segment_t {
    guint subprod;
    guint cchip[2];     // x, y
    guint ccc[2];       // first, last
    guint rack;
    guint crate;
    guint slot;
    guint inp_block[2]; // x, y
    guint recirc;
    guint baseband[2];  // x, y
    guint subband[2];   // x, y
    guint station[2];   // x, y
} lc_segment_t;

typedef struct stream_t {
    guint baseband;
    guint subband;
    guint station;
} stream_t;

GST_DEBUG_CATEGORY_STATIC(wcbe_sorter_debug);
#define GST_CAT_DEFAULT wcbe_sorter_debug

/* Filter signals and args */
enum
{
    SIGNAL_CREATE_CFG,
    SIGNAL_ACTIVATE_CFG,
    SIGNAL_DEACTIVATE_CFG,
    SIGNAL_DESTROY_CFG,
    LAST_SIGNAL
};

enum
{
    ARG_0,
    ARG_CONFIGS,
    ARG_ACTIVE_CONFIG
};


static GstStaticPadTemplate sink_factory = GST_STATIC_PAD_TEMPLATE(
    "sink",
    GST_PAD_SINK,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_MIME)
    );

static GstStaticPadTemplate src_factory = GST_STATIC_PAD_TEMPLATE(
    "src",
    GST_PAD_SRC,
    GST_PAD_ALWAYS,
    GST_STATIC_CAPS(WCBE_LAGSET_MIME)
    );

static guint wcbe_sorter_signals[LAST_SIGNAL] = {0};

GST_BOILERPLATE(WcbeSorter, wcbe_sorter, GstElement, GST_TYPE_ELEMENT);

static void wcbe_sorter_set_property(
    GObject *object, guint prop_id, const GValue *value, GParamSpec *pspec);
static void wcbe_sorter_get_property(
    GObject *object, guint prop_id, GValue *value, GParamSpec *pspec);
static void wcbe_sorter_finalize(GObject *object);
static GstFlowReturn wcbe_sorter_chain(GstPad *pad, GstBuffer *buf);
static gboolean wcbe_sorter_event(GstPad *pad, GstEvent *event);
static GstStateChangeReturn wcbe_sorter_change_state(
    GstElement *element, GstStateChange transition);

static gint cmp_timestamps(const lf_widar_time_t *ts0,
                           const lf_widar_time_t *ts1, guint64 block);
static GstFlowReturn push_queue_element(WcbeSorter *sorter,
                                        queue_element_t *qe);
static void queue_element_ref(queue_element_t *qe);
static void queue_element_unref(queue_element_t *qe);
static gpointer timeout_fun(gpointer data);
static void timeout_queue(gpointer data, gpointer user_data);
static guint64 set_timeouts(GSList *queues);
static guint frame_hash(gconstpointer key);
static gboolean frame_equal(gconstpointer a, gconstpointer b);
static void deactivate_config(WcbeSorter *sorter, gboolean flush);
static void destroy_config(gpointer data);
static void destroy_config_silently(wcbe_sorter_cfg_t *cfg);
static wcbe_sorter_cfg_t *create_config(const gchar *config,
                                        const gchar *name);
static void product_config(wcbe_sorter_cfg_t *cfg,
                           xmlXPathContextPtr context, xmlNodePtr product);
static gint cmp_segments(gconstpointer a, gconstpointer b);
static gint deref_stream(xmlNodePtr product, xmlXPathContextPtr context,
                         const gchar *name, stream_t *stream);
static void read_uint_prop(xmlNodePtr node, const xmlChar *name, guint *val);
static void read_all_segments(xmlXPathContextPtr context,
                              const stream_t *streamA, const stream_t *streamB,
                              GArray *lc_segments, guint subprod);
static void read_segment(xmlNodePtr segment, const stream_t *streamA,
                         const stream_t *streamB, GArray *lc_segments,
                         guint subprod);
static void flush_all_queues(WcbeSorter *sorter);

static void
wcbe_sorter_base_init(gpointer gclass)
{
    static GstElementDetails element_details = {
        "WIDAR lag set sorter",
        "Sorter/WIDAR",
        "Assemble and temporally sort WIDAR lag sets",
        "Martin Pokorny <mpokorny@nrao.edu>"
    };
    GstElementClass *element_class = GST_ELEMENT_CLASS(gclass);

    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&src_factory));
    gst_element_class_add_pad_template(
        element_class, gst_static_pad_template_get(&sink_factory));
    gst_element_class_set_details(element_class, &element_details);
}

/* initialize the plugin's class */
static void
wcbe_sorter_class_init(WcbeSorterClass *klass)
{
    GObjectClass *gobject_class;
    GstElementClass *gstelement_class;

    gobject_class = (GObjectClass *)klass;
    gstelement_class = (GstElementClass *)klass;

    gobject_class->set_property =
        GST_DEBUG_FUNCPTR(wcbe_sorter_set_property);
    gobject_class->get_property =
        GST_DEBUG_FUNCPTR(wcbe_sorter_get_property);
    gobject_class->finalize =
        GST_DEBUG_FUNCPTR(wcbe_sorter_finalize);

    gstelement_class->change_state =
        GST_DEBUG_FUNCPTR(wcbe_sorter_change_state);

    GParamSpec *config_spec =
        g_param_spec_string("config", "Configuration name",
                            "Configuration name", NULL,
                            G_PARAM_READABLE);
    g_object_class_install_property(
        gobject_class, ARG_CONFIGS,
        g_param_spec_value_array("configs", "Available configurations",
                                 "Get list of all available configurations",
                                 config_spec, G_PARAM_READABLE));
    g_object_class_install_property(
        gobject_class, ARG_ACTIVE_CONFIG,
        g_param_spec_string("active-config", "Active configuration",
                            "Get name of active configuration",
                            NULL, G_PARAM_READABLE));

    wcbe_sorter_signals[SIGNAL_CREATE_CFG] =
        g_signal_new("config-created", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    wcbe_sorter_signals[SIGNAL_ACTIVATE_CFG] =
        g_signal_new("config-activated", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);

    wcbe_sorter_signals[SIGNAL_DEACTIVATE_CFG] =
        g_signal_new("config-deactivated", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_marshal_VOID__STRING,
                     G_TYPE_NONE, 1, G_TYPE_STRING);

    wcbe_sorter_signals[SIGNAL_DESTROY_CFG] =
        g_signal_new("config-destroyed", G_TYPE_FROM_CLASS(klass),
                     G_SIGNAL_RUN_FIRST,
                     0, NULL, NULL,
                     g_cclosure_user_marshal_VOID__STRING_BOOLEAN,
                     G_TYPE_NONE, 2, G_TYPE_STRING, G_TYPE_BOOLEAN);
}

/* initialize the new element
 * instantiate pads and add them to element
 * set functions
 * initialize structure
 */
static void
wcbe_sorter_init(WcbeSorter *sorter,
                 WcbeSorterClass *gclass)
{
    GstElementClass *klass = GST_ELEMENT_GET_CLASS(sorter);

    sorter->sinkpad = gst_pad_new_from_template(
        gst_element_class_get_pad_template(klass, "sink"), "sink");

    sorter->srcpad = gst_pad_new_from_template(
        gst_element_class_get_pad_template(klass, "src"), "src");
    gst_pad_use_fixed_caps(sorter->srcpad);
    gst_pad_set_caps(sorter->srcpad,
                     gst_static_pad_template_get_caps(&src_factory));

    gst_element_add_pad(GST_ELEMENT(sorter), sorter->sinkpad);
    gst_element_add_pad(GST_ELEMENT(sorter), sorter->srcpad);
    gst_pad_set_chain_function(sorter->sinkpad, wcbe_sorter_chain);
    gst_pad_set_event_function(sorter->sinkpad, wcbe_sorter_event);

    sorter->config = NULL;
    sorter->dummy_buffer = gst_buffer_new();
    sorter->config_mutex = g_mutex_new();
    sorter->config_cond = g_cond_new();
    sorter->new_config = FALSE;
    sorter->configurations =
        g_hash_table_new_full(g_str_hash, g_str_equal, NULL, destroy_config);
    GError *err = NULL;
    sorter->timeout_thread = g_thread_create(timeout_fun, sorter, TRUE, &err);
    if (sorter->timeout_thread == NULL) {
        GST_ERROR("Failed to spawn timeout thread: %s", err->message);
    }
}

static void
wcbe_sorter_finalize(GObject *object)
{
    WcbeSorter *sorter = WCBE_SORTER(object);

    if (sorter->timeout_thread != NULL) {
        GThread *timeout_thread = sorter->timeout_thread;
        g_atomic_pointer_set(&(sorter->timeout_thread), NULL);
        g_thread_join(timeout_thread);
    }
    // now we may ignore locking, since timeout thread is gone
    deactivate_config(sorter, FALSE);
    if (sorter->configurations != NULL) {
        g_hash_table_destroy(sorter->configurations);
    }
    if (sorter->config_mutex != NULL) {
        g_mutex_free(sorter->config_mutex);
    }
    if (sorter->dummy_buffer != NULL) {
        gst_buffer_unref(sorter->dummy_buffer);
    }
    G_OBJECT_CLASS(parent_class)->finalize(object);
}

static void
wcbe_sorter_set_property(GObject *object, guint prop_id,
                         const GValue *value, GParamSpec *pspec)
{
    switch (prop_id) {
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

static void
wcbe_sorter_get_property(GObject *object, guint prop_id,
                         GValue *value, GParamSpec *pspec)
{
    WcbeSorter *sorter = WCBE_SORTER(object);

    switch (prop_id) {
      case ARG_ACTIVE_CONFIG:
        g_value_set_string(value, sorter->config ? sorter->config->name : "");
        break;
      case ARG_CONFIGS: {
        GValueArray *varray =
            g_value_array_new(g_hash_table_size(sorter->configurations));
        GValue str = {0,};
        g_value_init(&str, G_TYPE_STRING);
        void get_name(gpointer key, gpointer value, gpointer user_data) {
            const gchar *name = key;
            const wcbe_sorter_cfg_t *config = value;
            const gchar *fmt;
            if (config->destroy) {
                fmt = "[%s]";
            }
            else if (config == sorter->config) {
                fmt = "<%s>";
            }
            else {
                fmt = "(%s)";
            }
            gchar *tagged_name = g_strdup_printf(fmt, name);
            g_value_set_string(&str, tagged_name);
            g_value_array_append(varray, &str);
            g_free(tagged_name);
        }
        g_hash_table_foreach(sorter->configurations, get_name, NULL);
        g_value_take_boxed(value, varray);
        break;
      }
      default:
        G_OBJECT_WARN_INVALID_PROPERTY_ID(object, prop_id, pspec);
        break;
    }
}

/* GstElement vmethod implementations */

/* chain function
 * this function does the actual processing
 */

static GstFlowReturn
wcbe_sorter_chain(GstPad *pad, GstBuffer *buf)
{
    GstFlowReturn result = GST_FLOW_OK;
    WcbeSorter *sorter = WCBE_SORTER(gst_pad_get_parent_element(pad));

    if (sorter->config == NULL) {
        gst_buffer_unref(buf);
        GST_DEBUG("Sorter element is unconfigured...dropping lag frame");
        return result;
    }

    lf_lag_frame_t *lf = (lf_lag_frame_t *)GST_BUFFER_DATA(buf);
    g_assert(lf != NULL);
    frame_dest_t *dest = NULL;
    if (G_LIKELY(sorter->config->destinations != NULL)) {
        dest = g_hash_table_lookup(sorter->config->destinations, lf);
    }
    if (dest != NULL) {
        output_queue_t *oq = dest->oqueue;
        g_atomic_int_set(&(oq->timeouts_remaining), MAX_TIMEOUTS);
        lf_widar_time_t *earliest_ts = NULL;
        MUTEX_LOCK(oq->qmutex);
        queue_element_t *earliest_qe = g_queue_peek_tail(oq->queue);
        if (earliest_qe != NULL) {
            queue_element_ref(earliest_qe);
            earliest_ts = &(((lagset_t *)GST_BUFFER_DATA(
                                 earliest_qe->buffer))->timestamp);
        }
        // TODO: The timestamp comparisons don't account for phase of
        // the integration intervals; it's not in the config file yet!
        if (earliest_qe == NULL
            || cmp_timestamps(&(lf->timestamp), earliest_ts,
                              oq->hw_dump) >= 0) {
            lagset_t dummy_ls;
            dummy_ls.timestamp.epoch = lf->timestamp.epoch;
            dummy_ls.timestamp.seconds = lf->timestamp.seconds;
            dummy_ls.timestamp.clock_cycles = lf->timestamp.clock_cycles;
            queue_element_t dummy_qe = {.buffer = sorter->dummy_buffer};
            GST_BUFFER_DATA(dummy_qe.buffer) = (guint8 *)&dummy_ls;
            gint cmp(gconstpointer a, gconstpointer b) {
                const queue_element_t *qea = a;
                const queue_element_t *qeb = b;
                const lagset_t *lsa =
                    (const lagset_t *)GST_BUFFER_DATA(qea->buffer);
                const lagset_t *lsb =
                    (const lagset_t *)GST_BUFFER_DATA(qeb->buffer);
                return cmp_timestamps(&(lsa->timestamp), &(lsb->timestamp),
                                      oq->hw_dump);
            }
            GList *ls = g_queue_find_custom(oq->queue, &dummy_qe, cmp);
            queue_element_t *lagset_qe = NULL;
            lagset_t *lagset = NULL;
            if (ls != NULL) {
                lagset_qe = ls->data;
                queue_element_ref(lagset_qe);
                MUTEX_UNLOCK(oq->qmutex);
                g_assert(lagset_qe->buffer != NULL);
                lagset = (lagset_t *)GST_BUFFER_DATA(lagset_qe->buffer);
            }
            else {
                MUTEX_UNLOCK(oq->qmutex);
                lagset_qe = g_slice_new(queue_element_t);
                lagset_qe->unfilled_frames = oq->num_frames * oq->num_bins;
                lagset_qe->refcount = 1;
                result = gst_pad_alloc_buffer(
                    sorter->srcpad, 0,
                    SIZEOF_LAGSET_T(oq->num_frames, oq->num_lags,
                                    oq->num_bins, oq->nbits),
                    GST_PAD_CAPS(sorter->srcpad), &(lagset_qe->buffer));
                if (result == GST_FLOW_OK) {
                    lagset = (lagset_t *)GST_BUFFER_DATA(lagset_qe->buffer);
                    lagset_init(lagset, oq->num_frames, oq->num_lags,
                                oq->num_bins, oq->nbits);
                    lagset->id = oq->product_id;
                    lagset->timestamp.epoch = dummy_ls.timestamp.epoch;
                    lagset->timestamp.seconds = dummy_ls.timestamp.seconds;
                    lagset->timestamp.clock_cycles =
                        dummy_ls.timestamp.clock_cycles;
                    GST_BUFFER_TIMESTAMP(lagset_qe->buffer) =
                        WIDAR_TIME_TO_NS(TIMEVALUE(&(lagset->timestamp)));
                    queue_element_ref(lagset_qe); // on behalf of queue
                    MUTEX_LOCK(oq->qmutex);
                    g_queue_push_head(oq->queue, lagset_qe);
                    MUTEX_UNLOCK(oq->qmutex);
                }
                else {
                    GST_INFO("Failed to alloc buffer on source pad");
                }
            }
            if (lagset != NULL) {
                // The 'mod lagset->num_bins' below allows us to map
                // multiple LTA bins to a single, logical phase bin.
                // Whether this is useful beyond the case in which the
                // number of logical phase bins is one, I'm not sure.
                // Also note that the "bank" bit of the lta_bin value
                // is ignored.
                guint bin;
                if (G_LIKELY(LF_LAG_FRAME_FTYPE(lf)
                             == LF_LTA_DATA_FRAME_FTYPE)) {
                    bin = LF_LTA_FRAME_PHASE_BIN((lf_lta_lag_frame_t *)lf);
                }
                else {
                    bin = LF_CC_FRAME_LTA_BIN((lf_cc_lag_frame_t *)lf);
                }
                bin = bin % lagset->num_bins;
                guint header_idx = dest->header_idx + lagset->num_frames * bin;
                header_t *header = &(lagset->headers[header_idx]);
                if (header->num_lags == 0) {
                    guint lags_idx = dest->lags_idx + lagset->num_lags * bin;
                    lf_lag_t *lags = &(lagset->lags[lags_idx]);
                    memcpy(header->key, lf->key,
                           LF_KEY_LENGTH * sizeof(lf_data_frame_word_t));
                    header->lag_offset = lags_idx;
                    header->num_lags = lf->num_lags;
                    header->status = lf->status;
                    header->timestamp.epoch = lf->timestamp.epoch;
                    header->timestamp.seconds = lf->timestamp.seconds;
                    header->timestamp.clock_cycles =
                        lf->timestamp.clock_cycles;
                    header->data_valid_center = lf->data_valid_center;
                    header->data_valid_edge = lf->data_valid_edge;
                    if (G_LIKELY(HEADER_FTYPE(header)
                                 == LF_LTA_DATA_FRAME_FTYPE)) {
                        header->frame_count =
                            ((lf_lta_lag_frame_t *)lf)->frame_count;
                    }
                    else {
                        header->data_bias =
                            ((lf_cc_lag_frame_t *)lf)->data_bias;
                    }
                    memcpy(lags, lf->lags, lf->num_lags * sizeof(lf_lag_t));
                    lagset_qe->unfilled_frames--;
                }
                else {
                    GST_INFO("Received frame mapped to previously "
                             "filled frame in lagset.");
                }
            }
            else {
                GST_WARNING("Failed to find lagset for received frame");
            }
            if (lagset_qe != NULL) {
                queue_element_unref(lagset_qe);
            }
        }
        else {
            MUTEX_UNLOCK(oq->qmutex);
            GST_WARNING("Frame arrived too late...dropped");
            // TODO: emit a signal?
        }

        if (earliest_qe != NULL) {
            queue_element_unref(earliest_qe);
        }

        // Output lag sets if appropriate
        MUTEX_LOCK(oq->qmutex);
        queue_element_t *qe = g_queue_peek_tail(oq->queue);
        while ((qe != NULL && qe->unfilled_frames == 0)
               || (g_queue_get_length(oq->queue) > 2)) {
            qe = g_queue_pop_tail(oq->queue);
            push_queue_element(sorter, qe);
            qe = g_queue_peek_tail(oq->queue);
        }
        MUTEX_UNLOCK(oq->qmutex);
    }
    else {
        GST_INFO("Dropping frame unspecified by configuration");
    }

    gst_buffer_unref(buf);
    gst_object_unref(GST_OBJECT(sorter));

    return result;
}

static gboolean
wcbe_sorter_event(GstPad *pad, GstEvent *event)
{
    gboolean result;
    WcbeSorter *sorter = WCBE_SORTER(gst_pad_get_parent_element(pad));
    switch (GST_EVENT_TYPE(event)) {
      case GST_EVENT_EOS:
      case GST_EVENT_FLUSH_START:
        // Output all remaining lag sets
        flush_all_queues(sorter);
        result = gst_pad_push_event(sorter->srcpad, event);
        break;

      case WCBE_EVENT_CREATE_CONFIG:
        g_assert(gst_structure_has_name(event->structure, "config-create"));
        g_assert(gst_structure_has_field(event->structure, "name"));
        g_assert(gst_structure_has_field(event->structure, "config"));
        wcbe_sorter_create_config(
            sorter, gst_structure_get_string(event->structure, "config"),
            gst_structure_get_string(event->structure, "name"));
        result = gst_pad_push_event(sorter->srcpad, event);
        break;

      case WCBE_EVENT_ACTIVATE_CONFIG:
        g_assert(gst_structure_has_name(event->structure, "config-activate"));
        g_assert(gst_structure_has_field(event->structure, "name"));
        wcbe_sorter_activate_config(
            sorter, gst_structure_get_string(event->structure, "name"));
        result = gst_pad_push_event(sorter->srcpad, event);
        break;

      case WCBE_EVENT_DESTROY_CONFIG:
        g_assert(gst_structure_has_name(event->structure, "config-destroy"));
        g_assert(gst_structure_has_field(event->structure, "name"));
        wcbe_sorter_destroy_config(
            sorter, gst_structure_get_string(event->structure, "name"));
        result = gst_pad_push_event(sorter->srcpad, event);
        break;

      default:
        result = gst_pad_event_default(pad, event);
        break;
    }
    gst_object_unref(GST_OBJECT(sorter));
    return result;
}


static GstStateChangeReturn
wcbe_sorter_change_state(GstElement *element, GstStateChange transition)
{
    GstStateChangeReturn result = GST_STATE_CHANGE_SUCCESS;

    switch (transition) {
      case GST_STATE_CHANGE_READY_TO_PAUSED:
        result = GST_STATE_CHANGE_NO_PREROLL;
        break;
      case GST_STATE_CHANGE_READY_TO_NULL:
        deactivate_config(WCBE_SORTER(element), FALSE);
        break;
      default:
        break;
    }

    GstStateChangeReturn ret =
        GST_ELEMENT_CLASS(parent_class)->change_state(element, transition);
    if (result == GST_STATE_CHANGE_NO_PREROLL) {
        if (ret == GST_STATE_CHANGE_FAILURE) {
            result = GST_STATE_CHANGE_FAILURE;
        }
    }
    else {
        result = ret;
    }

    return result;
}

static gint
cmp_timestamps(const lf_widar_time_t *ts0, const lf_widar_time_t *ts1,
               guint64 block)
{
    g_return_val_if_fail(ts0 != NULL, 0);
    g_return_val_if_fail(ts1 != NULL, 0);

    gint result;
    if (G_LIKELY(ts0->epoch == ts1->epoch)) {
        if (block == 0) {
            block = 1;
        }
        guint64 tb0 = TIMEVALUE(ts0) / block;
        guint64 tb1 = TIMEVALUE(ts1) / block;
        if (tb0 < tb1) {
            result = -1;
        }
        else if (tb0 > tb1) {
            result = 1;
        }
        else {
            result = 0;
        }
    }
    else {
        result = ts0->epoch - ts1->epoch;
    }
    return result;
}

static GstFlowReturn
push_queue_element(WcbeSorter *sorter, queue_element_t *qe)
{
    g_return_val_if_fail(sorter != NULL, GST_FLOW_ERROR);
    g_return_val_if_fail(qe != NULL, GST_FLOW_ERROR);

    GstFlowReturn result;
    GST_DEBUG("Pushing lagset element with %u unfilled frames",
              qe->unfilled_frames);
    result = gst_pad_push(sorter->srcpad, qe->buffer);
    if (result != GST_FLOW_OK) {
        GST_INFO("Failed to push buffer");
    }
    qe->buffer = NULL; // do this since gst_pad_push steals the ref
    queue_element_unref(qe);
    return GST_FLOW_OK;
}

static void
queue_element_ref(queue_element_t *qe)
{
    g_return_if_fail(qe != NULL);
    g_atomic_int_inc(&(qe->refcount));
}

static void
queue_element_unref(queue_element_t *qe)
{
    g_return_if_fail(qe != NULL);
    if (g_atomic_int_dec_and_test(&(qe->refcount))) {
        if (qe->buffer != NULL) {
            gst_buffer_unref(qe->buffer);
        }
        g_slice_free(queue_element_t, qe);
    }
}

static gpointer
timeout_fun(gpointer data)
{
    WcbeSorter *sorter = data;
    gulong timeout = 0;

    while (g_atomic_pointer_get(&(sorter->timeout_thread)) != NULL) {
        if (timeout > 0) {
            g_usleep(timeout);
            MUTEX_LOCK(sorter->config_mutex);
            if (sorter->config != NULL) {
                g_slist_foreach(sorter->config->queues, timeout_queue, sorter);
            }
        }
        else {
            MUTEX_LOCK(sorter->config_mutex);
        }
        if (timeout == 0) {
            while (!sorter->new_config
                   && (g_atomic_pointer_get(&(sorter->timeout_thread))
                       != NULL)) {
                g_cond_wait(sorter->config_cond, sorter->config_mutex);
            }
        }
        if (sorter->new_config) {
            sorter->new_config = FALSE;
            guint64 dump;
            if (sorter->config != NULL) {
                dump = set_timeouts(sorter->config->queues);
            }
            else {
                dump = 0;
            }
            timeout = dump / (WIDAR_CLOCK_RATE / G_USEC_PER_SEC);
            if (timeout != 0) {
                timeout = CLAMP(timeout, 50000, 500000);
            }
            else {
                GST_WARNING("Unable to set non-zero timeout value "
                            "in dump event timeout generator");
            }
        }
        MUTEX_UNLOCK(sorter->config_mutex);
    }
    return NULL;
}

static void
timeout_queue(gpointer data, gpointer user_data)
{
    output_queue_t *oq = data;
    WcbeSorter *sorter = user_data;

    if (oq->timeout > 0 && g_timer_elapsed(oq->timer, NULL) >= oq->timeout) {
        g_timer_start(oq->timer);
        if (g_atomic_int_dec_and_test(&(oq->timeouts_remaining))) {
            MUTEX_LOCK(oq->qmutex);
            // Check "timeouts_remaining" once more in case something
            // entered the queue between our first test and mutex
            // acquisition.
            if (g_atomic_int_get(&(oq->timeouts_remaining)) == 0) {
                queue_element_t *qe = g_queue_pop_tail(oq->queue);
                MUTEX_UNLOCK(oq->qmutex);
                if (qe != NULL) {
                    GST_DEBUG("Pushing timed out lagset element with %u "
                              "unfilled frames", qe->unfilled_frames);
                    gst_pad_push(sorter->srcpad, qe->buffer);
                    qe->buffer = NULL;
                    queue_element_unref(qe);
                }
                g_atomic_int_compare_and_exchange(
                    &(oq->timeouts_remaining), 0, 1);
            }
            else {
                MUTEX_UNLOCK(oq->qmutex);
            }
        }
    }
}

/* set_timeouts
 *
 * Return the minimum of the hw_dump values of all the queues. It
 * might make some sense to return the GCD of the hw_dump values to
 * minimize the rate at which the timeout_fun loop runs; however,
 * since that loop imposes both upper and lower limits on its rate, I
 * instead opted for this simpler solution.
 */
static guint64
set_timeouts(GSList *queues)
{
    guint64 min = G_MAXUINT64;
    gboolean have_zero = FALSE;
    void fn(gpointer data, gpointer reserved) {
        output_queue_t *oq = data;
        // Ignore zero hw_dump values for now, but note their
        // presence.
        if (oq->hw_dump > 0) {
            if (oq->hw_dump < min) {
                min = oq->hw_dump;
            }
        }
        else {
            have_zero = TRUE;
        }
        oq->timeout = oq->hw_dump * (WIDAR_CLOCK_RATE / G_USEC_PER_SEC);
        g_atomic_int_set(&(oq->timeouts_remaining), MAX_TIMEOUTS);
        g_timer_start(oq->timer);
    }
    g_slist_foreach(queues, fn, NULL);
    // If all hw_dump values are zero, min will have its initial
    // value, so zero is the correct operational min value.
    if (min == G_MAXUINT64 && have_zero) {
        min = 0;
    }
    return min;
}

static guint
frame_hash(gconstpointer key)
{
    const lf_lag_frame_t *lf = key;
    const lf_data_frame_word_t *mask;

    if (G_LIKELY(LF_LAG_FRAME_FTYPE(lf) == LF_LTA_DATA_FRAME_FTYPE)) {
        mask = lta_hash_mask;
    }
    else {
        mask = cc_hash_mask;
    }
    return ((lf->key[0] & mask[0]) ^ (lf->key[1] & mask[1])
            ^ (lf->key[2] & mask[2]) ^ (lf->key[3] & mask[3]));
}

static gboolean
frame_equal(gconstpointer a, gconstpointer b)
{
    const lf_lag_frame_t *lfa = a;
    const lf_lag_frame_t *lfb = b;
    gboolean result;
    if (G_LIKELY(LF_LAG_FRAME_FTYPE(lfa) == LF_LAG_FRAME_FTYPE(lfb))) {
        const lf_data_frame_word_t *mask;
        if (G_LIKELY(LF_LAG_FRAME_FTYPE(lfa) == LF_LTA_DATA_FRAME_FTYPE)) {
            mask = lta_hash_mask;
        }
        else {
            mask = cc_hash_mask;
        }
        result = (((lfa->key[0] & mask[0]) == (lfb->key[0] & mask[0]))
                  && ((lfa->key[1] & mask[1]) == (lfb->key[1] & mask[1]))
                  && ((lfa->key[2] & mask[2]) == (lfb->key[2] & mask[2]))
                  && ((lfa->key[3] & mask[3]) == (lfb->key[3] & mask[3])));
    }
    else {
        result = FALSE;
    }
    return result;
}

static void
deactivate_config(WcbeSorter *sorter, gboolean flush)
{
    if (sorter->config != NULL) {
        if (flush) {
            flush_all_queues(sorter);
        }
        g_signal_emit(sorter, wcbe_sorter_signals[SIGNAL_DEACTIVATE_CFG],
                      0, sorter->config->name);
        if (sorter->config->destroy) {
            g_hash_table_remove(sorter->configurations, sorter->config->name);
        }
        sorter->config = NULL;
    }
}

static void
destroy_config(gpointer data)
{
    wcbe_sorter_cfg_t *cfg = data;
    gchar *name = g_strdup(cfg->name);
    destroy_config_silently(cfg);
    g_signal_emit(cfg->sorter, wcbe_sorter_signals[SIGNAL_DESTROY_CFG],
                  0, name, TRUE);
    g_free(name);
}

static void
destroy_config_silently(wcbe_sorter_cfg_t *cfg)
{
    g_free(cfg->name);
    while (cfg->queues != NULL) {
        void free_elt(gpointer data, gpointer user_data) {
            queue_element_t *qe = data;
            queue_element_unref(qe);
        }
        output_queue_t *oq = cfg->queues->data;
        cfg->queues = g_slist_remove(cfg->queues, oq);
        g_queue_foreach(oq->queue, free_elt, NULL);
        g_queue_free(oq->queue);
        g_mutex_free(oq->qmutex);
        g_timer_destroy(oq->timer);
        g_free(oq);
    }
    if (cfg->destinations != NULL) {
        g_hash_table_destroy(cfg->destinations);
    }
    g_free(cfg);
}

static wcbe_sorter_cfg_t *
create_config(const gchar *config, const gchar *name)
{
    gboolean success = FALSE;
    wcbe_sorter_cfg_t *result = g_new0(wcbe_sorter_cfg_t, 1);
    result->destroy = FALSE;
    xmlDocPtr doc = xmlParseMemory(config, strlen(config));
    if (doc != NULL) {
        xmlNodePtr root = xmlDocGetRootElement(doc);
        if (root != NULL
            && xmlStrEqual(root->name, BAD_CAST "config")) {
            result->name = g_strdup(name);
            xmlXPathInit();
            xmlXPathContextPtr context = xmlXPathNewContext(doc);
            g_assert(context != NULL);
            context->node = root;
            xmlXPathObjectPtr products = xmlXPathEvalExpression(
                BAD_CAST "//product4|//product7", context);
            if (products != NULL) {
                // Note: don't need a do anything when removing hash
                // table values except to free the frame_dest_t
                // memory; freeing the output_queue_t element within
                // the frame_dest_t is handled elsewhere (config
                // 'queues').
                result->destinations =
                    g_hash_table_new_full(frame_hash, frame_equal,
                                          g_free, g_free);
                guint i;
                for (i = 0; i < products->nodesetval->nodeNr; ++i) {
                    xmlNodePtr product = products->nodesetval->nodeTab[i];
                    context->node = product;
                    product_config(result, context, product);
                }
                xmlXPathFreeObject(products);
                success = TRUE;
            }
            else {
                GST_INFO("Failed to find any products in "
                         "configuration '%s'", result->name);
            }
            xmlXPathFreeContext(context);
        }
        else {
            GST_INFO("Configuration root element is missing or is "
                     "not a 'config' element");
        }
        xmlFreeDoc(doc);
    }
    else {
        GST_INFO("Failed to parse configuration");
    }
    if (!success) {
        destroy_config_silently(result);
        result = NULL;
    }
    return result;
}

static void
product_config(wcbe_sorter_cfg_t *cfg, xmlXPathContextPtr context,
               xmlNodePtr product)
{
    xmlChar *product_id = xmlGetProp(product, BAD_CAST "id");
    g_assert(product_id != NULL);
    output_queue_t *oq = g_new(output_queue_t, 1);
    oq->queue = g_queue_new();
    oq->qmutex = g_mutex_new();
    oq->timer = g_timer_new();
    cfg->queues = g_slist_prepend(cfg->queues, oq);
    oq->product_id = g_quark_from_string(product_id);

    read_uint_prop(product, BAD_CAST "bins", &(oq->num_bins));
    read_uint_prop(product, BAD_CAST "lags", &(oq->num_lags));
    guint lags_per_frame;
    if (oq->num_lags != MAX_LAGS_PER_FRAME / 2) {
        lags_per_frame = MAX_LAGS_PER_FRAME;
    }
    else {
        lags_per_frame = MAX_LAGS_PER_FRAME / 2;
    }
    oq->num_frames = oq->num_lags / lags_per_frame;

    // resolve stream XPath expressions
    stream_t streamA, streamB;
    deref_stream(product, context, "streamAref", &streamA);
    deref_stream(product, context, "streamBref", &streamB);

    // find dump time of frames from correlator for this product
    guint64 hw_dump = 0;
    xmlXPathObjectPtr intgns = xmlXPathEval(
        BAD_CAST "corrIntegration", context);
    if (intgns != NULL) {
        if (intgns->nodesetval->nodeNr == 1) {
            xmlNodePtr intgn = intgns->nodesetval->nodeTab[0];
            guint hw = 0, lta = 0;
            read_uint_prop(intgn, BAD_CAST "hw", &hw);
            read_uint_prop(intgn, BAD_CAST "lta", &lta);
            hw_dump = hw * lta * (WIDAR_CLOCK_RATE / G_USEC_PER_SEC);
            if (hw == 0 || lta == 0) {
                GST_WARNING("Product '%s' 'corrIntegration' element is "
                            "missing 'hw' and/or 'lta' attributes..."
                            "dump timeout events will not be generated "
                            "for this product.");
            }
        }
        else {
            GST_WARNING("Product '%s' has multiple "
                        "'corrIntegration' elements...dump timeout events "
                        "will not be generated for this product.");
        }
        xmlXPathFreeObject(intgns);
    }
    else {
        GST_WARNING("Product '%s' has missing "
                    "'corrIntegration' element...dump events"
                    "will not be generated for this product.");
    }

    // find all 'lagChainSegment's, create "lc_segment_t"s for them
    GArray *lc_segments = g_array_new(FALSE, FALSE, sizeof(lc_segment_t));

    // 'product4' and 'product7' elements need to be handled slight
    // differently because the 'subproductXX' children of 'product7'
    // contain ordering information.
    if (xmlStrEqual(product->name, BAD_CAST "product4")) {
        oq->nbits = 4;
        read_all_segments(context, &streamA, &streamB, lc_segments, 0);
    }
    else /* xmlStrEqual(product->name, BAD_CAST "product7") */ {
        oq->nbits = 7;
        xmlNodePtr child = product->children;
        while (child != NULL) {
            gint subprod = -1;
            if (xmlStrEqual(child->name, BAD_CAST "subProduct00")) {
                context->node = child;
                subprod = 0;
            }
            else if (xmlStrEqual(child->name, BAD_CAST "subProduct01")) {
                context->node = child;
                subprod = 1;
            }
            else if (xmlStrEqual(child->name, BAD_CAST "subProduct10")) {
                context->node = child;
                subprod = 2;
            }
            else if (xmlStrEqual(child->name, BAD_CAST "subProduct11")) {
                context->node = child;
                subprod = 3;
            }
            if (subprod != -1) {
                read_all_segments(context, &streamA, &streamB, lc_segments,
                                  subprod);
            }
            child = child->next;
        }
    }
    g_array_sort(lc_segments, cmp_segments);

    // number of frames needs is increased by a factor of four in
    // 7-bit correlation mode
    guint corr_factor = ((oq->nbits == 4) ? 1 : 4);
    oq->num_frames *= corr_factor;

    // Get total number of recirculation blocks
    guint lc_recirc = 0;
    guint total_recirc = 0;
    guint i;
    for (i = 0; i < lc_segments->len; ++i) {
        lc_segment_t *lc_segment =
            &g_array_index(lc_segments, lc_segment_t, i);
        total_recirc += lc_segment->recirc;
        // The following checks for an unlikely possibility. I think
        // we might be able work around this restriction, but it seems
        // unlikely to be necessary.
        if (i > 0 && lc_segment->recirc != lc_recirc) {
            GST_ERROR("'lagChainSegment's with differing 'recFactor' "
                      "attributes within one product are unsupported");
        }
        lc_recirc = lc_segment->recirc;
    }
    // TODO: The following isn't correct in the case of burst dumping
    // from the correlator chips. In that case, the rate at which
    // frames arrive at the backend depends on the rate at which they
    // can be transmitted over the network, not the (short-term) rate
    // at which they are created. This gets difficult since the rate
    // at which frames from the given product arrive at the backend
    // depends on the total data bandwidth used by the baseline board
    // as a whole.
    oq->hw_dump = oq->num_bins * lc_recirc * hw_dump;

    // Create mappings for frames in this product to lagset_t. This is
    // the heart of the algorithm for ordering the lags as they
    // arrive.
    guint header_idx = 0;
    guint lags_idx = 0;
    guint recirc_x = 0;
    guint recirc_y = total_recirc - 1;
    guint lc;
    for (lc = 0; lc < lc_segments->len; ++lc) {
        lc_segment_t *lc_segment =
            &g_array_index(lc_segments, lc_segment_t, lc);
        guint nblocks = lc_segment->recirc;
        guint start_y = recirc_y + 1 - nblocks;
        lf_data_frame_word_t board_id =
            ((lc_segment->rack << 16) | (lc_segment->crate << 8)
             | lc_segment->slot);
        lf_data_frame_word_t chip_id =
            ((lc_segment->cchip[1] << 3) | lc_segment->cchip[0]);
        guint blk;
        for (blk = 0; blk < nblocks; ++blk) {
            guint ccc;
            for (ccc = lc_segment->ccc[0]; ccc <= lc_segment->ccc[1];
                 ++ccc) {
                frame_dest_t *dest = g_new(frame_dest_t, 1);
                dest->oqueue = oq;
                dest->header_idx = header_idx;
                dest->lags_idx = lags_idx;
                header_idx++;
                lags_idx += lags_per_frame;
                lf_lag_frame_t *lf = g_new(lf_lag_frame_t, 1);
                lf_lta_key_value(
                    start_y, nblocks - 1,
                    chip_id,
                    ccc,
                    recirc_x, recirc_y,
                    lc_segment->baseband[0], lc_segment->baseband[1],
                    lc_segment->subband[0], lc_segment->subband[1],
                    lc_segment->station[0], lc_segment->station[1],
                    board_id,
                    0,    // lta_bin
                    lf->key);
                g_hash_table_insert(cfg->destinations, lf, dest);
            }
            ++recirc_x;
            --recirc_y;
        }
    }
    if (lags_idx != oq->num_lags * corr_factor) {
        GST_ERROR("Number of lags computed while ordering lag frames in "
                  "product '%s' (%u) disagrees with number given by "
                  "configuration (%u)", product_id, lags_idx,
                  oq->num_lags * corr_factor);
    }
    xmlFree(product_id);
    g_array_free(lc_segments, TRUE);
}

static gint
deref_stream(xmlNodePtr product, xmlXPathContextPtr context,
             const gchar *name, stream_t *stream)
{
    gint result = -1;

    xmlChar *streamRef = xmlGetProp(product, BAD_CAST name);
    if (streamRef != NULL) {
        xmlXPathObjectPtr stream_object = xmlXPathEval(streamRef, context);
        if (stream_object != NULL) {
            if (stream_object->nodesetval->nodeNr == 1) {
                xmlNodePtr node = stream_object->nodesetval->nodeTab[0];
                read_uint_prop(node, BAD_CAST "bbId", &(stream->baseband));
                read_uint_prop(node, BAD_CAST "sbId", &(stream->subband));
                read_uint_prop(node, BAD_CAST "sId", &(stream->station));
                result = 0;
            }
            else {
                GST_WARNING("Node set for XPath expression to stream element "
                            "('%s') is not a singleton", streamRef);
            }
            xmlXPathFreeObject(stream_object);
        }
        else {
            GST_WARNING("Failed to dereference XPath expression '%s'",
                        streamRef);
        }
        xmlFree(streamRef);
    }
    else {
        xmlChar *product_id = xmlGetProp(product, BAD_CAST "id");
        GST_WARNING("Missing '%s' attribute on product '%s'",
                    name, product_id);
        if (product_id != NULL) {
            xmlFree(product_id);
        }
    }
    if (result == -1) {
        memset(stream, 0, sizeof(stream_t));
    }
    return result;
}

static void
read_uint_prop(xmlNodePtr node, const xmlChar *name, guint *val)
{
    xmlChar *attr_val = xmlGetProp(node, name);
    if (attr_val != NULL) {
        errno = 0;
        *val = strtoul(attr_val, NULL, 10);
        if (errno != 0) {
            GST_WARNING("Failed to convert '%s' attribute value (%s) to uint "
                        "value: %s", name, attr_val, g_strerror(errno));
        }
        xmlFree(attr_val);
    }
    else {
        *val = 0;
        GST_WARNING("Failed to get '%s' attribute of '%s' node",
                    name, node->name);
    }
}

static void
read_all_segments(xmlXPathContextPtr context, const stream_t *streamA,
                  const stream_t *streamB, GArray *lc_segments,
                  guint subprod)
{
    xmlXPathObjectPtr segments = xmlXPathEval(
        BAD_CAST "lagChainSegment", context);
    if (segments != NULL) {
        guint i;
        for (i = 0; i < segments->nodesetval->nodeNr; ++i) {
            xmlNodePtr segment = segments->nodesetval->nodeTab[i];
            read_segment(segment, streamA, streamB, lc_segments, subprod);
        }
        xmlXPathFreeObject(segments);
    }
}

static void
read_segment(xmlNodePtr segment, const stream_t *streamA,
             const stream_t *streamB, GArray *lc_segments,
             guint subprod)
{
    lc_segment_t lc_segment;
    memset(&lc_segment, 0, sizeof(lc_segment_t));
    lc_segment.subprod = subprod;
    lc_segment.inp_block[0] = lc_segment.inp_block[1] = G_MAXUINT;

    // identify X and Y streams
    const stream_t *streamX = NULL, *streamY = NULL;
    xmlNodePtr inp = segment->children;
    while (inp != NULL) {
        if (xmlStrEqual(inp->name, BAD_CAST "xInp")) {
            read_uint_prop(inp, BAD_CAST "inpLagBlock",
                           &lc_segment.inp_block[0]);
            xmlChar *stream = xmlGetProp(inp, "stream");
            if (stream != NULL) {
                if (stream[0] == 'A') {
                    streamX = streamA;
                }
                else if (stream[0] == 'B') {
                    streamX = streamB;
                }
                xmlFree(stream);
            }
        }
        if (xmlStrEqual(inp->name, BAD_CAST "yInp")) {
            read_uint_prop(inp, BAD_CAST "inpLagBlock",
                           &lc_segment.inp_block[1]);
            xmlChar *stream = xmlGetProp(inp, "stream");
            if (stream != NULL) {
                if (stream[0] == 'A') {
                    streamY = streamA;
                }
                else if (stream[0] == 'B') {
                    streamY = streamB;
                }
                xmlFree(stream);
            }
        }
        inp = inp->next;
    }

    if (streamX != NULL && streamY != NULL && streamX != streamY) {
        lc_segment.baseband[0] = streamX->baseband;
        lc_segment.baseband[1] = streamY->baseband;
        lc_segment.subband[0] = streamX->subband;
        lc_segment.subband[1] = streamY->subband;
        lc_segment.station[0] = streamX->station;
        lc_segment.station[1] = streamY->station;

        if (lc_segment.inp_block[0] != G_MAXUINT
            && lc_segment.inp_block[1] != G_MAXUINT) {
            read_uint_prop(segment, BAD_CAST "rack", &lc_segment.rack);
            read_uint_prop(segment, BAD_CAST "crate", &lc_segment.crate);
            read_uint_prop(segment, BAD_CAST "slot", &lc_segment.slot);
            read_uint_prop(segment, BAD_CAST "x", &lc_segment.cchip[0]);
            read_uint_prop(segment, BAD_CAST "y", &lc_segment.cchip[1]);
            read_uint_prop(segment, BAD_CAST "recFactor", &lc_segment.recirc);
            read_uint_prop(segment, BAD_CAST "firstCCC", &lc_segment.ccc[0]);
            read_uint_prop(segment, BAD_CAST "lastCCC", &lc_segment.ccc[1]);
            g_array_append_val(lc_segments, lc_segment);
        }
        else {
            // replace unset inp_block values with 0, it's a safer value
            if (lc_segment.inp_block[0] == G_MAXUINT) {
                lc_segment.inp_block[0] = 0;
            }
            if (lc_segment.inp_block[1] == G_MAXUINT) {
                lc_segment.inp_block[1] = 0;
            }
            GST_WARNING("'xInp' and/or 'yInp' have missing or invalid "
                        "'inpLagBlock' values");
        }
    }
    else {
        GST_WARNING("'xInp' and 'yInp' either not both present or "
                    "not different");
    }
}

static gint
cmp_segments(gconstpointer a, gconstpointer b)
{
    const lc_segment_t *lca = a;
    const lc_segment_t *lcb = b;
    gint result;
    result = lcb->subprod - lca->subprod; // note reverse order
    if (result == 0) {
        result = lca->inp_block[0] - lcb->inp_block[0];
    }
    return result;
}

static void
flush_all_queues(WcbeSorter *sorter)
{
    static void flush_queue(gpointer data, gpointer user_data) {
        g_return_if_fail(data != NULL);
        output_queue_t *oq = data;
        MUTEX_LOCK(oq->qmutex);
        queue_element_t *qe = g_queue_pop_tail(oq->queue);
        while (qe != NULL) {
            push_queue_element(sorter, qe);
            qe = g_queue_pop_tail(oq->queue);
        }
        MUTEX_UNLOCK(oq->qmutex);
    }
    if (sorter->config != NULL) {
        g_slist_foreach(sorter->config->queues, flush_queue, NULL);
    }
}

gboolean
wcbe_sorter_create_config(WcbeSorter *sorter, const gchar *config,
                          const gchar *name)
{
    g_return_val_if_fail(sorter != NULL, FALSE);
    g_return_val_if_fail(config != NULL, FALSE);
    g_return_val_if_fail(name != NULL, FALSE);

    gboolean result = FALSE;
    if (g_hash_table_lookup(sorter->configurations, name) == NULL) {
        wcbe_sorter_cfg_t *cfg = create_config(config, name);
        if (cfg != NULL) {
            cfg->sorter = sorter;
            g_hash_table_insert(sorter->configurations, cfg->name, cfg);
            result = TRUE;
        }
        else {
            GST_INFO("Configuration '%s' failed", name);
        }
    }
    else {
        GST_INFO("Rejected duplicate configuration name '%s'", name);
    }
    g_signal_emit(sorter, wcbe_sorter_signals[SIGNAL_CREATE_CFG],
                  0, name, result);
    return result;
}

gboolean
wcbe_sorter_activate_config(WcbeSorter *sorter, const gchar *cfg_name)
{
    g_return_val_if_fail(sorter != NULL, FALSE);
    g_return_val_if_fail(cfg_name != NULL, FALSE);

    gboolean result;
    if (sorter->config == NULL
        || strcmp(cfg_name, sorter->config->name) != 0) {
        wcbe_sorter_cfg_t *cfg =
            g_hash_table_lookup(sorter->configurations, cfg_name);
        if (cfg != NULL) {
            MUTEX_LOCK(sorter->config_mutex);
            deactivate_config(sorter, TRUE);
            sorter->config = cfg;
            sorter->new_config = TRUE;
            g_cond_signal(sorter->config_cond);
            MUTEX_UNLOCK(sorter->config_mutex);
            result = TRUE;
        }
        else {
            GST_INFO("Unable to activate non-existent configuration '%s'",
                     cfg_name);
            result = FALSE;
        }
    }
    else {
        result = TRUE;
    }
    g_signal_emit(sorter, wcbe_sorter_signals[SIGNAL_ACTIVATE_CFG],
                  0, cfg_name, result);
    return result;
}

void
wcbe_sorter_destroy_config(WcbeSorter *sorter, const gchar *cfg_name)
{
    g_return_if_fail(sorter != NULL);
    g_return_if_fail(cfg_name != NULL);

    if (sorter->config != NULL
        && strcmp(cfg_name, sorter->config->name) == 0) {
        sorter->config->destroy = TRUE;
    }
    else {
        if (!g_hash_table_remove(sorter->configurations, cfg_name)) {
            g_signal_emit(sorter,
                          wcbe_sorter_signals[SIGNAL_DESTROY_CFG],
                          0, cfg_name, FALSE);
        }
    }
}
